/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

#include <iostream.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "inpus.h"


/******************************************************************_COMMENT

This file contains definitions of the member functions for the input class.
See comments in the file inpus.h for details

COMMENT_******************************************************************/


#ifdef INPUS

//////////////////////////////////////////////////////////////////////////////

input::input(register int size, register int extend)	// create symbol table
{

Next	=	new  char[size];	// get space
Last	=	Next;			// remember the
*((char **)Next) = 0;			//    first piece
Limit	=	Next + size;		// set the end of symbol table
Next    =       (char *)(((char **)Next) + 1);
Extend	=	extend;			// extension size
Line	=	1L;			// set line number to 1
}

//////////////////////////////////////////////////////////////////////////////

void input::IncreaseInput(register int i)// extend symbol table
{
Next	=	new  char[Extend + i];	// get space
*((char **)Next) = Last;		// link to the previous piece
Limit	=	Next + Extend + i;	// set end of symbol table
Next = (char *)(((char **)Next) + 1);
}


//////////////////// extend symbol table ////////////////////
//       pointer - beginning of the string that
//      	   doesn't fit into the buffer
//
//       alogrithm: new space is allocated and the string
//       	    which doesn't fit into the old area
//       	    is copied to the new area
/////////////////////////////////////////////////////////////

char *input::ExtendInput(register char *pointer)
{
register int length = (int)(Limit - pointer);// current length of symbol table

IncreaseInput(length);			// get new area
register char *n = Next;
register char *n0 = n;
while ( length-- )
 *n++ = *pointer++;
Next	=	n;			// adjust pointer
return n0;
}

//////////////////////////////////////////////////////////////////////////////

void input::FreeInput()
{
register char *last = Last;
register char *previous;
do
{
 previous = *((char **)last);
 delete last;
} while ( (last = previous) != NULL );
}

//////////////////////////////////////////////////////////////////////////////

void	input::SkipLine()
{
register char n = '\n';
register char sym;			// get symbol here

for ( ;; )				// skip until next line
{
 if ( cin.eof() )
   return;
 sym = cin.get();
 if ( sym != n )
   continue;
 Line++;
 return;
}
}

//////////////////////////////////////////////////////////////////////////////

void	input::ErrMessage(char *x)
{
cerr << "\ninput error:" << x << ':' << oct << (int)Inp << dec << Inp << Line;
SkipLine();
}

//////////////////////////////////////////////////////////////////////////////

void	input::SkipComment()
{
register char sym = GetSym();		// read symbol here

if ( cin.eof() )
 return;
if ( sym == '*' )			// usual C-like comment
 for ( ;; )				// catch the end of comment
 {
   if ( cin.eof() )
      return;
   sym = GetSym();
   for (;;)
   {
      if ( sym != '*' )			// skip it ...
	  break;
      if ( cin.eof() )
	  return;
      sym = GetSym();
      if ( sym == '/' )
	  return;			// the end of comment is reached
   }
 }
if ( sym == '/' )			// C++-like comment
{
 SkipLine();
 return;
}
ErrMessage("skipping comment");	// error ...
}

//////////////////////////////////////////////////////////////////////////

char	*input::GetName()
{
register char   sym;			// get symbol here
register char  *pointer = Next;

if ( Next == Limit )			// overflow ...
 pointer = ExtendInput(pointer);
*Next++	=	Inp;			// store first letter
while ( !cin.eof() )			// read a word
{
 sym = GetSym();			// get the symbol
 if ( !isalnum(sym) && (sym != '_') )	// end of the name
 {
   if ( sym == '\n' )
      Line--;
   cin.putback(sym);			// stop ...
   break;
 }
 if ( Next == Limit )			// overflow ...
   pointer = ExtendInput(pointer);
 *Next++=	sym;			// store next symbol
}
if ( Next == Limit )			// overflow ...
 pointer = ExtendInput(pointer);
*Next++	=	0;			// append by a zero
return pointer;
}

//////////////////////////////////////////////////////////////////////////

char	*input::GetNumber()
{
register char   sym;			// get symbol here
register char  *pointer = Next;

if ( Next == Limit )			// overflow ...
 pointer = ExtendInput(pointer);
*Next++	=	Inp;			// store first letter
while ( !cin.eof() )			// read a word
{
 sym = GetSym();			// get the symbol
 if ( !(isdigit(sym)) )			// end of the name
 {
   if ( sym == '\n' )
      Line--;
   cin.putback(sym);			// stop ...
   break;
 }
 if ( Next == Limit )			// overflow ...
   pointer = ExtendInput(pointer);
 *Next++=	sym;			// store next symbol
}
if ( Next == Limit )			// overflow ...
 pointer = ExtendInput(pointer);
*Next++	=	0;			// append by a zero
return pointer;
}

////////////////////////////////////////////////////////////////////////////

void	input::Store()
{
if ( Next == Limit )			// overflow ...
 (void)ExtendInput(Next);
*Next++	=	Inp;			// store the letter
if ( Next == Limit )
 (void)ExtendInput(--Next);
*Next++ =	0;			// store the 0
}

#endif
