/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

#ifndef BMON
# define BMON
#endif

#include <iostream.h>
#include <stddef.h>
#include <fstream.h>
#include <stdlib.h>
#include <string.h>

#include "nodus.h"


/******************************************************************_COMMENT

This file defines methods for the standard Base Framework:

mskxBits			defines the array of bit values.

lr1TablesBMon::ParseBMon	this is a State Transition Engine.

errorBMon::RecoverBMon		defines error recovery procedure.

lexBMon::FindToken		finds the binary value of a symbol
				from that symbol's name.

COMMENT_******************************************************************/


//
// mskxBits
//
//////////////////////////////////////////////////////////////////////////
unsigned char mskxBits[8] = { 1, 2, 4, 8, 16, 32, 64, 128 };
//////////////////////////////////////////////////////////////////////////

#ifdef __V10_COMPATIBILITY__
 extern stack *CurrentStack;
#endif


//
// ParseBMon
//
//////////////////////////////////////////////////////////////////////////
brick	lr1TablesBMon::ParseBMon(register stack *s,
			         register lexBMon *l,
				 register errorBMon *er,
				 register typeProto **n)
{
#ifdef __V10_COMPATIBILITY__
CurrentStack = s;
#endif

register stackElement *stck = s->Stack;
register stackElement *high = s->StackHigh;
register const ubrick dummy = Dummy;	// pure shift indicator
register ubrick token2 = dummy;		// dummy - pure shift or after goto
register ubrick token = dummy;   	// the token
register ubrick save;
register ubrick tok;
register void *a = s->LexAtribut;
register void *a2 = s->LexAtribut2;
register typeProto *node = *n;		// current node
register int post;
register ubrick defTrigger = Tokens[Trigger].Trigger;
register int defTriggerDefined = (defTrigger != BRICKFFFF);

register int reductorDefined = (Tokens[Reductor].Trigger != BRICKFFFF);

if ( reductorDefined )
 node->DefReductor = Tokens[Reductor].Trigger;

stackSave 	     stckSave[STACK];	  	// save stack here on gotos
register stackSave  *sSaveC = stckSave;	  	// current stckSave
register stackSave  *sSaveHigh = stckSave + STACK;// limit for stckSave

s->Save = stckSave;
s->SaveEnd = stckSave;				// remember stack save

actionSave	     actSave[STACK];	  	// save actions here
register actionSave *aSaveC = actSave;	  	// current actionSave

for ( ;; )
{
 register ubrick x;		// index in various tables...
 register ubrick transit;	// transition state
 register packIt *p;            // packed column
 register ubrick r;		// row number
 register rule *act;		// action
 register ubrick trigger;

 p = Tokens + token;

 post = ( !(p->Hidden) && node->CurrentState != (long)token );// new and not hidden?
 if ( post && !(p->Skip) )
   node->CurrentState = token;				// remember it
 if ( Flag & F_ACTION )
 {
   StoreAtr(stck, a, token);				// store the attribute
   if ( defTriggerDefined && token != dummy )	      	// default trigger
   {
      s->Stack = stck;
      node->ExecuteTrigger(defTrigger, s);
   }
   trigger = p->Trigger;
   if ( trigger != BRICKFFFF )				// execute trigger
   {
      s->Stack = stck;
      node->ExecuteTrigger(trigger, s);
   }
 }

 if ( p->Skip )						// skip the token?
 {
#if defined(DEBUGON)
   if ( Flag & F_DEBUG )
      DebugSkip(token);
#endif
   stck--;						// yes...
 }
 else
   if ( post )
      if ( BIT(node->AutoSignalBits, token) )		// send auto signals
          node->SendAutoSignals(token, a);		// to some parents

cont0:

///////////////////////
// perform reduction in real time - without getting next token
///////////////////////

 if ( (ShiftHash[stck->State] >> BRICKHEADSHIFT) == B_FINALSTATE_ )
 {

#if defined(DEBUGON)		// printout the stack
   if ( Flag & F_DEBUG )
      Debug(stck);
#endif

   s->Stack = stck;		// record current stack

   transit = ReduceHash[stck->State];
   x = transit & ~BRICKHEADREDUCE;
   switch ( transit >> BRICKHEADREDUCESHIFT )
   {
      case B_REDUCE0_ACTION_:           goto action0;
      case B_REDUCEN_ACTION_:           goto actionN;
      case B_REDUCEN_ACTION_SINGLE_:    goto singleAct1;
      case B_REDUCE1_:                  goto reduce1;
      case B_REDUCE1_SINGLE_: 		goto single1;
      case B_REDUCE2MORE_:    if ( token2 != dummy )
			      {
				moveAttribute(Tokens + token2, a, a2);
				save = token = token2;
			      }
			      else
			      {
				save = dummy;
				token = l->LexBMon(a, n, node);
				if ( *n != node )       // token for another node
				{

#                                  if defined(DEBUGON)	// printout the stack
				       if ( Flag & F_DEBUG )
					    Debug(stck);
#                                  endif

				   s->Stack = stck;
				   return token;
				}
				
				moveAttribute(Tokens + token, a2, a);

#                               if defined(DEBUGON)
				   if ( Flag & F_DEBUG )
				       DebugLex(node->GetSymbol(token),
                                                node->LastSender->ObjectName);
#				endif

				if ( node->IsTranslateOn() )// translation (select)?
                                   token = TranslateToken(token);
				token2 = token;
				
				sSaveC = stckSave, aSaveC = actSave;
                                s->Save = stckSave, s->SaveEnd = stckSave;
				sSaveC->Stack = stck;
				(sSaveC++)->State = stck->State;
			      }
			      goto reduce2_1;
      case B_SINGLE0_:        goto single;
      case B_SINGLE0_ACTION_: goto singleA;
      case B_ACCEPT_:         goto accept;
      default:                goto error_;
   }
 }

 if ( token2 != dummy )		// get the next token
 {
   moveAttribute(Tokens + token2, a, a2);
   save=token=token2, token2=dummy;  // if after goto - we have one already
 }
 else
 {
   save = dummy;
   token = l->LexBMon(a, n, node);// if pure shift - get from the input stream
   if ( *n != node )		// token for another node
   {

#     if defined(DEBUGON)	// printout the stack
          if ( Flag & F_DEBUG )
	       Debug(stck);
#     endif

      s->Stack = stck;
      return token;
   }

#  if defined(DEBUGON)
      if ( Flag & F_DEBUG )
          DebugLex(node->GetSymbol(token), node->LastSender->ObjectName);
#  endif
   
   if ( node->IsTranslateOn() )// translation is on (select)?
      token = TranslateToken(token);

   sSaveC = stckSave, aSaveC = actSave, s->SaveEnd = stckSave;
 }
 

recover:			// come here after error recovery - no need to
				// get the token - it has been set by recovery

 if ( stck > high )		// stack overflow?
 {
   s->Stack = stck;
   er->StackOverflow(this);     // sorry -
   return l->Other;             //          bye-bye...
 }

#if defined(DEBUGON)		// printout the stack
   if ( Flag & F_DEBUG )
      Debug(stck);
#endif

cont:				// come here after reduction - no need to
				// check stack, get token, or printout stack

 s->Stack = stck;		// record current stack

 transit = ShiftHash[stck->State];
 switch ( transit >> BRICKHEADSHIFT )
 {

   case B_REGSTATE_:
      p = Tokens + token;			// packed column
      r = (Rows[transit + p->Local]>>p->Shift)&p->Mask;
      if ( r )
      {
	  (++stck)->State = p->Offset - r;
	  continue;
      }
      goto reduction;

   case B_SIMPLE1_:
      if ( (transit & ~BRICKHEAD) == token )
      {
	  (stck+1)->State = ReduceHash[stck->State];
	  stck++;
	  continue;
      }
      goto reduction;

   case B_STATE1_:
      SHIFT(Out1, 0, H1);
      goto reduction;

   case B_STATE2_:
      SHIFT(Out2, 0, H2); SHIFT(Out2, 2, H2);
      goto reduction;

   case B_STATE3_:
      SHIFT(Out3, 0, H3); SHIFT(Out3, 2, H3); SHIFT(Out3, 4, H3);
      goto reduction;

   case B_STATE4_:
      SHIFT(Out4, 0, H4); SHIFT(Out4, 2, H4);
      SHIFT(Out4, 4, H4); SHIFT(Out4, 6, H4);
      goto reduction;

   case B_FINALSTATE_:
      if ( token2 == dummy )
	  token2 = token;
      transit = ReduceHash[stck->State];
      x = transit & ~BRICKHEADREDUCE;
      switch ( transit >> BRICKHEADREDUCESHIFT )
      {

       case B_ERROR_:
       error_:
	  if ( Tokens[token].Skip )		// skip with no error
	  {
#if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		       DebugSkip(token);
#endif
		if ( token2 == token )
		       token2 = dummy;
		stck++;
		continue;
	  }
	  s->Stack = stck;
	  if ( !er->RecoverBMon(this, s, l, n, &token, a) )
		return token;
	  stck     = s->Stack;
	  token2   = dummy;
	  goto recover;

       case B_REDUCE2MORE_:
	  REDUCE2(x, token, ex2);
	  switch ( x >> BRICKHEADREDUCESHIFT )
	  {
	       case B_ERROR_2MORE_:
				goto error_;
	       case B_REDUCE0_ACTION2MORE_:
				x &= ~BRICKHEADREDUCE; goto action0;
	       case B_REDUCEN_ACTION2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto singleAct1;
	       case B_REDUCEN_ACTION2MORE_:
				x &= ~BRICKHEADREDUCE; goto actionN;
	       case B_REDUCE1_2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto single1;
	       case B_REDUCE1_2MORE_:
				x &= ~BRICKHEADREDUCE; // fall thru...
	  }

       case B_REDUCE1_:
       reduce1:
	  if ( reductorDefined )
	  {
		s->Stack = stck;
		node->ExecuteReductor(s, Reduce1[x]);
	  }
	  stck  -= Reduce1[x];
	  token  = Reduce1[x+1];
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  goto cont;

       case B_REDUCE1_SINGLE_:
       single1:
	  if ( reductorDefined )
	  {
		s->Stack = stck;
		node->ExecuteReductor(s, Reduce1[x] + 1);
	  }
	  stck        -= Reduce1[x];
	  tok          = Reduce1[x+1];
	  stck->State  = tok;
	  token        = DebugHash[tok]->Result;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  continue;

       case B_REDUCE0_ACTION_:
       action0:
	  if ( Flag & F_ACTION )
	  {
	      s->Stack      = stck;
	      token         = node->Action0(x, s);
	      stck->Atribut = s->Atribut;
	  }
	  else
	      token         = node->ActionSkip0(x);
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  goto cont;

       case B_REDUCEN_ACTION_SINGLE_:
       {
	  singleAct1:
	  register rule *r;
	  if ( Flag & F_ACTION )
	  {
	      s->Stack     = stck;
	      r            = node->ActionO(x, s);
	      stck        -= r->Reduce;
	      tok          = r->State;
	      stck->State  = tok;
	      token        = DebugHash[tok]->Result;
	      moveAttribute(Tokens + token, a, s->Atribut);
	  }
	  else
	  {
	      r             = node->ActionSkipN(x);
	      stck         -= r->Reduce;
	      tok           = r->State;
	      stck->State   = tok;
	      token         = DebugHash[tok]->Result;
	  }
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(r->Reduce), Debug(stck), PrintGoto();
#         endif
	  continue;
       }

       case B_REDUCEN_ACTION_:
       actionN:
	  if ( Flag & F_ACTION )
	  {
	      s->Stack      = stck;
	      act           = node->ActionN(x, s);
	      token         = act->State;
	      stck         -= act->Reduce;
	      moveAttribute(Tokens + token, a, s->Atribut);
	  }
	  else
	  {
	      act           = node->ActionSkipN(x);
	      token         = act->State;
	      stck         -= act->Reduce;
	  }
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(act->Reduce), Debug(stck), PrintGoto();
#         endif
	  goto cont;

       case B_SINGLE0_:
       single:
	  if ( reductorDefined )
	  {
		s->Stack = stck;
		node->ExecuteReductor(s, 0);
	  }

#         if defined(DEBUGON)
		if ( Flag&F_DEBUG && !(DebugFlag&F_goto) )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif

	  transit = ShiftHash[stck->State] & ~BRICKHEAD;
	  (++stck)->State = Out1[transit + 1];
	  token2 = save;
	  tok = DebugHash[stck->State]->Result;
	  p = Tokens + tok;
	  post = ( !(p->Hidden) && node->CurrentState != (long)tok );
	  if ( post && !(p->Skip) )
	      node->CurrentState = tok;			// remember it
	  if ( Flag & F_ACTION )
	  {
	      if ( defTriggerDefined )			// default trigger
   	      {
      	           s->Stack = stck;
		   node->ExecuteTrigger(defTrigger, s);
	      }
	      trigger = p->Trigger;
	      if ( trigger != BRICKFFFF )
	      {
                   s->Stack = stck;
		   node->ExecuteTrigger(trigger, s);
	      }
	  }
	  if ( p->Skip )
	  {
#if defined(DEBUGON)
	      if ( Flag & F_DEBUG )
		   DebugSkip(tok);
#endif
	      stck--;
	  }
	  else
	      if ( post )
		   if ( BIT(node->AutoSignalBits, tok) )// send auto signals
			 node->SendAutoSignals(tok, a);	// to some parents
	  goto cont0;

       case B_SINGLE0_ACTION_:
       singleA:
	  if ( Flag & F_ACTION )
	  {
	      s->Stack      = stck;
	      (void)(node->Action0(x, s));
	  }
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG && !(DebugFlag&F_goto) )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  transit = ShiftHash[stck->State] & ~BRICKHEAD;
	  (++stck)->State = Out1[transit + 1];
	  token2 = save;
	  
	  tok = DebugHash[stck->State]->Result;
	  p = Tokens + tok;
	  post = ( !(p->Hidden) && node->CurrentState != (long)tok );
	  if ( post && !(p->Skip) )
	      node->CurrentState = tok;			// remember it
	  if ( Flag & F_ACTION )
	  {
	      StoreAtr(stck, s->Atribut, tok);
	      if ( defTriggerDefined )			// default trigger
	      {
      	           s->Stack = stck;
      	           node->ExecuteTrigger(defTrigger, s);
	      }
	      trigger = p->Trigger;
	      if ( trigger != BRICKFFFF )
	      {
                   s->Stack = stck;
		   node->ExecuteTrigger(trigger, s);
	      }
	  }
	  if ( p->Skip )
	  {
#if defined(DEBUGON)
	      if ( Flag & F_DEBUG )
		   DebugSkip(tok);
#endif
	      stck--;
	  }
	  else
   	      if ( post )
		   if ( BIT(node->AutoSignalBits, tok) )// send auto signals
          	         node->SendAutoSignals(tok, s->Atribut);//2 some parents
	  goto cont0;

       case B_ACCEPT_:
       accept:
	  if ( reductorDefined )
	  {
	       s->Stack = stck;
	       node->ExecuteReductor(s, Reduce1[x]);
	  }
	  stck     -= Reduce1[x];
	  s->Stack  = stck;
	  token     = Reduce1[x+1];
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  {
		if ( defTriggerDefined )		// default trigger
		{
			s->Stack = stck;
			node->ExecuteTrigger(defTrigger, s);
		}
		register ubrick trigger = Tokens[token].Trigger;
		if ( trigger != BRICKFFFF )
		{
			s->Stack = stck;
			node->ExecuteTrigger(trigger, s);
		}
	  }
	  return token;
      }
 }

reduction:		// Come here at the state
			// where no shift is possible: reduce, but do not
			// execute actions until the series of reductions
			// finishes.  If error occures - restore stack.
			// By that we return to the state were reductions
			// has originated from: then we may proceed to
			// error recovery.  If there will be no errors -
			// execute all postponed actions and proceed in
			// usual manner.  This causious approach (contrary
			// to the non real time parsing) is essential
			// for error recovery - when erroneous token
			// comes we do not want to execute actions, but
			// want to recover at the state we are currently at.
 {
      s->Save = sSaveC;
      s->SaveEnd = sSaveC;				// remember stack save
      sSaveC->Stack = stck;		// remember current stack
      (sSaveC++)->State = stck->State;

#     if defined(DEBUGON)		// printout the stack
	  if ( Flag&F_DEBUG && Flag&F_DEBUGRECOVER)
	       DebugRecover(stckSave, stckSave);
#     endif

      moveAttribute(Tokens + token, a2, a);
      token2        = token;		// save the token

reducert:

      transit = ReduceHash[stck->State];
      x = transit & ~BRICKHEADREDUCE;
      switch ( transit >> BRICKHEADREDUCESHIFT )
      {

       case B_ERROR_:				// restore stack and token
       stoprt:
	  while ( --sSaveC >= stckSave )
	  {
#             if defined(DEBUGON)
		   if ( Flag&F_DEBUG && Flag&F_DEBUGRECOVER )
			 DebugRecover(sSaveC, stckSave);
#             endif
	      sSaveC->Stack->State = sSaveC->State;
	  }
	  sSaveC++;
	  s->SaveEnd = stckSave;
	  stck = stckSave[0].Stack;
	  token = token2;
	  moveAttribute(Tokens + token2, a, a2);
	  goto error_;

       case B_REDUCE2MORE_:			// 2 or more reductions
       reduce2_1:
	  REDUCE2(x, token, ex3);
	  switch ( x >> BRICKHEADREDUCESHIFT )
	  {
	       case B_ERROR_2MORE_:
				goto stoprt;
	       case B_REDUCEN_ACTION2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto singleAct1_1;
	       case B_REDUCE0_ACTION2MORE_:
				x &= ~BRICKHEADREDUCE; goto action0_1;
	       case B_REDUCEN_ACTION2MORE_:
				x &= ~BRICKHEADREDUCE; goto actionN_1;
	       case B_REDUCE1_2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto single1_1;
	       case B_REDUCE1_2MORE_:
				x &= ~BRICKHEADREDUCE; // fall through
	  }

       case B_REDUCE1_:				// exactly one reduction
	  if ( reductorDefined )
	  {
		aSaveC->Stack     = stck;
		aSaveC->Action    = REDUCTOR;
		aSaveC->Save      = sSaveC;
		(aSaveC++)->Index = Reduce1[x];
	  }
	  stck  -= Reduce1[x];
	  token  = Reduce1[x+1];
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  break;				// now perform goto

       case B_REDUCE1_SINGLE_:
       single1_1:
	  if ( reductorDefined )
	  {
		aSaveC->Stack     = stck;
		aSaveC->Action    = REDUCTOR;
		aSaveC->Save      = sSaveC;
		(aSaveC++)->Index = Reduce1[x] + 1;
	  }
	  stck        -= Reduce1[x];
	  tok          = Reduce1[x+1];
	  sSaveC->Stack = stck;
	  (sSaveC++)->State = stck->State;
	  stck->State  = tok;
	  token        = DebugHash[tok]->Result;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  if ( Flag & F_ACTION )
	  {
	      if ( defTriggerDefined )
	      {
			 aSaveC->Stack     = stck;
			 aSaveC->Action    = DEFTRIGGER;
			 aSaveC->Save      = sSaveC;
			 (aSaveC++)->Index = defTrigger;
	      }
	      trigger = Tokens[token].Trigger;
	      if ( trigger != BRICKFFFF )
	      {
			 aSaveC->Stack     = stck;
			 aSaveC->Action    = TRIGGER;
			 aSaveC->Save      = sSaveC;
			 (aSaveC++)->Index = trigger;
	      }
	  }
	  token = token2;
	  moveAttribute(Tokens + token2, a, a2);
	  goto shiftrt;

       case B_REDUCE0_ACTION_:
       action0_1:
	  if ( Flag & F_ACTION )
	  {
	      aSaveC->Stack     = stck;
	      aSaveC->Action    = ACTION0;
	      aSaveC->Save      = sSaveC;
	      (aSaveC++)->Index = x;
	  }
	  token = node->ActionSkip0(x);
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  break;				// now perform goto

       case B_REDUCEN_ACTION_:
       actionN_1:
	  act = node->ActionSkipN(x);
	  if ( Flag & F_ACTION )
	  {
	      aSaveC->Stack     = stck;
	      aSaveC->Action    = ACTIONN;
	      aSaveC->Save      = sSaveC;
	      (aSaveC++)->Index = x;
	  }
	  token  = act->State;
	  stck  -= act->Reduce;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(act->Reduce), Debug(stck), PrintGoto();
#         endif
	  break;                                // now perform goto

       case B_REDUCEN_ACTION_SINGLE_:
       {
	  singleAct1_1:
	  act = node->ActionSkipN(x);
	  if ( Flag & F_ACTION )
	  {
	      aSaveC->Stack     = stck;
	      aSaveC->Action    = ACTIONNS;
	      aSaveC->Save      = sSaveC;
	      (aSaveC++)->Index = x;
	  }
	  tok               = act->State;
	  stck             -= act->Reduce;
	  sSaveC->Stack     = stck;
	  (sSaveC++)->State = stck->State;
	  stck->State       = tok;
	  token             = DebugHash[tok]->Result;
	  if ( Flag & F_ACTION )			
	  {
	        if ( defTriggerDefined )
	        {
			 aSaveC->Stack     = stck;
			 aSaveC->Action    = DEFTRIGGER;
			 aSaveC->Save      = sSaveC;
			 (aSaveC++)->Index = defTrigger;
		}
		trigger = Tokens[token].Trigger;
		if ( trigger != BRICKFFFF )		
		{
			aSaveC->Stack     = stck;
			aSaveC->Action    = TRIGGER;
			aSaveC->Save      = sSaveC;
			(aSaveC++)->Index = trigger;
		}
	  }
	  token = token2;
	  moveAttribute(Tokens + token2, a, a2);
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(act->Reduce), Debug(stck), PrintGoto();
#         endif
	  goto shiftrt;               // now perform goto
       }

       case B_SINGLE0_:
	  if ( reductorDefined )
	  {
		aSaveC->Stack     = stck;
		aSaveC->Action    = REDUCTOR;
		aSaveC->Save      = sSaveC;
		(aSaveC++)->Index = 0;
	  }
#         if defined(DEBUGON)
		if ( Flag&F_DEBUG && !(DebugFlag&F_goto))
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  transit           = ShiftHash[stck->State] & ~BRICKHEAD;
	  sSaveC->Stack     = ++stck;
	  (sSaveC++)->State = stck->State;
	  stck->State       = Out1[transit + 1];
	  tok = DebugHash[stck->State]->Result;
	  p = Tokens + tok;
	  if ( Flag & F_ACTION )
	  {
	      if ( defTriggerDefined )
	      {
			 aSaveC->Stack     = stck;
			 aSaveC->Action    = DEFTRIGGER;
			 aSaveC->Save      = sSaveC;
			 (aSaveC++)->Index = defTrigger;
	      }
	      trigger = p->Trigger;
	      if ( trigger != BRICKFFFF )
	      {
		   aSaveC->Stack     = stck;
		   aSaveC->Action    = TRIGGER;
		   aSaveC->Save      = sSaveC;
		   (aSaveC++)->Index = trigger;
	      }
	  }
	  if ( p->Skip )
	  {
#if defined(DEBUGON)
	      if ( Flag & F_DEBUG )
		   DebugSkip(tok);
#endif
	      stck--;
	  }
	  else
	      if ( !(p->Hidden) && node->CurrentState != (long)tok )//newandnothidden?
	      {
		   node->CurrentState = tok;		// remember it
		   if ( BIT(node->AutoSignalBits, tok) )// send auto signals
		   {
			 aSaveC->Stack     = stck;
			 aSaveC->Action    = AUTOSIGNAL;
			 aSaveC->Save      = sSaveC;
			 (aSaveC++)->Index = tok;
		   }
	      }
	  goto shiftrt;

       case B_SINGLE0_ACTION_:
	  if ( Flag & F_ACTION )
	  {
	      aSaveC->Stack      = stck;
	      aSaveC->Action     = ACTION0;
	      aSaveC->Save       = sSaveC;
	      (aSaveC++)->Index  = x;
	  }
#         if defined(DEBUGON)
		if ( Flag&F_DEBUG && !(DebugFlag&F_goto) )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  transit           = ShiftHash[stck->State] & ~BRICKHEAD;
	  sSaveC->Stack     = ++stck;
	  (sSaveC++)->State = stck->State;
	  stck->State       = Out1[transit + 1];
	  tok = DebugHash[stck->State]->Result;
	  p = Tokens + tok;
	  if ( Flag & F_ACTION )
	  {
	      if ( defTriggerDefined )
	      {
			 aSaveC->Stack     = stck;
			 aSaveC->Action    = DEFTRIGGER;
			 aSaveC->Save      = sSaveC;
			 (aSaveC++)->Index = defTrigger;
	      }
	      trigger = p->Trigger;
	      if ( trigger != BRICKFFFF )
	      {
		   aSaveC->Stack     = stck;
		   aSaveC->Action    = TRIGGER;
		   aSaveC->Save      = sSaveC;
		   (aSaveC++)->Index = trigger;
	      }
	  }
	  if ( p->Skip )
	  {
#if defined(DEBUGON)
	      if ( Flag & F_DEBUG )
		   DebugSkip(tok);
#endif
	      stck--;
	  }
	  else
	      if ( !(p->Hidden) && node->CurrentState != (long)tok )//newandnothidden?
	      {
		   node->CurrentState = tok;		// remember it
		   if ( BIT(node->AutoSignalBits, tok) )// send auto signals
			{
		         aSaveC->Stack     = stck;
			 aSaveC->Action    = AUTOSIGNAL;
			 (aSaveC++)->Index = tok;
		   }
	      }
	  goto shiftrt;

       case B_ACCEPT_:
	  goto stoprt;
      }

#define GOTORT(out, x, h)  if ( out[transit+(x-h)] == token )	          \
			   {                                              \
			    sSaveC->Stack = ++stck;                       \
			    (sSaveC++)->State = stck->State;              \
			    stck->State = out[transit+(x-h+1)];           \
				 break;                                   \
			   }

      transit = ShiftHash[stck->State];
      switch ( transit >> BRICKHEADSHIFT )
      {
       case B_REGSTATE_:
	 p = Tokens + token;			   // packed column
	 r = (Rows[transit + p->Local]>>p->Shift)&p->Mask;
	 if ( r )
	 {
	    sSaveC->Stack = ++stck;
	    (sSaveC++)->State = stck->State;
	    stck->State = p->Offset - r;
	    break;
	 }
	 goto stoprt;

       case B_SIMPLE1_:
	 if ( (transit & ~BRICKHEAD) == token )
	 {
	    sSaveC->Stack = ++stck;
	    (sSaveC++)->State = stck->State;
	    stck->State = ReduceHash[stck->State];
	    break;
	 }
	 goto stoprt;

       case B_STATE1_:
       b_state1_:
	 GOTORT(Out1, 0, H1);
	 goto stoprt;

       case B_STATE2_:
	 GOTORT(Out2, 0, H2); GOTORT(Out2, 2, H2);
	 goto stoprt;

       case B_STATE3_:
	 GOTORT(Out3, 0, H3); GOTORT(Out3, 2, H3); GOTORT(Out3, 4, H3);
	 goto stoprt;

       case B_STATE4_:
	 GOTORT(Out4, 0, H4); GOTORT(Out4, 2, H4);
	 GOTORT(Out4, 4, H4); GOTORT(Out4, 6, H4);
	 goto stoprt;

       case B_FINALSTATE_:
	 {
	    register brick trans = ReduceHash[stck->State] >> BRICKHEADREDUCESHIFT;
	    if ( trans == B_SINGLE0_ACTION_ || trans == B_SINGLE0_ )
		goto b_state1_;
	 }                      	// fall through
       default:
	 goto stoprt;
      }


      p = Tokens + token;
      if ( Flag & F_ACTION )
      {
	  if ( defTriggerDefined )
	  {
		aSaveC->Stack     = stck;
		aSaveC->Action    = DEFTRIGGER;
		aSaveC->Save      = stckSave; //sSaveC;
		(aSaveC++)->Index = defTrigger;
	  }
	  trigger = p->Trigger;
	  if ( trigger != BRICKFFFF )
	  {
	       aSaveC->Stack     = stck;
	       aSaveC->Action    = TRIGGER;
	       aSaveC->Save      = stckSave; //sSaveC;
	       (aSaveC++)->Index = trigger;
	  }
      }
      if ( p->Skip )		// skip the token?
      {
#if defined(DEBUGON)
	  if ( Flag & F_DEBUG )
	       DebugSkip(token);
#endif
	  stck--;
      }
      else
	  if ( !(p->Hidden) && node->CurrentState != (long)token )//newandnothidden?
	  {
	       node->CurrentState = token;		// remember it
	       if ( BIT(node->AutoSignalBits, token) )	// send auto signals
	       {
		    aSaveC->Stack     = stck;
		    aSaveC->Action    = AUTOSIGNAL;
		    (aSaveC++)->Index = token;
	       }
	  }
      token = token2;
      moveAttribute(Tokens + token2, a, a2);

shiftrt:

      s->Stack = stck;			// record current stack

#     if defined(DEBUGON)		// printout the stack
	  if ( Flag & F_DEBUG )
	       Debug(stck);
#     endif

      if ( stck > high )	       // stack overflow?
      {
	  s->Stack = stck;
	  er->StackOverflow(this);     // sorry -
	  return token;                //          bye-bye...
      }
      if ( sSaveC >= sSaveHigh )       // save stack overflow?
      {
	  s->Stack = stck;
	  er->SaveStackOverflow(this); // sorry -
	  return token;                //          bye-bye...
      }

#define SHIFTRT(out, x, h) if ( out[transit+(x-h)] == token )	        \
			   {                                            \
			    if ( Flag & F_ACTION )                      \
			      actionSave::Execute(node, actSave, aSaveC,\
							       this, s);\
			    (++stck)->State = out[transit+(x-h+1)];     \
			    token2 = dummy;                             \
			    continue;                                   \
			   }

      transit = ShiftHash[stck->State];
      switch ( transit >> BRICKHEADSHIFT )
      {
       case B_REGSTATE_:
	 p = Tokens + token;			   // packed column
	 r = (Rows[transit + p->Local]>>p->Shift)&p->Mask;
	 if ( r )
	 {
	    if ( Flag & F_ACTION )
		actionSave::Execute(node, actSave, aSaveC, this, s);
	    (++stck)->State = p->Offset - r;
	    token2 = dummy;
	    continue;
	 }
	 goto reducert;

       case B_SIMPLE1_:
	 if ( (transit & ~BRICKHEAD) == token )
	 {
	    if ( Flag & F_ACTION )
		actionSave::Execute(node, actSave, aSaveC, this, s);
	    (stck+1)->State = ReduceHash[stck->State];
	    stck++;
	    token2 = dummy;
	    continue;
	 }
	 goto reducert;

       case B_STATE1_:
	 SHIFTRT(Out1, 0, H1);
	 goto reducert;

       case B_STATE2_:
	 SHIFTRT(Out2, 0, H2); SHIFTRT(Out2, 2, H2);
	 goto reducert;

       case B_STATE3_:
	 SHIFTRT(Out3, 0, H3); SHIFTRT(Out3, 2, H3); SHIFTRT(Out3, 4, H3);
	 goto reducert;

       case B_STATE4_:
	 SHIFTRT(Out4, 0, H4); SHIFTRT(Out4, 2, H4);
	 SHIFTRT(Out4, 4, H4); SHIFTRT(Out4, 6, H4);
	 goto reducert;

       default:
	 goto reducert;
      }
 }
}
}
//////////////////////////////////////////////////////////////////////////

brick mskxBMon::Success(typeProto **n, stack *s, errorBMon *e)
{
Lex->Flag |= (F_realTime|F_BMon);
return Lr1Tables->ParseBMon(s, Lex, e, n);
}
/////////////////////////////////////////////////////////////////////////

//
// RecoverBMon
//
//////////////////////////////////////////////////////////////////////////
int errorBMon::RecoverBMon(register lr1TablesBMon *tab, register stack *s,
		           register lexBMon *l, register typeProto **n,
			   register ubrick *token, register void *a)
{
register ubrick t = *token;
if ( t == tab->Accept )    			        // success - no error
 return 0;					        //        finish...

{                                                       // need translation?
 register ubrick r = tab->TranslateToken(*token);
 if ( r != *token )
 {
   *token = r;
   return 1;
 }
}

if ( t < l->Size && tab->Tokens[t].Wait )		// need to repost
{
#if defined(DEBUGON)
  if ( tab->Flag&F_DEBUG )
    tab->DebugPostPone(s->Stack, t), (*n)->PrintName();
#endif
 (*n)->PostPone(*token, a);				// repost
 return 0;                                              // exit parsing
}							// caller gets nxt msg

#if defined(DEBUGON)
 if ( tab->Flag&F_DEBUG )
   tab->DebugError(s->Stack), (*n)->PrintName();
#endif

Number++;						// one more error

ReportError(tab, s, l, token);				// report error

if ( tab->Flag&F_INTERACTIVE )				// report on cerr
{
 ostream *o = tab->Cerr;
 tab->Cerr = &cerr;
 ReportError(tab, s, l, token);
 tab->Cerr = o;
}

if ( (*n)->UserRecover(token) )				// user recover
 return 1;

if ( tab->ErrorMessage >= 0 )				// errorMessage defined
 if ( tab->FindEdge(s->Stack->State, tab->ErrorMessage) )//valid token!
 {
   *token = tab->ErrorMessage;
   return 1;                                     	// continue parsing
 }

return 0;
}
//////////////////////////////////////////////////////////////////////////


//
// FindToken
//
//////////////////////////////////////////////////////////////////////////
ubrick lexBMon::FindToken(register char *w)		// find a token
{
register int m, c, s = 0, e = NonTerminalsSize - 1;
while ( s <= e )
{
 m = (s + e) >> 1;
 if ( !(c = strcmp(w, NonTerminals[m].Word)) )
 {
   return NonTerminals[m].Result;
 }
 if ( c < 0 )
   e = m - 1;
 else
   s = m + 1;
}

s = 0, e = TerminalsSize - 1;
while ( s <= e )
{
 m = (s + e) >> 1;
 if ( !(c = strcmp(w, Terminals[m].Word)) )
 {
   return Terminals[m].Result;
 }
 if ( c < 0 )
   e = m - 1;
 else
   s = m + 1;
}

s = 0, e = LiteralsSize - 1;
while ( s <= e )
{
 m = (s + e) >> 1;
 if ( !(c = strcmp(w, Literals[m].Word)) )
 {
   return Literals[m].Result;
 }
 if ( c < 0 )
   e = m - 1;
 else
   s = m + 1;
}
return Other;
}
//////////////////////////////////////////////////////////////////////////
