/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

#include <string.h>
#include <iostream.h>
#include <ctype.h>

#include "skeletus.h"


/******************************************************************_COMMENT

This file defines data and methods for the State Transition Engine:

lr1Tables::Parse()			a stand-alone State Transition Engine.
					Used when MUSKOX compiler is invoked
					without -b and -n flags. Scans symbols
					from the input stream ahead of executing
					the actions in order to optimize the
					speed of parsing. May scan the next
					symbol and may execute an action 
					before issuing an error in the
					case of parsing errors. This is the
					fastest State Transition Engine
					
lr1Tables::ParseInRealTime()		a stand-alone real-time State Transition
					Engine. Used when MUSKOX compiler is 
					invoked without -b and -n flags. Parses
					in real-time: scans next symbol only 
					after execution of the actions,
					triggers, and reductors. Always issues
					the error before scanning next symbol
					and before executing any possible
					actions in the case of parsing errors.

actionSave::Execute			executes deferred actions, triggers,
					and reductors for real-time engine.


					search functions for symbol names
--------------------------------------------------------------------
		name --> code

lex::LookUpTerminal			finds the code of a message (terminal)
lex::GetTerminal			for a message given by its character 
					name.

lex::LookUpLiteral			finds the code of a literal message
lex::GetLiteral				(terminal) for a message given by its
					literal character representation.

lex::LookUpNonTerminal			finds the code of a state (nonterminal)
lex::GetNonTerminal			given by its character name.


		code --> name

lex::GetToken(ubrick)			finds the name of a symbol given by
					that symbol's code.
--------------------------------------------------------------------

lex::GetToken(void *)			scans next symbol from the input stream.
					Makes sense only if INPUS is defined.

lex::Lex				lexical analyzer for standalone parser

error::Recover				error recovery procedure. Used only
					for the stand-alone engine.

lr1Tables::FindEdgeAndMakeTransition	finds symbol and makes a transition.
					Used in Recover.

lr1Tables::FindEdge			finds symbol in a given state.

debugging methods			define debugging methods for lr1tables
					class.

stack::Init				stack class initializer.

stack::CalcMaxAtrSize			calculates maximum size of the 
					type's attributes.

COMMENT_******************************************************************/

#ifdef __V10_COMPATIBILITY__
#define DBTRACE
 int Pflag = 0;
 stack *CurrentStack;
#endif

//
// Parse
//
//////////////////////////////////////////////////////////////////////////
ubrick	lr1Tables::Parse(register stack *s,
			 register lex *l,
			 register ubrick (lex::*lx)(void *),
			 register actions *t,
			 register error *er)
{
#ifdef __V10_COMPATIBILITY__
CurrentStack = s;
#endif

register const ubrick dummy = Dummy;
register ubrick token = dummy;    	// next token
register ubrick token2 = dummy;		// dummy - shift, otherwise - goto
register void *a = 0;			// synthesized attribute
register void *a0 = s->LexAtribut;	// synthesized attribute
register ubrick tok;			// for 0 productions
register stackElement *stck = s->Stack;
register stackElement *high = s->StackHigh;

#ifndef FAST_PARSER

 register brick defTrigger = Tokens[Trigger].Trigger;
 register int defTriggerDefined = (defTrigger != BRICKFFFF);
 register int reductorDefined = (Tokens[Reductor].Trigger != BRICKFFFF);
 
 if ( reductorDefined )
   t->DefReductor = Tokens[Reductor].Trigger;

#endif

for ( ;; )
{
 register ubrick x;			// index in various tables...

 {
   if ( a )
      StoreAtr(stck, a, token);				// store the attribute

conta:;

#ifndef FAST_PARSER

   if ( defTriggerDefined && token != dummy )	       	// default trigger
   {
      s->Stack = stck;
      t->ExecuteTrigger(defTrigger, s);
   }
   register ubrick trigger = Tokens[token].Trigger;
   if ( trigger != BRICKFFFF )				// execute trigger
   {
      s->Stack = stck;
      t->ExecuteTrigger(trigger, s);
   }

#endif

 }

#ifndef FAST_PARSER

 if ( Tokens[token].Skip )		// skip the token?
 {
#if defined(DEBUGON)
   if ( Flag & F_DEBUG )
      DebugSkip(token);
#endif
   stck--;
 }
#endif

cont0:

 a = a0;			// synthesized attribute
 if ( token2 != dummy )		// get the next token
 {
   token=token2;		// if after goto - we have one already
   token2=dummy;
 }
 else
 {
   token = (l->*lx)(a);		// if shift - get from the input stream
   if ( t->IsTranslateOn() )      // translation (select)?
      token = TranslateToken(token);
 }

recover:			// come here after error recovery - no need to
				// get the token - it has been set by recovery

 if ( stck > high )		// stack overflow?
 {
   s->Stack = stck;
   er->StackOverflow(this);     // sorry -
   return token;                //          bye-bye...
 }

#if defined(DEBUGON)		// printout the stack
   if ( Flag & F_DEBUG )
      Debug(stck);
#endif

cont:				// come here after reduction - no need to
				// check stack, get token, or printout stack

 register ubrick transit = ShiftHash[stck->State];
 switch ( transit >> BRICKHEADSHIFT )
 {

   case B_REGSTATE_:
   {
      register packIt *p = Tokens + token;
      register ubrick r = (Rows[transit + p->Local]>>p->Shift)&p->Mask;
      if ( r )
      {
	  (++stck)->State = p->Offset - r;
	  continue;
      }
      goto reduction;
   }

   case B_SIMPLE1_:
      if ( (transit & ~BRICKHEAD) == token )
      {
	  (stck+1)->State = ReduceHash[stck->State];
	  stck++;
	  continue;
      }
      goto reduction;

   case B_STATE1_:
      SHIFT(Out1, 0, H1);
      goto reduction;

   case B_STATE2_:
      SHIFT(Out2, 0, H2); SHIFT(Out2, 2, H2);
      goto reduction;

   case B_STATE3_:
      SHIFT(Out3, 0, H3); SHIFT(Out3, 2, H3);
      SHIFT(Out3, 4, H3);
      goto reduction;

   case B_STATE4_:
      SHIFT(Out4, 0, H4); SHIFT(Out4, 2, H4);
      SHIFT(Out4, 4, H4); SHIFT(Out4, 6, H4);
						// fall through...
   reduction:					// attempt to reduce...

   case B_FINALSTATE_:
      if ( token2 == dummy )
	  token2 = token;
      transit = ReduceHash[stck->State];
      x = transit & ~BRICKHEADREDUCE;
      switch ( transit >> BRICKHEADREDUCESHIFT )
      {

       case B_ERROR_:
       error_:

#ifndef FAST_PARSER

	  if ( Tokens[token].Skip )		// skip with no error
	  {
#if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		       DebugSkip(token);
#endif
		if ( token2 == token )
		       token2 = dummy;
		stck++;				// adjust the stack
		continue;			// possible trigger...?
	  }

#endif

	  s->Stack = stck;
	  a        = a0;
	  if ( !er->Recover(this, s, l, lx, &token, a) )
		return token;
	  stck     = s->Stack;
	  token2   = dummy;
	  goto recover;

       case B_REDUCE2MORE_:
	  REDUCE2(x, token, ex1);
	  switch ( x >> BRICKHEADREDUCESHIFT )
	  {
	       case B_ERROR_2MORE_:
				goto error_;
	       case B_REDUCE0_ACTION2MORE_:
				x &= ~BRICKHEADREDUCE; goto action0;
	       case B_REDUCEN_ACTION2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto singleAct1;
	       case B_REDUCEN_ACTION2MORE_:
				x &= ~BRICKHEADREDUCE; goto actionN;
	       case B_REDUCE1_2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto single1;
	       case B_REDUCE1_2MORE_:
				x &= ~BRICKHEADREDUCE; //fall through...
	  }

       case B_REDUCE1_:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
		s->Stack = stck;
		t->ExecuteReductor(s, Reduce1[x]);
	  }
#endif
	  stck  -= Reduce1[x];
	  token  = Reduce1[x+1];
	  a      = 0;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  goto cont;

       case B_REDUCE1_SINGLE_:
       single1:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
		s->Stack = stck;
		t->ExecuteReductor(s, Reduce1[x] + 1);
	  }
#endif
	  stck        -= Reduce1[x];
	  stck->State  = Reduce1[x+1];
	  token        = DebugHash[stck->State]->Result;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  goto conta;

       case B_REDUCE0_ACTION_:
       action0:
	  s->Stack = stck;
	  token    = t->Action0(x, s);
	  a        = s->Atribut;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  goto cont;

       case B_REDUCEN_ACTION_:
       {
	  actionN:
	  register rule *r;
	  s->Stack = stck;
	  r        = t->ActionN(x, s);
	  token    = r->State;
	  stck    -= r->Reduce;
	  a        = s->Atribut;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(r->Reduce), Debug(stck), PrintGoto();
#         endif
	  goto cont;
       }

       case B_REDUCEN_ACTION_SINGLE_:
       {
	  singleAct1:
	  register rule *r;
	  s->Stack     = stck;
	  r            = t->ActionO(x, s);
	  stck        -= r->Reduce;
	  stck->State  = r->State;
	  token        = DebugHash[stck->State]->Result;
	  StoreAtr(stck, s->Atribut, token); // store the attribute
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(r->Reduce), Debug(stck), PrintGoto();
#         endif
	  goto conta;
       }

       case B_SINGLE0_:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
		s->Stack = stck;
		t->ExecuteReductor(s, 0);
	  }
#endif
#         if defined(DEBUGON)
		if ( Flag&F_DEBUG && !(DebugFlag&F_goto))
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  transit = Out1[(ShiftHash[stck->State] & ~BRICKHEAD) + 1];
	  (++stck)->State = transit;

#ifndef FAST_PARSER

	  tok = DebugHash[stck->State]->Result;
	  {
		if ( defTriggerDefined )		// default trigger
		{
			s->Stack = stck;
			t->ExecuteTrigger(defTrigger, s);
		}
		register ubrick trigger = Tokens[tok].Trigger;
		if ( trigger != BRICKFFFF )
		{
			s->Stack = stck;
			t->ExecuteTrigger(trigger, s);
		}
	  }
	  if ( Tokens[tok].Skip )
	  {
#if defined(DEBUGON)
		if ( Flag & F_DEBUG )
			DebugSkip(tok);
#endif
		stck--;
	  }

#endif

	  goto cont0;

	case B_SINGLE0_ACTION_:
	  transit = Out1[(ShiftHash[stck->State] & ~BRICKHEAD) + 1];
	  if ( token != transit )
	  {
		s->Stack = stck;
		(void)(t->Action0(x, s));
	  }
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG && !(DebugFlag&F_goto) )
			Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  (++stck)->State = transit;

	  tok = DebugHash[stck->State]->Result;
	  {
		StoreAtr(stck, s->Atribut, tok);

#ifndef FAST_PARSER

		if ( defTriggerDefined )		// default trigger
		{
			s->Stack = stck;
			t->ExecuteTrigger(defTrigger, s);
		}
		register ubrick trigger = Tokens[tok].Trigger;
		if ( trigger != BRICKFFFF )
		{
			s->Stack = stck;
			t->ExecuteTrigger(trigger, s);
		}

#endif

	  }

#ifndef FAST_PARSER

	  if ( Tokens[tok].Skip )
	  {
#if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		       DebugSkip(tok);
#endif
		stck--;
	  }

#endif

	  goto cont0;

       case B_ACCEPT_:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
	       s->Stack = stck;
	       t->ExecuteReductor(s, Reduce1[x]);
	  }
#endif
	  stck     -= Reduce1[x];
	  s->Stack  = stck;
	  token     = Reduce1[x+1];
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
#ifndef FAST_PARSER

	  {
		if ( defTriggerDefined )		// default trigger
		{
			s->Stack = stck;
			t->ExecuteTrigger(defTrigger, s);
		}
		register ubrick trigger = Tokens[token].Trigger;
		if ( trigger != BRICKFFFF )
		{
			s->Stack = stck;
			t->ExecuteTrigger(trigger, s);
		}
	  }

#endif
	  return token;
      }
 }
}
}
//////////////////////////////////////////////////////////////////////////


//
// ParseInRealTime
//
//////////////////////////////////////////////////////////////////////////
ubrick	lr1Tables::ParseInRealTime(register stack *s,
				   register lex *l,
				   register ubrick (lex::*lx)(void *),
				   register actions *t,
				   register error *er)
{
register stackElement *stck = s->Stack;
register stackElement *high = s->StackHigh;
register ubrick const dummy = Dummy;// error returned by the lexer
register ubrick token2 = dummy;	// dummy - pure shift, otherwise - after goto
register ubrick save;
register ubrick token = dummy;   // the token
register ubrick tok;
register void *a = s->LexAtribut;
register void *a2 = s->LexAtribut2;

#ifndef FAST_PARSER
register brick defTrigger = Tokens[Trigger].Trigger;
register int defTriggerDefined = (defTrigger != BRICKFFFF);

register int reductorDefined = (Tokens[Reductor].Trigger != BRICKFFFF);

if ( reductorDefined )
 t->DefReductor = Tokens[Reductor].Trigger;

#endif

stackSave 	     stckSave[STACK];	  	// save stack here on gotos
register stackSave  *sSaveC = stckSave;	    	// current stckSave
register stackSave  *sSaveHigh = stckSave + STACK;// limit for stckSave

s->Save = stckSave;
s->SaveEnd = stckSave;				// remember stack save

actionSave	     actSave[STACK];	  	// save actions here
register actionSave *aSaveC = actSave;	   	// current actionSave

for ( ;; )
{
 register ubrick x;		// index in various tables...
 register ubrick transit;	// transition state
 register packIt *p;            // packed column
 register ubrick r;		// row number
 register rule *act;		// action

#ifndef FAST_PARSER
 register ubrick trigger;
#endif

 if ( Flag & F_ACTION )
 {
   StoreAtr(stck, a, token);				// store the attribute

#ifndef FAST_PARSER

   if ( defTriggerDefined && token != dummy )	      	// default trigger
   {
      s->Stack = stck;
      t->ExecuteTrigger(defTrigger, s);
   }
   trigger = Tokens[token].Trigger;
   if ( trigger != BRICKFFFF )				// execute trigger
   {
      s->Stack = stck;
      t->ExecuteTrigger(trigger, s);
   }

#endif

 }

#ifndef FAST_PARSER

 if ( Tokens[token].Skip )		// skip the token?
 {
#if defined(DEBUGON)
   if ( Flag & F_DEBUG )
      DebugSkip(token);
#endif
   stck--;
}

#endif

cont0:

///////////////////////
// perform reduction in real time - without getting next token
///////////////////////

 if ( (ShiftHash[stck->State] >> BRICKHEADSHIFT) == B_FINALSTATE_ )
 {

#if defined(DEBUGON)		// printout the stack
   if ( Flag & F_DEBUG )
      Debug(stck);
#endif

   transit = ReduceHash[stck->State];
   x = transit & ~BRICKHEADREDUCE;
   switch ( transit >> BRICKHEADREDUCESHIFT )
   {
      case B_REDUCE0_ACTION_: 		goto action0;
      case B_REDUCEN_ACTION_: 		goto actionN;
      case B_REDUCEN_ACTION_SINGLE_:    goto singleAct1;
      case B_REDUCE1_:        		goto reduce1;
      case B_REDUCE1_SINGLE_: 		goto single1;
      case B_REDUCE2MORE_:    if ( token2 != dummy )
			      {
				moveAttribute(Tokens + token2, a, a2);
				save = token = token2;
			      }
			      else
			      {
				save = dummy;
				token = (l->*lx)(a);
				if ( t->IsTranslateOn() )// translation (select)?
                                   token = TranslateToken(token);
				token2 = token;
				moveAttribute(Tokens + token, a2, a);
			      }
			      sSaveC = stckSave, aSaveC = actSave;
			      s->Save = stckSave, s->SaveEnd = stckSave;
			      sSaveC->Stack = stck;
			      (sSaveC++)->State = stck->State;
			      goto reduce2_1;
      case B_SINGLE0_:        goto single;
      case B_SINGLE0_ACTION_: goto singleA;
      case B_ACCEPT_:         goto accept;
      default:                goto error_;
   }
 }

 if ( token2 != dummy )		// get the next token
 {
   moveAttribute(Tokens + token2, a, a2);
   save=token=token2, token2=dummy;  // if after goto - we have one already
 }
 else
 {
   save = dummy, token = (l->*lx)(a);//if pure shift - get from the input stream
   if ( t->IsTranslateOn() )// translation (select)?
      token = TranslateToken(token);
 }

 sSaveC = stckSave, aSaveC = actSave, s->SaveEnd = s->Save;

recover:			// come here after error recovery - no need to
				// get the token - it has been set by recovery

 if ( stck > high )		// stack overflow?
 {
   s->Stack = stck;
   er->StackOverflow(this);     // sorry -
   return token;                //          bye-bye...
 }

#if defined(DEBUGON)		// printout the stack
   if ( Flag & F_DEBUG )
      Debug(stck);
#endif

cont:				// come here after reduction - no need to
				// check stack, get token, or printout stack

 transit = ShiftHash[stck->State];
 switch ( transit >> BRICKHEADSHIFT )
 {

   case B_REGSTATE_:
      p = Tokens + token;			// packed column
      r = (Rows[transit + p->Local]>>p->Shift)&p->Mask;
      if ( r )
      {
	  (++stck)->State = p->Offset - r;
	  continue;
      }
      goto reduction;

   case B_SIMPLE1_:
      if ( (transit & ~BRICKHEAD) == token )
      {
	  (stck+1)->State = ReduceHash[stck->State];
	  stck++;
	  continue;
      }
      goto reduction;

   case B_STATE1_:
      SHIFT(Out1, 0, H1);
      goto reduction;

   case B_STATE2_:
      SHIFT(Out2, 0, H2); SHIFT(Out2, 2, H2);
      goto reduction;

   case B_STATE3_:
      SHIFT(Out3, 0, H3); SHIFT(Out3, 2, H3); SHIFT(Out3, 4, H3);
      goto reduction;

   case B_STATE4_:
      SHIFT(Out4, 0, H4); SHIFT(Out4, 2, H4);
      SHIFT(Out4, 4, H4); SHIFT(Out4, 6, H4);
      goto reduction;

   case B_FINALSTATE_:
      if ( token2 == dummy )
	  moveAttribute(Tokens + token, a2, a), token2 = token;
      transit = ReduceHash[stck->State];
      x = transit & ~BRICKHEADREDUCE;
      switch ( transit >> BRICKHEADREDUCESHIFT )
      {

       case B_ERROR_:
       error_:

#ifndef FAST_PARSER

	  if ( Tokens[token].Skip )		// skip with no error
	  {
#if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		       DebugSkip(token);
#endif
		if ( token2 == token )
		       token2 = dummy;
		stck++;				// adjust the stack
		continue;			// possible trigger...?
	  }

#endif

	  s->Stack = stck;
	  if ( !er->Recover(this, s, l, lx, &token, a) )
		return token;
	  stck     = s->Stack;
	  token2   = dummy;
	  goto recover;

       case B_REDUCE2MORE_:
	  REDUCE2(x, token, ex2);
	  switch ( x >> BRICKHEADREDUCESHIFT )
	  {
	       case B_ERROR_2MORE_:
				goto error_;
	       case B_REDUCE0_ACTION2MORE_:
				x &= ~BRICKHEADREDUCE; goto action0;
	       case B_REDUCEN_ACTION2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto singleAct1;
	       case B_REDUCEN_ACTION2MORE_:
				x &= ~BRICKHEADREDUCE; goto actionN;
	       case B_REDUCE1_2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto single1;
	       case B_REDUCE1_2MORE_:
				x &= ~BRICKHEADREDUCE; // fall thru...
	  }

       case B_REDUCE1_:
       reduce1:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
		s->Stack = stck;
		t->ExecuteReductor(s, Reduce1[x]);
	  }
#endif
	  stck  -= Reduce1[x];
	  token  = Reduce1[x+1];
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  goto cont;

       case B_REDUCE1_SINGLE_:
       single1:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
		s->Stack = stck;
		t->ExecuteReductor(s, Reduce1[x] + 1);
	  }
#endif
	  stck        -= Reduce1[x];
	  tok          = Reduce1[x+1];
	  stck->State  = tok;
	  token        = DebugHash[tok]->Result;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  continue;

       case B_REDUCE0_ACTION_:
       action0:
	  if ( Flag & F_ACTION )
	  {
	      s->Stack      = stck;
	      token         = t->Action0(x, s);
	      moveAttribute(Tokens + token, a, s->Atribut);
	  }
	  else
	      token         = t->ActionSkip0(x);
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  goto cont;

       case B_REDUCEN_ACTION_:
       actionN:
	  if ( Flag & F_ACTION )
	  {
	      s->Stack      = stck;
	      act           = t->ActionN(x, s);
	      token         = act->State;
	      stck         -= act->Reduce;
	      moveAttribute(Tokens + token, a, s->Atribut);
	  }
	  else
	  {
	      act           = t->ActionSkipN(x);
	      token         = act->State;
	      stck         -= act->Reduce;
	  }
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(act->Reduce), Debug(stck), PrintGoto();
#         endif
	  goto cont;

       case B_REDUCEN_ACTION_SINGLE_:
       {
	  singleAct1:
	  register rule *r;
	  if ( Flag & F_ACTION )
	  {
	      s->Stack     = stck;
	      r            = t->ActionO(x, s);
	      stck        -= r->Reduce;
	      tok          = r->State;
	      stck->State  = tok;
	      token        = DebugHash[tok]->Result;
	      moveAttribute(Tokens + token, a, s->Atribut);
	  }
	  else
	  {
	      r             = t->ActionSkipN(x);
	      stck         -= r->Reduce;
	      tok           = r->State;
	      stck->State   = tok;
	      token         = DebugHash[tok]->Result;
	  }
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(r->Reduce), Debug(stck), PrintGoto();
#         endif
	  continue;
       }

       case B_SINGLE0_:
       single:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
		s->Stack = stck;
		t->ExecuteReductor(s, 0);
	  }
#endif
#         if defined(DEBUGON)
		if ( Flag&F_DEBUG && !(DebugFlag&F_goto) )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  transit = Out1[(ShiftHash[stck->State] & ~BRICKHEAD) + 1];
	  (++stck)->State = transit;
	  token2 = save;

#ifndef FAST_PARSER

	  tok = DebugHash[stck->State]->Result;
	  if ( Flag & F_ACTION )
	  {
	      if ( defTriggerDefined )			// default trigger
	      {
		   s->Stack = stck;
		   t->ExecuteTrigger(defTrigger, s);
	      }
	      register ubrick trigger = Tokens[tok].Trigger;
	      if ( trigger != BRICKFFFF )
	      {
		   s->Stack = stck;
		   t->ExecuteTrigger(trigger, s);
	      }
	  }
	  if ( Tokens[tok].Skip )
	  {
#if defined(DEBUGON)
	      if ( Flag & F_DEBUG )
		   DebugSkip(tok);
#endif
	      stck--;
	  }
#endif

	  goto cont0;

       case B_SINGLE0_ACTION_:
       singleA:
	  if ( Flag & F_ACTION )
	  {
	      s->Stack      = stck;
	      (void)(t->Action0(x, s));
	  }
#         if defined(DEBUGON)
		if ( Flag&F_DEBUG && !(DebugFlag&F_goto) )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  transit = Out1[(ShiftHash[stck->State] & ~BRICKHEAD) + 1];
	  (++stck)->State = transit;
	  tok = DebugHash[stck->State]->Result;
	  token2 = save;

	  if ( Flag & F_ACTION )
	  {
	      StoreAtr(stck, s->Atribut, tok);

#ifndef FAST_PARSER

	      if ( defTriggerDefined )			// default trigger
	      {
		   s->Stack = stck;
		   t->ExecuteTrigger(defTrigger, s);
	      }
	      register ubrick trigger = Tokens[tok].Trigger;
	      if ( trigger != BRICKFFFF )
	      {
		   s->Stack = stck;
		   t->ExecuteTrigger(trigger, s);
	      }

#endif

	  }

#ifndef FAST_PARSER

	  if ( Tokens[tok].Skip )
	  {
#if defined (DEBUGON)
	      if ( Flag & F_DEBUG )
		   DebugSkip(tok);
#endif
	      stck--;
	  }

#endif

	  goto cont0;

       case B_ACCEPT_:
       accept:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
	       s->Stack = stck;
	       t->ExecuteReductor(s, Reduce1[x]);
	  }
#endif
	  stck     -= Reduce1[x];
	  s->Stack  = stck;
	  token     = Reduce1[x+1];
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
#ifndef FAST_PARSER

	  {
		if ( defTriggerDefined )		// default trigger
		{
			s->Stack = stck;
			t->ExecuteTrigger(defTrigger, s);
		}
		register ubrick trigger = Tokens[token].Trigger;
		if ( trigger != BRICKFFFF )
		{
			s->Stack = stck;
			t->ExecuteTrigger(trigger, s);
		}
	  }

#endif
	  return token;
      }
 }

reduction:		// Come here at the state
			// where no shift is possible: reduce, but do not
			// execute actions until the series of reductions
			// finishes.  If error occures - restore stack.
			// By that we return to the state were reductions
			// has originated from: then we may proceed to
			// error recovery.  If there will be no errors -
			// execute all postponed actions and proceed in
			// usual manner.  This causious approach (contrary
			// to the non real time parsing) is essential
			// for error recovery - when erroneous token
			// comes we do not want to execute actions, but
			// want to recover at the state we are currently at.
 {
      s->Save = sSaveC;
      s->SaveEnd = sSaveC;				// remember stack save
      sSaveC->Stack = stck;		// remember current stack
      (sSaveC++)->State = stck->State;

      moveAttribute(Tokens + token, a2, a);
      token2 = token;		// save the token

reducert:

      transit = ReduceHash[stck->State];
      x = transit & ~BRICKHEADREDUCE;
      switch ( transit >> BRICKHEADREDUCESHIFT )
      {

       case B_ERROR_:				// restore stack and token
       stoprt:
	  while ( --sSaveC >= stckSave )
	      sSaveC->Stack->State = sSaveC->State;
	  sSaveC++;
	  s->SaveEnd = stckSave;
	  stck = stckSave[0].Stack;
	  token = token2;
	  moveAttribute(Tokens + token2, a, a2);
	  goto error_;

       case B_REDUCE2MORE_:			// 2 or more reductions
	  reduce2_1:
	  REDUCE2(x, token, ex3);
	  switch ( x >> BRICKHEADREDUCESHIFT )
	  {
	       case B_ERROR_2MORE_:
				goto stoprt;
	       case B_REDUCEN_ACTION2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto singleAct1_1;
	       case B_REDUCE0_ACTION2MORE_: 
				x &= ~BRICKHEADREDUCE; goto action0_1;
	       case B_REDUCEN_ACTION2MORE_:
				x &= ~BRICKHEADREDUCE; goto actionN_1;
	       case B_REDUCE1_2MORE_SINGLE_:
				x &= ~BRICKHEADREDUCE; goto single1_1;
	       case B_REDUCE1_2MORE_:
				x &= ~BRICKHEADREDUCE; // fall throuhg
	  }

       case B_REDUCE1_:				// exactly one reduction
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
		aSaveC->Stack     = stck;
		aSaveC->Action    = REDUCTOR;
		aSaveC->Save      = sSaveC;
		(aSaveC++)->Index = Reduce1[x];
	  }
#endif
	  stck  -= Reduce1[x];
	  token  = Reduce1[x+1];
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif
	  break;				// now perform goto

       case B_REDUCE1_SINGLE_:
       single1_1:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
		aSaveC->Stack     = stck;
		aSaveC->Action    = REDUCTOR;
		aSaveC->Save      = sSaveC;
		(aSaveC++)->Index = Reduce1[x] + 1;
	  }
#endif
	  stck        -= Reduce1[x];
	  tok          = Reduce1[x+1];
	  sSaveC->Stack = stck;
	  (sSaveC++)->State = stck->State;
	  stck->State  = tok;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(Reduce1[x]), Debug(stck), PrintGoto();
#         endif

#ifndef FAST_PARSER

	  token        = DebugHash[tok]->Result;
	  if ( Flag & F_ACTION )
	  {
	      if ( defTriggerDefined )
	      {
			 aSaveC->Stack     = stck;
			 aSaveC->Action    = DEFTRIGGER;
			 aSaveC->Save      = sSaveC;
			 (aSaveC++)->Index = defTrigger;
	      }
	      trigger = Tokens[token].Trigger;
	      if ( trigger != BRICKFFFF )
	      {
			 aSaveC->Stack     = stck;
			 aSaveC->Action    = TRIGGER;
			 aSaveC->Save      = sSaveC;
			 (aSaveC++)->Index = trigger;
	      }
	  }

#endif

	  token = token2;
	  moveAttribute(Tokens + token2, a, a2);
	  goto shiftrt;

       case B_REDUCE0_ACTION_:
       action0_1:
	  if ( Flag & F_ACTION )
	  {
	      aSaveC->Stack     = stck;
	      aSaveC->Action    = ACTION0;
	      aSaveC->Save      = sSaveC;
	      (aSaveC++)->Index = x;
	  }
	  token = t->ActionSkip0(x);
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  break;				// now perform goto

       case B_REDUCEN_ACTION_:
       actionN_1:
	  act = t->ActionSkipN(x);
	  if ( Flag & F_ACTION )
	  {
	      aSaveC->Stack     = stck;
	      aSaveC->Action    = ACTIONN;
	      aSaveC->Save      = sSaveC;
	      (aSaveC++)->Index = x;
	  }
	  token  = act->State;
	  stck  -= act->Reduce;
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(act->Reduce), Debug(stck), PrintGoto();
#         endif
	  break;                                // now perform goto

       case B_REDUCEN_ACTION_SINGLE_:
       {
	  singleAct1_1:
	  act = t->ActionSkipN(x);
	  if ( Flag & F_ACTION )
	  {
	      aSaveC->Stack     = stck;
	      aSaveC->Action    = ACTIONNS;
	      aSaveC->Save      = sSaveC;
	      (aSaveC++)->Index = x;
	  }
	  tok               = act->State;
	  stck             -= act->Reduce;
	  sSaveC->Stack     = stck;
	  (sSaveC++)->State = stck->State;
	  stck->State       = tok;

#ifndef FAST_PARSER

	  token             = DebugHash[tok]->Result;
	  if ( Flag & F_ACTION )
	  {
		if ( defTriggerDefined )
		{
			aSaveC->Stack     = stck;
			aSaveC->Action    = DEFTRIGGER;
			aSaveC->Save      = sSaveC;
			(aSaveC++)->Index = defTrigger;
		}
		trigger = Tokens[token].Trigger;
		if ( trigger != BRICKFFFF )
		{
			aSaveC->Stack     = stck;
			aSaveC->Action    = TRIGGER;
			aSaveC->Save      = sSaveC;
			(aSaveC++)->Index = trigger;
		}
	  }

#endif

	  token = token2;
	  moveAttribute(Tokens + token2, a, a2);
#         if defined(DEBUGON)
		if ( Flag & F_DEBUG )
		      Debug(act->Reduce), Debug(stck), PrintGoto();
#         endif
	  goto shiftrt;               // now perform goto
       }

       case B_SINGLE0_:
#ifndef FAST_PARSER
	  if ( reductorDefined )
	  {
		aSaveC->Stack     = stck;
		aSaveC->Action    = REDUCTOR;
		aSaveC->Save      = sSaveC;
		(aSaveC++)->Index = 0;
	  }
#endif
#         if defined(DEBUGON)
		if ( Flag&F_DEBUG && !(DebugFlag&F_goto) )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  transit           = ShiftHash[stck->State] & ~BRICKHEAD;
	  sSaveC->Stack     = ++stck;
	  (sSaveC++)->State = stck->State;
	  stck->State       = Out1[transit + 1];

#ifndef FAST_PARSER

	  tok = DebugHash[stck->State]->Result;
	  if ( Flag & F_ACTION )
	  {
	      if ( defTriggerDefined )
	      {
		    aSaveC->Stack     = stck;
		    aSaveC->Action    = DEFTRIGGER;
		    aSaveC->Save      = sSaveC;
		    (aSaveC++)->Index = defTrigger;
	      }
	      trigger = Tokens[tok].Trigger;
	      if ( trigger != BRICKFFFF )
	      {
		   aSaveC->Stack     = stck;
		   aSaveC->Action    = TRIGGER;
		   aSaveC->Save      = sSaveC;
		   (aSaveC++)->Index = trigger;
	      }
	  }
	  if ( Tokens[tok].Skip )
	  {
#if defined(DEBUGON)
	      if ( Flag & F_DEBUG )
		   DebugSkip(tok);
#endif
	      stck--;
	  }

#endif

	  goto shiftrt;

       case B_SINGLE0_ACTION_:
	  if ( Flag & F_ACTION )
	  {
	      aSaveC->Stack      = stck;
	      aSaveC->Action     = ACTION0;
	      aSaveC->Save       = sSaveC;
	      (aSaveC++)->Index  = x;
	  }
#         if defined(DEBUGON)
		if ( Flag&F_DEBUG && !(DebugFlag&F_goto) )
		      Debug((ubrick)0), Debug(stck), PrintGoto();
#         endif
	  transit           = ShiftHash[stck->State] & ~BRICKHEAD;
	  sSaveC->Stack     = ++stck;
	  (sSaveC++)->State = stck->State;
	  stck->State       = Out1[transit + 1];

#ifndef FAST_PARSER

	  tok = DebugHash[stck->State]->Result;
	  if ( Flag & F_ACTION )
	  {
	      if ( defTriggerDefined )		          
	      {						  
		    aSaveC->Stack     = stck;		
		    aSaveC->Action    = DEFTRIGGER;
		    aSaveC->Save      = sSaveC;
		    (aSaveC++)->Index = defTrigger;
	      }
	      trigger = Tokens[tok].Trigger;
	      if ( trigger != BRICKFFFF )
	      {
		   aSaveC->Stack     = stck;
		   aSaveC->Action    = TRIGGER;
		   aSaveC->Save      = sSaveC;
		   (aSaveC++)->Index = trigger;
	      }
	  }
	  if ( Tokens[tok].Skip )
	  {
#if defined(DEBUGON)
	      if ( Flag & F_DEBUG )
		   DebugSkip(tok);
#endif
	      stck--;
	  }

#endif

	  goto shiftrt;

       case B_ACCEPT_:
	  goto stoprt;
      }

#ifndef FAST_PARSER

#define GOTORT(out, x, h)  if ( out[transit+(x-h)] == token )	          \
			   {                                              \
			    sSaveC->Stack = ++stck;                       \
			    (sSaveC++)->State = stck->State;              \
			    stck->State = out[transit+(x-h+1)];           \
	    		    if ( Flag & F_ACTION )			  \
			    {                                             \
			      if ( defTriggerDefined )		          \
			      {						  \
				 aSaveC->Stack     = stck;		  \
				 aSaveC->Action    = DEFTRIGGER; 	  \
				 aSaveC->Save      = sSaveC;              \
				 (aSaveC++)->Index = defTrigger;	  \
			      }                                           \
			      trigger = Tokens[token].Trigger;            \
			      if ( trigger != BRICKFFFF )	          \
			      {						  \
				 aSaveC->Stack     = stck;		  \
				 aSaveC->Action    = TRIGGER;		  \
				 aSaveC->Save      = sSaveC;              \
				 (aSaveC++)->Index = trigger;		  \
			      }                                           \
			    }						  \
			    token = token2;                               \
			    moveAttribute(Tokens + token2, a, a2);	  \
			    break;                                        \
			   }

#else

#define GOTORT(out, x, h)  if ( out[transit+(x-h)] == token )	          \
			   {                                              \
			    sSaveC->Stack = ++stck;                       \
			    (sSaveC++)->State = stck->State;              \
			    stck->State = out[transit+(x-h+1)];           \
			    token = token2;                               \
	                    moveAttribute(Tokens + token2, a, a2);	  \
			    break;                                        \
			   }

#endif

      tok = token;
      transit = ShiftHash[stck->State];
      switch ( transit >> BRICKHEADSHIFT )
      {
       case B_REGSTATE_:
	 p = Tokens + token;			   // packed column
	 r = (Rows[transit + p->Local]>>p->Shift)&p->Mask;
	 if ( r )
	 {
	    sSaveC->Stack = ++stck;
	    (sSaveC++)->State = stck->State;
	    stck->State = p->Offset - r;

#ifndef FAST_PARSER

	    if ( Flag & F_ACTION )
	    {
		if ( defTriggerDefined )
		{
		     aSaveC->Stack     = stck;		
		     aSaveC->Action    = DEFTRIGGER;
		     (aSaveC++)->Index = defTrigger;	
		}
		trigger = Tokens[token].Trigger;
		if ( trigger != BRICKFFFF )
		{
	             aSaveC->Stack     = stck;
		     aSaveC->Action    = TRIGGER;
	             (aSaveC++)->Index = trigger;
		}
	    }

#endif

	    token = token2;
	    moveAttribute(Tokens + token2, a, a2);
	    break;
	 }
	 goto stoprt;

       case B_SIMPLE1_:
	 if ( (transit & ~BRICKHEAD) == token )
	 {
	    sSaveC->Stack = ++stck;
	    (sSaveC++)->State = stck->State;
	    stck->State = ReduceHash[stck->State];

#ifndef FAST_PARSER

	    if ( Flag & F_ACTION )
	    {
		if ( defTriggerDefined )
		{
		     aSaveC->Stack     = stck;
		     aSaveC->Action    = DEFTRIGGER;
		     (aSaveC++)->Index = defTrigger;
		}
		trigger = Tokens[token].Trigger;
		if ( trigger != BRICKFFFF )
		{
		     aSaveC->Stack     = stck;
		     aSaveC->Action    = TRIGGER;
		     (aSaveC++)->Index = trigger;
		}
	    }

#endif

	    token = token2;
	    moveAttribute(Tokens + token2, a, a2);
	    break;
	 }
	 goto stoprt;

       case B_STATE1_:
       b_state1_:
	 GOTORT(Out1, 0, H1);
	 goto stoprt;

       case B_STATE2_:
	 GOTORT(Out2, 0, H2); GOTORT(Out2, 2, H2);
	 goto stoprt;

       case B_STATE3_:
	 GOTORT(Out3, 0, H3); GOTORT(Out3, 2, H3);
	 GOTORT(Out3, 4, H3);
	 goto stoprt;

       case B_STATE4_:
	 GOTORT(Out4, 0, H4); GOTORT(Out4, 2, H4);
	 GOTORT(Out4, 4, H4); GOTORT(Out4, 6, H4);
	 goto stoprt;

       case B_FINALSTATE_:
	 {
	    register brick trans = ReduceHash[stck->State] >> BRICKHEADREDUCESHIFT;
	    if ( trans == B_SINGLE0_ACTION_ || trans == B_SINGLE0_ )
		goto b_state1_;
	 }                      	// fall through
       default:
	 goto stoprt;
      }

#ifndef FAST_PARSER

      if ( Tokens[tok].Skip )		// skip the token?
      {
#if defined(DEBUGON)
	  if ( Flag & F_DEBUG )
	       DebugSkip(tok);
#endif
	  stck--;
      }

#endif

shiftrt:

#     if defined(DEBUGON)		// printout the stack
	  if ( Flag & F_DEBUG )
	       Debug(stck);
#     endif

      if ( stck > high )	       // stack overflow?
      {
	  s->Stack = stck;
	  er->StackOverflow(this);     // sorry -
	  return token;                //          bye-bye...
      }
      if ( sSaveC >= sSaveHigh )       // save stack overflow?
      {
	  s->Stack = stck;
	  er->SaveStackOverflow(this); // sorry -
	  return token;                //          bye-bye...
      }

#define SHIFTRT(out, x, h) if ( out[transit+(x-h)] == token )	      \
			   {                                          \
			    if ( Flag & F_ACTION )                    \
			      actionSave::Execute(t, actSave, aSaveC, \
							     this, s);\
			    (++stck)->State = out[transit+(x-h+1)];   \
			    token2 = dummy;                           \
			    continue;                                 \
			   }

      transit = ShiftHash[stck->State];
      switch ( transit >> BRICKHEADSHIFT )
      {
       case B_REGSTATE_:
	 p = Tokens + token;			   // packed column
	 r = (Rows[transit + p->Local]>>p->Shift)&p->Mask;
	 if ( r )
	 {
	    if ( Flag & F_ACTION )
		actionSave::Execute(t, actSave, aSaveC, this, s);
	    (++stck)->State = p->Offset - r;
	    token2 = dummy;
	    continue;
	 }
	 goto reducert;

       case B_SIMPLE1_:
	 if ( (transit & ~BRICKHEAD) == token )
	 {
	    if ( Flag & F_ACTION )
		actionSave::Execute(t, actSave, aSaveC, this, s);
	    (stck+1)->State = ReduceHash[stck->State];
	    stck++;
	    token2 = dummy;
	    continue;
	 }
	 goto reducert;

       case B_STATE1_:
	 SHIFTRT(Out1, 0, H1);
	 goto reducert;

       case B_STATE2_:
	 SHIFTRT(Out2, 0, H2); SHIFTRT(Out2, 2, H2);
	 goto reducert;

       case B_STATE3_:
	 SHIFTRT(Out3, 0, H3); SHIFTRT(Out3, 2, H3);
	 SHIFTRT(Out3, 4, H3);
	 goto reducert;

       case B_STATE4_:
	 SHIFTRT(Out4, 0, H4); SHIFTRT(Out4, 2, H4);
	 SHIFTRT(Out4, 4, H4); SHIFTRT(Out4, 6, H4);
	 goto reducert;

       default:
	 goto reducert;
      }
 }
}
}
//////////////////////////////////////////////////////////////////////////


//
// Execute
//
//////////////////////////////////////////////////////////////////////////
void actionSave::Execute(register actions *t, register actionSave *b,
			 register actionSave *e, register lr1Tables *tb,
			 register stack *s)
{
register stackElement *se = s->Stack;	// save
register stackSave *ss = s->SaveEnd;
register rule *r;
register ubrick token;

while ( b < e )
{
 s->Stack = b->Stack;
 s->SaveEnd = b->Save;
 switch ( b->Action )
 {
	case ACTIONN:
	{
		r = t->ActionN((b++)->Index, s);
		moveAttribute(tb->Tokens + r->State,
				 s->Stack[1 - r->Reduce].Atribut, s->Atribut);
		break;
	}

	case ACTIONNS:
	{
		r = t->ActionO((b++)->Index, s);
		moveAttribute(tb->Tokens + tb->DebugHash[r->State]->Result,
				 s->Stack[- r->Reduce].Atribut, s->Atribut);
		break;
	}

	case ACTION0:
		(void)(t->Action0(b->Index, s));
		moveAttribute(tb->Tokens + tb->DebugHash[(b++)->Stack->State]
								       ->Result,
			      s->Stack[1].Atribut, s->Atribut);
		break;

	case DEFTRIGGER:
	{
		t->ExecuteTrigger((b++)->Index, s);
		break;
	}

	case TRIGGER:
	{
		token = tb->DebugHash[b->Stack->State]->Result;
		if ( !(tb->Tokens[token].Hidden) && t->GetCurrentState() !=
									token )
			 t->SetState(token);
		t->ExecuteTrigger((b++)->Index, s);
		break;
	}

	case REDUCTOR:
	{
		t->ExecuteReductor(s, (b++)->Index);
		break;
	}

	case AUTOSIGNAL:
		t->SendAutoSignals((b++)->Index, s->Atribut);
		break;
 }
}
s->Stack = se;				// restore
s->SaveEnd = ss;
}
//////////////////////////////////////////////////////////////////////////

ubrick stack::GetSave(register stackElement *s)
{
register stackSave *sb = Save;
register stackSave *se = SaveEnd;
while ( sb < se )
{
 if (sb->Stack == s)
   return sb->State;
 sb++;
};
return s->State;
}


//
// LookUpLiteral
//
//////////////////////////////////////////////////////////////////////////
ubrick lex::LookUpLiteral(void *)
{
register int m, c, s = 0, e = LiteralsSize - 1;
register char *w = Token->Word;

while ( s <= e )
{
 m = (s + e) >> 1;
 if ( !(c = strcmp(w, Literals[m].Word)) )
 {
   Result = Literals[m].Result;
   return 1;
 }
 if ( c < 0 )  e = m - 1;
 else          s = m + 1;
}
Result = Other;
return 0;
}
//////////////////////////////////////////////////////////////////////////

//
// GetLiteral
//
//////////////////////////////////////////////////////////////////////////
ubrick lex::GetLiteral(register char *w)
{
register int m, c, s = 0, e = LiteralsSize - 1;

while ( s <= e )
{
 m = (s + e) >> 1;
 if ( !(c = strcmp(w, Literals[m].Word)) )
   return Literals[m].Result;
 if ( c < 0 )  e = m - 1;
 else          s = m + 1;
}
return Other;
}
//////////////////////////////////////////////////////////////////////////

//
// LookUpTerminal
//
//////////////////////////////////////////////////////////////////////////
ubrick lex::LookUpTerminal(void *)
{
register int m, c, s = 0, e = TerminalsSize - 1;
register char *w = Token->Word;

while ( s <= e )
{
 m = (s + e) >> 1;
 if ( !(c = strcmp(w, Terminals[m].Word)) )
 {
   Result = Terminals[m].Result;
   return 1;
 }
 if ( c < 0 )  e = m - 1;
 else          s = m + 1;
}
Result = Other;
return 0;
}
//////////////////////////////////////////////////////////////////////////

//
// GetTerminal
//
//////////////////////////////////////////////////////////////////////////
ubrick lex::GetTerminal(register char *w)
{
register int m, c, s = 0, e = TerminalsSize - 1;

while ( s <= e )
{
 m = (s + e) >> 1;
 if ( !(c = strcmp(w, Terminals[m].Word)) )
   return Terminals[m].Result;
 if ( c < 0 )  e = m - 1;
 else          s = m + 1;
}
return Other;
}
//////////////////////////////////////////////////////////////////////////

//
// LookUpNonTerminal
//
//////////////////////////////////////////////////////////////////////////
ubrick lex::LookUpNonTerminal(void *)
{
register int m, c, s = 0, e = NonTerminalsSize - 1;
register char *w = Token->Word;

while ( s <= e )
{
 m = (s + e) >> 1;
 if ( !(c = strcmp(w, NonTerminals[m].Word)) )
 {
   Result = NonTerminals[m].Result;
   return 1;
 }
 if ( c < 0 )  e = m - 1;
 else          s = m + 1;
}
Result = Other;
return 0;
}
//////////////////////////////////////////////////////////////////////////

//
// GetNonTerminal
//
//////////////////////////////////////////////////////////////////////////
ubrick lex::GetNonTerminal(register char *w)
{
register int m, c, s = 0, e = NonTerminalsSize - 1;

while ( s <= e )
{
 m = (s + e) >> 1;
 if ( !(c = strcmp(w, NonTerminals[m].Word)) )
   return NonTerminals[m].Result;
 if ( c < 0 )  e = m - 1;
 else          s = m + 1;
}
return Other;
}
//////////////////////////////////////////////////////////////////////////


ubrick lex::SetAlternative(void *)
{
register char c = *(Token->Word);
if ( isalpha(c) || c == '_' )
 return Alternative;
return Other;
}


//
// GetToken
//
//////////////////////////////////////////////////////////////////////////
char* lex::GetToken(ubrick val)
{
  ubrick i;

  i = 0;
  while ( i < LiteralsSize )
  {
     if ( Literals[i].Result == val )
	 return Literals[i].Word;
     i++;
  }

  i = 0;
  if ( Terminals != Literals )
     while ( i < TerminalsSize )
     {
	 if ( Terminals[i].Result == val )
	      return Terminals[i].Word;
	 i++;
     }
  i = 0;
  while ( i < NonTerminalsSize )
  {
     if ( NonTerminals[i].Result == val )
	 return NonTerminals[i].Word;
     i++;
  }
  return "U N K N O W N";
}
//////////////////////////////////////////////////////////////////////////


//
// GetToken
//
//////////////////////////////////////////////////////////////////////////
ubrick lex::GetToken(void *)
{
#ifdef INPUS
for ( ;; )
{
 switch( (int)(Input->GetSym()) )
 {
  case 0   :    Result = EndOfStream; return 1;	   // return to Parse on EOF
  case 1   :
  case 2   :
  case 3   :
  case 4   :
  case 5   :
  case 6   :
  case 7:
  case '\b':    Store(); return 0;
  case '\t':	
  case '\n':	
  case '\v':
  case '\f':
  case '\r':    break;
  case 14  :
  case 15  :
  case 16  :
  case 17  :
  case 18  :
  case 19  :
  case 20  :
  case 21  :
  case 22  :
  case 23  :
  case 24  :
  case 25  :
  case 26  :
  case 27  :
  case 28  :
  case 29  :
  case 30  :
  case 31  :    Store(); return 0;
  case ' ' :	break;
  case '!' :
  case '\"':
  case '#' :
  case '$' :
  case '%' :
  case '&' :
  case '\'':
  case '(' :
  case ')' :
  case '*' :
  case '+' :
  case ',' :	Store(); return 0;
  case '-' :	Store(Input->GetName()); return 0;
  case '.' :	Store(); return 0;
  case '/' :	Input->SkipComment(); break;
  case '0' :
  case '1' :
  case '2' :
  case '3' :
  case '4' :
  case '5' :
  case '6' :
  case '7' :
  case '8' :
  case '9' :	Store(Input->GetNumber()); return 0;
  case ':' :
  case ';' :
  case '<' :
  case '=' :
  case '>' :
  case '?' :
  case '@' :	Store(); return 0;
  case 'A' :
  case 'B' :
  case 'C' :
  case 'D' :
  case 'E' :
  case 'F' :
  case 'G' :
  case 'H' :
  case 'I' :
  case 'J' :
  case 'K' :
  case 'L' :
  case 'M' :
  case 'N' :
  case 'O' :
  case 'P' :
  case 'Q' :
  case 'R' :
  case 'S' :
  case 'T' :
  case 'U' :
  case 'V' :
  case 'W' :
  case 'X' :
  case 'Y' :
  case 'Z' :	Store(Input->GetName()); return 0;
  case '[' :
  case '\\':
  case ']' :
  case '^' :	Store(); return 0;
  case '_' :	Store(Input->GetName()); return 0;
  case '`' :	Store(); return 0;
  case 'a' :
  case 'b' :
  case 'c' :
  case 'd' :
  case 'e' :
  case 'f' :
  case 'g' :
  case 'h' :
  case 'i' :
  case 'j' :
  case 'k' :
  case 'l' :
  case 'm' :
  case 'n' :
  case 'o' :
  case 'p' :
  case 'q' :
  case 'r' :
  case 's' :
  case 't' :
  case 'u' :
  case 'v' :
  case 'w' :
  case 'x' :
  case 'y' :
  case 'z' :    Store(Input->GetName()); return 0;
  case '{' :
  case '|' :
  case '}' :
  case '~' :
  default  :	Store(); return 0;
 }					// end of switch
}					// end of read loop
#else
return 0;
#endif
}
//////////////////////////////////////////////////////////////////////////

//
// GetToken
//
//////////////////////////////////////////////////////////////////////////
ubrick   lex::GetToken(register char *w)
{
register ubrick r;
if ( (r = GetLiteral(w)) == Other )
 if ( (r = GetTerminal(w)) == Other )
   if ( (r = GetNonTerminal(w)) == Other )
       ;
return r;
}
//////////////////////////////////////////////////////////////////////////


//
// Lex
//
//////////////////////////////////////////////////////////////////////////
ubrick lex::Lex(void *a)
{
if ( Result == EndOfStream )	// if lex is called after EndOfstream
{                               // has been swallowed
 if ( !(Flag&F_realTime) )      // and we do not parse in real time
 {
   Result = Other;		// return Other - we will force
   return Other;		// Parse() to go to the accept state.
 }
}
if ( (this->*Lex1)(a) )
 return Result;
if ( (this->*Lex2)(a) )
 return Result;
if ( (this->*Lex3)(a) )
 return Result;
if ( (this->*Lex4)(a) )
 return Result;
if ( (this->*Lex5)(a) )
 return Result;
return SetAlternative(a);
}
//////////////////////////////////////////////////////////////////////////

//
// Recover
//
//////////////////////////////////////////////////////////////////////////
int	error::Recover(register lr1Tables *tab, register stack *s,
		       register lex *l, register ubrick (lex::*lx)(void *),
		       register ubrick *token, register void *a)
{
if ( *token == (ubrick)(tab->Accept) )    		// success - no error
 return 0;					        //        finish...

#if defined(DEBUGON)
 if ( tab->Flag & F_DEBUG )
   tab->DebugError(s->Stack);
#endif

{                                                       // need translation?
 register ubrick r = tab->TranslateToken(*token);
 if ( r != *token )
 {
   *token = r;
   return 1;
 }
}

Number++;						// one more error

ReportError(tab, s, l, token);				// report error

if ( Number >= ErrorLimit() )					// too many errors
 return 0;

if ( UserRecover(tab, s, l, lx, token) )		// user defined
 return UserResult;                                     //        recovery...

if ( tab->ErrorMessage >= 0 )				// errorMessage defined
 if ( tab->FindEdge(s->Stack->State, tab->ErrorMessage) )//valid token!
 {
   *token = tab->ErrorMessage;
   return 1;                                     	// continue parsing
 }

if ( Flag & F_NOSKIP )		// stay at the same state
{
#if defined(DEBUGON)
   if ( tab->Flag & F_DEBUG )
      tab->Debug(s->Stack), tab->DebugWait();
#endif
 register ubrick t;
 *token = t = (l->*lx)(a);
 return t != tab->EndOfStream;// stop parsing on eof
}

register stackElement *stck = s->Stack;
register stackElement *limit = s->StackLimit;

for ( ;; )		    // find a state in the stack with outgoing ERROR
{                           // or reach the start state (limit == stck)
			    // or NoPop option is on.

 if ( tab->FindEdgeAndMakeTransition(&stck, tab->Error) ||
      limit == stck || Flag & F_NOPOP )
 {
   register ubrick t;					// next input token
   register int loop = 0;                               // skipped tokens cnt

#  if defined(DEBUGON)
      if ( tab->Flag & F_DEBUG )
	  tab->Debug(stck), tab->DebugWait();
#  endif

   while ( (t = (l->*lx)(a)) != tab->EndOfStream )      // now skip tokens
   {

      if ( tab->FindEdge(stck->State, t) )		// valid token!
      {
	  *token = t;					// return next token
	  s->Stack = stck;				// update stack
	  return 1;                                     // continue parsing
      }

      if ( ++loop >= SkipTokens )			// too many tokens
	  return 0;					// - stop parsing

#     if defined(DEBUGON)
	  if ( tab->Flag & F_DEBUG )
	       tab->Debug(stck), tab->DebugSkip(t), tab->DebugWait();
#     endif

   }
   return 0;		// end of stream has been reached - stop parsing
 }

#if defined(DEBUGON)
 if ( tab->Flag & F_DEBUG )
   tab->DebugBack(stck);
#endif

 stck--;                // pop the stack
}
}
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
#define TRANSIT(out, x) if ( out[transit+x] == token )		  \
			{                                         \
			 (++(*stck))->State=out[transit+x+1];     \
			 return 1;                                \
			}
//////////////////////////////////////////////////////////////////////////


//
// FindEdgeAndMakeTransition
//
// We do not undertake any action when making a transition.
// So, for example, we discourage to define a rule ERROR : {action};
//
//////////////////////////////////////////////////////////////////////////

int	lr1Tables::FindEdgeAndMakeTransition(register stackElement **stck,
					     register ubrick token)
{
register ubrick transit = ShiftHash[(*stck)->State];

switch ( transit >> BRICKHEADSHIFT )
{
   case B_REGSTATE_:
	{
	  register packIt *p = Tokens + token;
	  register int r = (Rows[transit + p->Local]>>p->Shift)&p->Mask;
	  if ( r )
	  {
	     (++(*stck))->State = p->Offset - r;
	     return 1;
	  }
	  return 0;
	}

   case B_SIMPLE1_:
      if ( (transit & ~BRICKHEAD) == token )
      {
	 (*stck+1)->State = ReduceHash[(*stck)->State];
	 (*stck)++;
	 return 1;
      }
      return 0;

   case B_STATE1_:
	transit &= ~BRICKHEAD; 
	TRANSIT(Out1, 0);
        return 0;

   case B_STATE2_:
        transit &= ~BRICKHEAD; 
	TRANSIT(Out2, 0); 
	TRANSIT(Out2, 2);
	return 0;

   case B_STATE3_:
	transit &= ~BRICKHEAD;
        TRANSIT(Out3, 0); 
	TRANSIT(Out3, 2);
	TRANSIT(Out3, 4); 
	return 0;

   case B_STATE4_:
	transit &= ~BRICKHEAD;
	TRANSIT(Out4, 0);
	TRANSIT(Out4, 2);
	TRANSIT(Out4, 4);
	TRANSIT(Out4, 6);
	return 0;

   default:
	transit = ReduceHash[(*stck)->State];
	switch ( transit >> BRICKHEADREDUCESHIFT )
	{
	 case B_SINGLE0_:
	 case B_SINGLE0_ACTION_:
	  transit = Out1[(ShiftHash[(*stck)->State] & ~BRICKHEAD) + 1];
	  if ( token == DebugHash[transit]->Result )
	  {
		 (++(*stck))->State = transit;
		 return 1;
	  }
	}

   return 0;
}
}
//////////////////////////////////////////////////////////////////////////

//
// FindEdge
//
// We do not make a transition for SINGLE0_ or SINGLE0_ACTION_
// states. This is because we discourage to accept from the input stream
// nonterminals that normally should appear as a result of
// reductions of the rules like N :; or N : {action};
//
//////////////////////////////////////////////////////////////////////////
int	lr1Tables::FindEdge(register ubrick state, register ubrick token)
{
register ubrick transit = ShiftHash[state];

switch ( transit >> BRICKHEADSHIFT )
{
   case B_REGSTATE_:
   {
      register packIt *p = Tokens + token;
      return (Rows[transit + p->Local]>>p->Shift)&p->Mask;
   }

   case B_SIMPLE1_:
      return (transit & ~BRICKHEAD) == token;

   case B_STATE1_:
      transit &= ~BRICKHEAD;
      return Out1[transit] == token;

   case B_STATE2_:
      transit &= ~BRICKHEAD;
      return Out2[transit] == token || Out2[transit+2] == token;

   case B_STATE3_:
      transit &= ~BRICKHEAD;
      return Out3[transit] == token || Out3[transit+2] == token ||
	     Out3[transit+4] == token;

   case B_STATE4_:
      transit &= ~BRICKHEAD;
      return Out4[transit] == token || Out4[transit+2] == token ||
	     Out4[transit+4] == token || Out4[transit+6] == token;

   default:
      return 0;
}
}
//////////////////////////////////////////////////////////////////////////

//
// Translation methods
//
//////////////////////////////////////////////////////////////////////////
ubrick lr1Tables::TranslateToken(register ubrick r)
{
register ubrick *u = Translate;
if ( u )
{
 register ubrick d = Dummy;
 for ( ;; )
 {
   register ubrick v = *u;
   if ( v == d )
      break;
   if ( v == r )
   {
      r = *++u;
      break;
   }
   u += 2;
 }
}
return r;
}

//
// Debugging methods
//
//////////////////////////////////////////////////////////////////////////
#ifdef __V10_COMPATIBILITY__
 static int FirstError = 1;
#endif

void error::ReportError(register lr1Tables *tab, register stack *stk,
			register lex *l, register ubrick *token)
{

#ifndef __V10_COMPATIBILITY__
 *(tab->Cerr) << endl << "MUSKOX v1.2 : invalid symbol #" << Number;
#else
 *(tab->Cerr) << endl << "syntax error #" << Number;
#endif

*(tab->Cerr) << ", " << l->GetToken(*token) << '=' << *token;

if ( *(tab->LanguageName) )
 *(tab->Cerr) << " <type " << tab->LanguageName << '>';

#ifdef INPUS
 *(tab->Cerr) << " line " << l->Input->Line;
#endif

*(tab->Cerr) << " \nStack:";
for (stackElement* sp = stk->StackLimit; sp <= stk->Stack; sp++)
 *(tab->Cerr) << "\n\t" << tab->DebugHash[sp->State]->Word
	      << "->s" << sp->State;
 
*(tab->Cerr) << endl;   //flush the output

#ifdef __V10_COMPATIBILITY__
if (FirstError) 
{
 FirstError = 0;
 cerr << endl << "^Gmessage error - see the current log file for run time messages(default message.log)" 
      << endl;
}
#endif

}

void error::StackOverflow(register lr1Tables *tab)
{
*(tab->Cerr) << "\nstack overflow" << endl;
}

void error::SaveStackOverflow(register lr1Tables *tab)
{
*(tab->Cerr) << "\nsave stack overflow" << endl;
}

//////////////////////
#ifdef __V10_COMPATIBILITY__
#if defined(BMON)                       // for base monitor...
extern char* CurrentNodeName;
#endif

void  lr1Tables::Debug(register stackElement *stck) // printout shift/goto
{
#ifdef  DBTRACE
if(strcmp(LanguageName, "InHeartBeat")
	&& strcmp(LanguageName, "admin_cmd")
        && strcmp(LanguageName, "CommDaemon")
  )
{
 if (Pflag) 
 {
   int first_time = 1;
   extern stack* CurrentStack;
   *Cerr << "\n---------------\n";
   *Cerr << ' ' << DebugHash[stck->State]->Word << "->s" << stck->State;
   #if defined(BMON)                       // for base monitor...
   *Cerr << "\t\t\t" << LanguageName << "::" << CurrentNodeName;
   #endif
   for (stackElement* sp=CurrentStack->StackLimit; sp<=stck; sp++)
   {
      if (first_time) 
      {
         *Cerr << "\n ParseStack: ";
         first_time = 0;
      } 
      else
	 *Cerr << "\n             ";
      *Cerr << DebugHash[sp->State]->Word ;
   }
   *Cerr << "\n---------------" << endl;
 }
}
#endif
}

#else

void  lr1Tables::Debug(register stackElement *stck) // printout shift/goto
{
*Cerr << ' ' << DebugHash[stck->State]->Word << "->s" << stck->State << endl;
DebugFlag &= ~F_goto;

if ( !Stack )			// regular ARM
 return;

stackElement *s = Stack->StackLimit;
*Cerr << "\n Stack: ";
for ( ;; )
{
 *Cerr << 's' << s->State << ' ' << DebugHash[s->State]->Word ;
 if ( ++s <= stck )
   *Cerr << "\n             ";
 else
   break;
}
*Cerr << "\n---------------" << endl;
}
#endif
///////////////////////

///////////////////////
#ifdef __V10_COMPATIBILITY__
void  lr1Tables::PrintGoto()				// printout goto sign
{
#ifdef  DBTRACE
if(strcmp(LanguageName, "InHeartBeat")
        && strcmp(LanguageName, "admin_cmd")
        && strcmp(LanguageName, "CommDaemon")
  )
 if (Pflag)
   *Cerr << " !!!";
#endif
}

void  lr1Tables::Debug(register ubrick x)  		// printout reduction
{
#ifdef  DBTRACE
if(strcmp(LanguageName, "InHeartBeat")
        && strcmp(LanguageName, "admin_cmd")
        && strcmp(LanguageName, "CommDaemon")
  )
 if (Pflag)
   *Cerr << endl << "...Reduce " << x;
#endif
}

void  lr1Tables::DebugError(register stackElement *stck)// printout error
{                                                       // condition
#ifdef  DBTRACE
if (Pflag)
 *Cerr << endl << "...Error:<" << DebugHash[stck->State]->Word << '='
       << DebugHash[stck->State]->Result << '>';
#endif
}

#else

void  lr1Tables::PrintGoto()				// printout goto sign
{
*Cerr << "...Goto:";
DebugFlag |= F_goto;
}

void  lr1Tables::Debug(register ubrick x)  		// printout reduction
{
*Cerr << endl << "...Reduce " << x;
}

void  lr1Tables::DebugError(register stackElement *stck)// printout error
{                                                       // condition
*Cerr << "\n...Error:<" << DebugHash[stck->State]->Word << '='
       << DebugHash[stck->State]->Result << '>' << endl;
}

#endif
////////////////////

void  lr1Tables::DebugPostPone(register stackElement *stck,// printout postponed
                               register brick t)           // message
{
*Cerr << "\n...PostPone:<" << DebugHash[stck->State]->Word << '='
       << DebugHash[stck->State]->Result << '>'
       << " token = <" << t << '>' << endl;
}


void  lr1Tables::DebugSkip(register brick t)            // skip token
{							// (may be on error)
*Cerr << "\n...Skip token = <" << t << '>' << endl;
}

void  lr1Tables::DebugBack(register stackElement *stck) // pop the stack
{							// on error
*Cerr << endl << "...Error: Pop the stack ";
Debug(stck);
}

void  lr1Tables::DebugWait()				// wait for token
{                                                       // on error
*Cerr << "\n...Error:waiting for token " << endl;
}

void  lr1Tables::DebugRecover(stackSave *from, stackSave *to)
{
*Cerr << "\n...Next recover from '" << (void *)from << "'";
if ( from )
{
 if ( from->Stack )
 {
   *Cerr << " from->Stack->State '" << from->Stack->State << "'";
   Debug(from->Stack);
 }
 *Cerr << "\n from->State '" << from->State << "'";
}
*Cerr << "\n -- to '" << (void *)to << "'";
if ( to )
{
 if ( to->Stack )
 {
   *Cerr << " to->Stack->State '" << to->Stack->State << "'";
   Debug(to->Stack);
 }
 *Cerr << "\n to->State '" << to->State << "'";
}
*Cerr << endl;
}

void lr1Tables::DebugLex(register char *s, register char *o)
{
*Cerr << '<' << s << '@' << o << '>';
}
//////////////////////////////////////////////////////////////////////////


//
// stack
//
//////////////////////////////////////////////////////////////////////////
long stack::TotalAtrMaxSize = 0;
long stack::Flag = PASSBYREFERENCE;

void stack::Init(register int depth, register packIt *p,
	         register long sz, register ubrick state)
{
Depth = depth;
PackIt = p;
Size = sz;
State = state;

register int mxs = CalcMaxAtrSize(p, sz);
if ( mxs > TotalAtrMaxSize )		   // the total max size of attributes
 TotalAtrMaxSize = mxs;
StackLimit = Stack = new stackElement[depth];
StackHigh = Stack + depth - 3;		   // reserve for SHIFTRT (+ 1 more)
AtrMaxSize = mxs;

Atribut = new char[mxs];
LexAtribut = new char[mxs];
LexAtribut2 = new char[mxs];
register int max = mxs*depth;
AtrStackLimit = new char[max];

register stackElement *s = Stack + depth; // fill in attribute positions
register char *sab = AtrStackLimit;
register char *sae = sab + max;
while ( (sae -= mxs) >= sab )
 (--s)->Atribut = sae;

Stack->State = state;
Stack->Atribut = AtrStackLimit;

Save = 0;
SaveEnd = 0;
}
//////////////////////////////////////////////////////////////////////////


//
// CalcMaxAtrSize
//
//////////////////////////////////////////////////////////////////////////
int stack::CalcMaxAtrSize(register packIt *p, register long s)
{
register ubrick result = 0;
register ubrick mx;

while ( s-- )
{
 if ( !p->Copy )				// no copy constructor
 {
   if ( p->Move.SizeOfAttribute > result )
      result = p->Move.SizeOfAttribute;
 }
 else						// copy constructor
 {
   if ( (mx = (int)(*(p->Move.Method))(0, 0)) > result ) // length of attribute
      result = mx;
 }
 p++;
}
return result;
}
//////////////////////////////////////////////////////////////////////////

///////////////////
int mskx::Success()
{ return Lr1Tables->Success(Stack, Lex, Lx, Node, Error); }
///////////////////

//
// real-time parsing
//
///////////////////
int mskx::SuccessInRealTime()
{ return Lr1Tables->SuccessInRealTime(Stack, Lex, Lx, Node, Error); }
///////////////////

//
// generator
//
///////////////////
#ifdef GENON
int mskx::Generate(long length, long number, ostream &f)
{ return Lr1Tables->Generate(Lex, length, number, f); }
#endif
///////////////////

//
// actions virtual functions
//
///////////////////
ubrick actions::GetCurrentState() { return BRICKFFFF; } // redefine these two in
void   actions::SetState(ubrick) {}		        // typeproto
void   actions::SendAutoSignals(brick, void *) {}       // defined in typeProto

brick actions::Action0(ubrick, stack *){return 0;}; // a:{ action };
brick actions::ActionSkip0(ubrick){return 0;};      // skip a:{ action };
rule *actions::ActionN(ubrick, stack *){return 0;}; // a:something { action };
rule *actions::ActionO(ubrick, stack *){return 0;}; // single goto ActionN
rule *actions::ActionSkipN(ubrick){return 0;};      // skip a:something{action};
void  actions::ExecuteTrigger(ubrick, stack *) {}   // execute a trigger
void  actions::ExecuteReductor(stack *, ubrick) {}  // execute a reductor

ubrick lex::UserLex(void *){return 0;}
ubrick lex::Nop(void *){return 1;}
ubrick lex::PureUserLex(void *){return Result;}

int error::ErrorLimit(){return 25;}

int lr1Tables::Success(register stack *s, register lex *l,
                       register ubrick (lex::*lx)(void *),
                       register actions *t, register error *e)
{  
l->Flag &= ~F_realTime;
return Parse(s, l, lx, t, e) == Accept && !ErrorProc->Number;
}  

int lr1Tables::SuccessInRealTime(register stack *s, register lex *l,
                                 register ubrick (lex::*lx)(void *),
                                 register actions *t, register error *e)
{  
l->Flag |= F_realTime;
return ParseInRealTime(s, l, lx, t, e) == Accept && !ErrorProc->Number;
}  

keyWord *lr1Tables::Find(register ubrick token)
{
register packIt *p = Tokens + token;
return DebugHash[p->Offset - (p->Mask != 0)];
}

