/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <iostream.h>
#include <ctype.h>
#include <string.h>

#include "atribus.h"
#include "inpus.h"
#include "skeletus.h"

#include "sym.h"
#include "sym.m"				// grammar symbols

#ifdef __MWERKS__
#include <console.h>
#endif

//
// redefine error reporting
//

extern symLex sLex;

int	myError::UserRecover(lr1Tables *, stack *, lex *, // no user recovery
			ubrick (lex::*)(void *), ubrick *)
{
return 0;
}

						// report error and print stack
void	myError::ReportError(lr1Tables *tab, stack *stck, lex *l, ubrick *token)
{
cerr << "\nsym: Syntax error at <" << l->GetToken(*token)
     << ">, ParseStack:";

register stackElement *sp = stck->StackLimit;

while ( sp <= stck->Stack )
{
 cerr << "\n\t" << sp->State << " - " << tab->DebugHash[sp->State]->Word;
 sp++;
}
}


//
// redefine error limit
//
						// up to 10 errors
int myError::ErrorLimit()
{
return 10;
}


//
// redefine lexical analyzer
//

ubrick myLex::PureUserLex(void *) {return 0;}

ubrick symLex::PureUserLex(register void *a)
{
for ( ;; )
{
 register char *s = Pointer;
 switch (*Pointer++ )
 {
   case '+': return __l_43__sym;
   case '-': return __l_45__sym;
   case '(': return __l_40__sym;
   case ')': return __l_41__sym;
   case '*': return __l_42__sym;
   case '/': return __l_47__sym;
   case ':': Expression = Pointer; return __l_58__sym; // remember expression
   case ';': return __l_59__sym;
   case '=': return __l_61__sym;
   case ' ':
   case '\t':
   case '\n': continue;
   case 0:    return _ENDoFsTREAM_sym;
   default:   if ( isalpha(*s) )		// name
              {
                while ( *Pointer )
                {
                   if ( !isalnum(*Pointer) )
                       break;
                   Pointer++;
                }
		register int off = (int)(Pointer - s);
		*(char **)a = new char [off + 1]; // remember name
		strncpy(*(char **)a, s, off);
		(*(char **)a)[off] = 0;
		return _NAME_sym;
              }
              if ( !isdigit(*s) )		// lexical error
		return _OTHER_sym;
	      while ( isdigit(*Pointer) )	// number
		Pointer++;
	      *((long *)a) = atol(s);           // remember number
	      return _NUMBER_sym;
 }
}
}


//
// declare generated objects
//

extern  mskx  mskx_sym;			// generated parser,lexer,error
extern  lex   lexer_sym;
extern  error error_sym;

extern  sym  this_sym;			// generated type object

//
// redefine lexer and error objects
//

symLex sLex(&lexer_sym);			// redefine lexer and error
myError myError_sym(&error_sym);

//
// initialize static members
//

int lr1Tables::Flag = 0;
ostream *lr1Tables::Cerr = &cerr;

//
// sym parser
//
//
// accepts arithmetic formulae that are then used 
// to calculate an arithmetic expression 
//
// a formula begins with a name separated by ':' from the expression 
// which may involve names of other formalue, long numbers,
// operations addition, subtraction, unary plus 
// and minus, and grouping in parenthesis; for example:
//
//	x : a + b / 2 ;
//
// an arithmetic expression is the last item in the input;
// it looks like the right handside of a formula followed by '=';
// it is calculated recursively, for example:
//
// 	x : a + b / 2;
//	y : x * 2;
//	a : 2;
//	b : 8;
//	y + 9 =			----	yeilds 21
//
// prepare the formalue and expressions in a string on 
// the command line, for example:
//
//		sym "x:a+b/2; y:x*2; a:2; b:8; y+9 ="
//


main(register int argc, register char *argv[])	// parser
{

#ifdef __MWERKS__
argc = ccommand(&argv);
#endif

int ret = 0;					// exit status

if ( argc < 2 )
{
 cerr << "expression is missing\n";
 exit(1);
}

char e[1024];					// prepare the expression

e[0] = 0;
register int i = 0;
while ( ++i < argc )
 strcat(e, argv[i]);

sLex.Pointer = e;				// input buffer

//
// prepare for parsing
//

mskx_sym.Lex = &sLex;				// reset lexer
mskx_sym.Error = &myError_sym;			// reset error
myError_sym.Flag &= ~(F_NOPOP|F_NOSKIP);	// continue parsing on errors

mskx_sym.Lr1Tables->DebugReset();		// no debugging printouts
mskx_sym.Lr1Tables->ErrorProc = &myError_sym;   // for error reporting

//
// parse the expression
//

cerr << endl << "SYM: ";

if ( !mskx_sym.Success() )			// parse!!!
{
 cerr << endl << mskx_sym.Error->Number << " syntax error(s)";
 ret = 1;
}
else
{
 cerr << "Result = " << this_sym.Result;
 ret = 0;
}
cerr << endl;

return ret;
}
