/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

// 5.16

{
#include <stdio.h>
#include <stdlib.h>
}

abstract type gen
[
	attribute {char *};

	message public skip G;

	trigger { 
                     if ( !Verify() )	// verify links constantly
                        cerr << "\n"
                             << GetSymbol(LastSymbol) << " - "
			     << " broken - " << ObjectName; 
                };

	trigger G { Generate(10, atol($$), cerr); };

];

type file : public gen
{
public:
	char *FileName;
	FILE *Fd;
	long  WriteCount;
	file(char *f = "") { FileName = f; Fd = 0; WriteCount = 0; }
}
[
	state public open_ closed ioDone jobDone;
	message public Open Close
		       Read Write {char *};

	LIFE : jobDone+;
	jobDone : open_ ioDone closed;
	open_   : Open
			{Fd = fopen(FileName, "w+");};
	closed  : Close
			{(void)fclose(Fd); Fd = 0;};
	ioDone  : <
		   Read
		  |
		   Write {(void)fputs($1, Fd); ++WriteCount;}
		  >*;
];

type data(file ...) : public gen
[
	attribute {char *};

	state public on off waitOn waitOff;
	state public open_ closed jobDone;
	state private any;

	message public Open {char *} Close;

	trigger <Open Close> (->); //propagate messages to children

	LIFE : <on | off>+;

	any : on | off | waitOn | waitOff;

	waitOn   : Open;	   //become on when all children are open
	select <waitOn on> : any open_@?
				{Send(AllChildrenState(open_)?on:waitOn);};

	waitOff  : Close;          //become off when all children are off
	select <waitOff off> : any closed@?
				{Send(AllChildrenState(jobDone)?off:waitOff);};
];

type sysadm(sysadm peer ..., data server ...) : public gen
{
public:
	typeProto *Sender;
	sysadm() {Sender = 0;}
}
[
	attribute {char *};

	state   public working vacation quit fired;

	message public Transmit;

	message VacationReturn VacationGone;

	trigger Transmit {Sender = LastSender->LastSender;};

	trigger VacationReturn {UnLink(Find($$), "peer");};

	trigger <on@server off@server> {Sender = LastSender;};


	begin working;


	LIFE : working working* <quit | fired>;

	vacation : VacationGone { Link(Find($1), "peer"); };

	working : vacation VacationReturn;

	repeat <working> :
			  <
			   on@server  Transmit-
				       {
					cerr << "\nsysadm " << ObjectName
					     << ": data server "
					     << Sender->ObjectName
					     << " just went on";
				       }
			   |
			   off@server Transmit-
				       {
					cerr << "\nsysadm " << ObjectName
					     << ": data server "
					     << Sender->ObjectName
					     << " just went off";
				       }
			  >;
	repeat <vacation quit fired> : <on@server | off@server> 
					                     (, Transmit - peer)
			   {
			    cerr << "\n\n\tsysadm " << ObjectName
				 << " is not working,"
				 << "\n\ttransmitted received symbol "
				 << GetSymbol(LastSymbol) << " from server "
				 << Sender->ObjectName << " to peers\n";
			   };

];


file { FileName = "t516_1.exr"; }  f516_1,
     { FileName = "t516_2.exr"; }  f516_2,
     { FileName = "t516_3.exr"; }  f516_3,
     { FileName = "t516_4.exr"; }  f516_4,
     { FileName = "t516_5.exr"; }  f516_5,
     { FileName = "t516_6.exr"; }  f516_6,
     { FileName = "t516_7.exr"; }  f516_7,
     { FileName = "t516_8.exr"; }  f516_8,
     { FileName = "t516_9.exr"; }  f516_9,
     { FileName = "t516_10.exr"; } f516_10,
     { FileName = "t516_11.exr"; } f516_11,
     { FileName = "t516_12.exr"; } f516_12;

data(f516_1,  f516_2,  f516_3)  d516_1;
data(f516_4,  f516_5,  f516_6)  d516_2;
data(f516_7,  f516_8,  f516_9)  d516_3;
data(f516_10, f516_11, f516_12) d516_4;

sysadm(server d516_1, server d516_2) s516_1;
sysadm(server d516_3, server d516_4) s516_2;
