/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

#ifndef BMON
# define BMON
#endif

#include <iostream.h>
#include <stddef.h>
#include <fstream.h>
#include <stdlib.h>
#include <new.h>

#ifdef unix
#include <sys/types.h>
#endif

#ifdef __MWERKS__
#include <console.h>
#endif

#include <string.h>
#include <ctype.h>
#include <stdio.h>

#include "pc.h"

#ifdef V10_COMPATIBILITY__
# include "bmlog.h"
#else
# include "bml.h"
#endif


/******************************************************************_COMMENT

This file contains the main() function of the standard framework.
The following are all the items defined:

bmlg                            an ofstream where to print out state
				transition log

lr1Tables::Flag			static fields from the lr1Tables class
lr1Tables::Cerr                 are initialized here

bmLogLevel			a level of framework logging, set with the
				-l flag

bmLog				ofstream to the file where bml() macro
				will log the debugging information, and where
				state transition log will be stored by the
				State Transition Engine.

allNodes			a pointer to the list of all MUSKOX objects
				for this ARM. Each element of the list
				is a handle to the MUSKOX object. This handle
				class is bmNode defined in pc.h. This list
				contains both statically defined and dynamically
				defined objects. It is created by
				the standard framework in the
				beginning of the main() from the nodeList
				(see below). allNodes is used by the
				Copy() and Clone() members of typeProto
				to create a dynamic node, and by Delete() to
				delete a dynamic object. A standalone
				lookUpObjects() is used to search for a object
				in the list using the object's name.

typeList			a list of all MUSKOX types in the ARM generated
				by the MUSKOX compiler from one or from a number
				of MUSKOX specs. Each element of the list is of
				class initStaticTypes which is a handle to
				an array of type descriptors typeDescriptorBMon
				(see nodus.h for the definition of
				typeDescriptorBMon).

				Each array represents
				types defined in a single MUSKOX spec.
				Each element of the array points to
				an instance of the C++ class mskxBMon that
				contains pointers to the parsing tables of the
				type, lexical analyzer, error object, etc. Class
				mskxBMON is defined in the file nodus.h.
				The last element of the array is NULL.

nodeList			a list of handles of C++ class initStaticNodes
				(defined in nodus.h). Each handle points to the
				array of pointers to the MUSKOX objects
				statically defined in a particular MUSKOX spec.
				Each array is generated by the MUSKOX compiler.

local				a local queue of tokens (states & messages)
				addressed to the MUSKOX objects of this object
				module.

out				an object that is responsible for getting
				messages from outside of the system when the
				local queue is empty.

flagBM				flags that control execution of the standard
				base framework. These flags are defined in
				pc.h.

lookupTypes()			finds a particular type by its name. Searches
				for the type in the list typeList.

lookupObjects()			finds a particular object by its name. Searches
				the object in the list allNodes.

exitBM()			finishes the execution of the configuration
				monitor.

inputFile			FILE to ask for symbols if the
				local queue is empty and F_askBM is set in
				flagBM.

----------------------------------------------------------------------------

main()				The main function of the standard framework.
				The purpose of main is:

	Step 1				Get the command line flags.

	Step 2				Set the stack::Flag to PASSBYVALUE.
					This ensures that attributes are copied
					into the tokenBMon elements of the
					local queue in the methods Make() and
					FillContent(), FillContent0()
					of the tokenBMon class.

-------------------- RESOLUTION PHASE ---------------------

	Step 3				Initialize the allNodes list by the
					static MUSKOX objects pointed to
					by the nodeList. Initialization is
					done by calling the Init() member
					function of typeProto. nodeList is
					formed by the constructors of the
					initStaticNodes class. These
					constructors work for the MUSKOX
					generated instances before the OS
					invokes main().

	Step 4				Build the pointers between parent
					and children MUSKOX objects. This is
					done by calling Resolve() member
					generated by the MUSKOX compiler.

	Step 5				Fill auto signal bits for all the
					MUSKOX objects defined in MUSKOX specs.
					Auto signal bits are set if an
					automatic state change notification
					shall take place. Filling is done
					by calling FillAutoSignals() member
					generated by the MUSKOX compiler.

	Step 6				Set various static flags and fields
					in the lr1Tables class for debugging
					and logging.

-------------------- LIFECYCLE MANAGEMENT PHASE -------------------

	Step 7				Proceed to the main loop:

						1.Get a symbol from local
						  queue (if F_askBM flag is set
						  in flagBM and the queue is
						  empty, then ask for the
						  symbol from the input stream
						  by calling askForMessage()).

						2.Invoke a State Transition 
						  Engine for the object to which
						  the symbol is addressed.

------------------------------------------------------------------------------

outside::askForMessage()		a method from the outside class. asks
					for a symbol from the input stream
					defined in inputFile if inputFile is
					not NULL. Otherwise, asks it from cin.
					inputFile is defined in pc.cpp.

LexBMon()				is a member of lexBMon class. This
					is a lexical analyzer for the standard
					base framework. It retrieves a symbol
					from the local queue. It resets the
					address of the object on behalf of which
					it is called if the symbol
					is addressed to the object
					different from the current object
					passed to LexBMon() as a second
					argument. Otherwise it returns the value
					of the state or message retrieved from
					the queue. It also retrieves the
					attribute if it is present in the token.

GetMessage()				is a member of localQueue class. It
					moves a symbol in front of the local
					queue. If the object is blocked, it
					does it for the first symbol sent
					by this object itself.

localQue::Print()			prints out the local queue.

typeProto::~typeProto()			destructor of typeProto class.

COMMENT_******************************************************************/


///////////////////////////////////////////////////////////////////////////

void	exitBM(int);

static char nomem[] = "\n\nS1: MUSKOX v1.2 Standard Fremework: no more memory, exit!";

#if defined(__MWERKS__) || defined(__BCPLUSPLUS__) || defined(__sparc) ||\
    defined(__pyrsoft)
 void mskx_new_handler();
 void mskx_new_handler()
#elif defined(_MSC_VER) 
 int mskx_new_handler(size_t);
 int mskx_new_handler(size_t)
#endif
{
bml(BMLEV) << nomem << endl;
cerr << nomem << endl;
exitBM(7);

#if defined(_MSC_VER)
 return 0;
#endif
}

///////////////////////////////////////////////////////////////////////////

ofstream *bmLog = 0;
ostream *lr1Tables::Cerr = 0;			// debugging device
int bmLogLevel = 0;				// log level

int      lr1Tables::Flag = F_ACTION|F_DEBUG;	// flag (F_DEBUG, F_ACTION)



///////////////////////////////////////////////////////////////////////////

bmNode	*allNodes = 0;				// list of all objects

initStaticTypes *typeList = 0;

#ifndef __V10_COMPATIBILITY__
 initStaticNodes *nodeList = 0;
#endif

localQue local;					// local message queue

outside  out;					// object to get messages
						// sent from the outside

long	flagBM = F_askBM|F_askAttr|F_debugBM;	// ARM flags



///////////////////////////////////////////////////////////////////////////

mskxBMon *lookUpTypes(register char *n)
{
if ( !n )
 return 0;
register initStaticTypes *t = typeList;
while ( t )
{
 register typeDescriptorBMon *d = t->Types;
 register char *c;
 while ( (c = d->TypeName) != NULL )
   if ( !strcmp(n, c) )
      return d->Tables;
   else
      d++;
 t = t->Link;
}
bml(BMLEV) << "cannot find tables for type <" << n << ">\n";
return 0;
}


///////////////////////////////////////////////////////////////////////////

typeProto *lookUpObjects(register char *n)
{
if ( !n )
 return 0;
register bmNode *bmn = allNodes;
while ( bmn )
{
 register typeProto *tpn = bmn->Node;    // get to the actual object
 if ( strcmp(tpn->ObjectName, n) )         // not yet...
 {
   bmn = bmn->Link;
   continue;
 }
 return tpn;
}
bml(BMLEV) << "cannot find object <" << n << ">\n";
return 0;
}


///////////////////////////////////////////////////////////////////////////

void	exitBM(register int e)			// exit
{
bmLog->close();

cerr << "\n\nMUSKOX V1.2 : Standard Framework ARM finished" << endl;

exit(e);
}


///////////////////////////////////////////////////////////////////////////

static FILE *inputFile = 0;

///////////////////////////////////////////////////////////////////////////
void printHelp()
{
 cerr << "command line flags:\n"
         "             -l L ...    printout log messages for level N\n"
         "                                  L is one of 2, 4, 7, 8, or 10\n"
         "                                  defined in bml.h\n"
         "             -i file     file to read symbols from (stdin default)\n"
         "             -h          print this help\n";
}

///////////////////////////////////////////////////////////////////////////
main(register int argc, register char *argv[])
{

#ifdef __MWERKS__
argc = ccommand(&argv);
#endif

cerr << "\nMASTER INFORMATION SYSTEMS CORPORATION"
        "\nMUSKOX Standard Framework v1.2, Copyright (c) 1995, 1996\n"
        "\nARM <" << argv[0] << "> started up\n" << endl;

#if defined(__MWERKS__)

 (void)set_new_handler(&mskx_new_handler);    

#elif defined(__BCPLUSPLUS__) || defined(__sparc)

 (void)set_new_handler(mskx_new_handler);

#elif defined(_MSC_VER)

 (void)_set_new_handler(&mskx_new_handler);

#endif


if ( argc == 2 && !strcmp("-h", argv[1]) )	// check arguments
{
 printHelp();
 exit(1);
}

///////////////////////////////////////////////////////////////////////////

register int i = strlen(argv[0]);		// make ARM log file name
register char *logName = new char[i + 5];
strcpy(logName, argv[0]);

#ifdef unix
 strcat(logName, ".log");
#else
 register int j = (i - 4 > 0)?(i-4):0;
 while ( i-- >= j )
 {
  if ( logName[i] == '.' )
  {
    i++;
    strcpy(logName + i, "log");
    goto cont;
  }
 }
 strcat(logName, ".log");
cont:
#endif

bmLog = new ofstream(logName);
lr1Tables::Cerr = bmLog;			// ARM log file

//
// Step 1		Get command line flags
//

i = 0;
while ( ++i < argc )				// process flags
 if (!strcmp("-i", argv[i]) )			// redirect stdin (cin)
   inputFile = fopen(argv[++i], "r");
 else
   if ( !strcmp("-l", argv[i]) )		// get the log level
      while ( i < argc-1 )
	  if ( !isdigit(*argv[i+1]) )
	       break;
	  else
	       bmLogLevel |= (1 << atoi(argv[++i]));
   else
   {
      printHelp();
      exitBM(1);
   }

//
// Step 2		Set stack::Flag to PASSBYVALUE
//

stack::Flag     &= ~PASSBYREFERENCE;
stack::Flag     |= PASSBYVALUE;

///////////////////////////////////////////////////////////////////////////
//
// RESOLUTION PHASE
//
///////////////////////////////////////////////////////////////////////////

//
// Step 3		Initialize allNodes list of static MUSKOX objects
//

cerr << "\tInitialize static MUSKOX objects" << endl;

register initStaticNodes *nodes = nodeList;
register bmNode *bmn = allNodes;
while ( nodes )
{
 register typeProto **n = nodes->Nodes;		// create a list of nodes and
 while ( *n )					// initialize every node
 {
   bmn = new bmNode(*n, bmn);
   (*n++)->Init();
 }
 nodes = nodes->Link;
}
allNodes = bmn;

//
// Step 4		Resolve references between parents and children
//
 
cerr << "\tResolve references between parents and children" << endl;

nodes = nodeList;
while ( nodes )
{
 register typeProto **n = nodes->Nodes;		// resolve all references
 while ( *n )
   (*n++)->Resolve();
 nodes = nodes->Link;
}

//
// Step 5		Fill in auto signal bits for state change notification
//

cerr << "\tSet up notification by marked symbols" << endl;

nodes = nodeList;
while ( nodes )
{
 register typeProto **n = nodes->Nodes;		// fill auto signal bits
 while ( *n )
   (*n++)->FillAutoSignals();
 nodes = nodes->Link;
}

//
// Step 6		Set flags and static fields
//

lr1Tables::Flag = F_ACTION|F_DEBUG;	// flag (F_DEBUG, F_ACTION)
lr1Tables::Cerr = bmLog;		// debugging device

if ( flagBM%F_askBM )
 lr1Tables::Flag |= F_INTERACTIVE;


///////////////////////////////////////////////////////////////////////////
//
// END OF RESOLUTION PHASE
//
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
//
// LIFECYCLE MANAGEMENT PHASE
//
///////////////////////////////////////////////////////////////////////////

//
// Step 7		main loop - retrieve a symbol and dispatch it
//

cerr << "\tStart controlling object lifecycles\n\n" << endl;

register tokenBMon *t = 0;			// first symbol
register typeProto *n;				// target object
if ( flagBM&F_askBM )
{
 for ( ;; )
   if ( (t = out.AskForMessage()) != NULL )	// ask the first symbol
      break;
 n = t->Destination.Address;			// target object
}
else
 n = 0;

register localQue *lq = &local;
register typeProto *m = n;			// now get a LIFE...
for (;;)
{
 if ( n )					// the object is there
 {
   *(lr1Tables::Cerr) << endl << "==\t\t== object == "
                      << n->ObjectName;
   if ( n->Parse(&m) < 0 )			// Parse! end of job?...
      break;
 }
 else						// undefined object or startup
 {
   lq->Free();					// free symbol for undef. object
   for ( ;; )					// find a symbol for
   {						// existing object
      register tokenBMon *t;

      if ( (t = lq->GetMessage()) == NULL )	// no symbols
	  if ( flagBM&F_askBM )			// debug mode?
	       for ( ;; )
	             if ( (t = out.AskForMessage()) != NULL )// ask for symbol
                            break;
      if ( (m = t->Destination.Address) == NULL )// destination - unknown...
      {
	  lq->Free();				// free symbol for undef. object
	  continue;				// keep looking...
      }
      break;					// now got the real object...
   }
 }
 n = m;						// new object will parse...
}

exitBM(0);

return 0;	// to avoid compiler warning
}


///////////////////////////////////////////////////////////////////////////

static int readString(char *, int, char *);	// read into string and get rid off the 
					// trailining newline
static int readString(register char *string, register int length, 
 		    register char *prompt)
{
*string = '\0';
cerr << prompt;
if ( !fgets(string, length, stdin) )
 exitBM(3);
register int l = strlen(string) - 1;
if ( l >= 0 && string[l] == '\n')
 string[l] = 0;
return l;
}

///////////////////////////////////////////////////////////////////////////

tokenBMon *outside::AskForMessage() 	// ask for symbol from cin
{
static sAtribus attrib[2];		// to ask the value of attribute

char node[512], symbol[512];		// buffers for input symbol

if ( inputFile )			// get symbols from the input file
{
 char buf[512];                         // read next line here
 register char *b, *b0;

 for ( ;; )				// loop reading the file
 {
   b = buf;
   do
   {
      Line++;
      *b = 0;
      if ( !fgets(b, 512, inputFile) && !strlen(b) )  	// eof or error
	  exitBM(0);
      while ( *b && iscntrl(*b) ) 	      	        // skip cntrls
	  b++;
   }
   while ( (b[0] == '/' && b[1] == '/') );		// skip comment

   b0 = b;

   register int l = strlen(buf) - 1;
   if ( l >= 0 && buf[l] == '\n')
      buf[strlen(buf) - 1] = 0;		// get rid of newline

   while ( isspace(*b) )		// skip initial spaces
      b++;
   if ( !*b )				// skip blank line
      continue;
   break;                               // something on the line
 }

 while ( *b && !isspace(*b) )		// isolate object
   b++;
 if ( !*b )
   goto exit4;

 *b++ = 0;
 strcpy(node, b0);
 while ( isspace(*b) )			// spaces after object
   b++;
 if ( !*b )
 {
exit4:

   cerr << "\nS4: no symbol for object <" << buf 
	<< "> in .inp file, line = " << Line;
   exitBM(4);
 }

 b0 = b;

 while ( *b && !isspace(*b) )		// isolate symbol
   b++;
 *b++ = 0;
 strcpy(symbol, b0);

 while ( *b && isspace(*b) )		// spaces after symbol
   b++;

 attrib[0].Word = new char[128];
 strncpy(attrib[0].Word, b, 127);
}
else
{
  Line++;
  cerr << "\n\nPlease enter separately object, symbol, and optional attribute (eof to exit):\n\n";
  register int l = 0;
  
  do
     l = readString(node, 512, "object> ");
  while ( !l );
  
  do
     l = readString(symbol, 512, "symbol> ");
  while ( !l );
  
  if ( flagBM&F_askAttr )
  {
    attrib[0].Word = new char[128];
    readString(attrib[0].Word, 128, "\noptional attribute (character string)> ");
  }
}

register typeProto *n = lookUpObjects(node);	// find the object
if ( !n )
{
 cerr << "\nS2: object <" << node << "> is absent -- repeat . . .\n";
 return 0;
}
register tokenBMon *m = local.Append();		// get the symbol
if ( !flagBM&F_askAttr )
 m->Make(n, n, symbol);				// send the symbol
else
 m->Make(n, n, symbol, &(attrib[0].Word));	// send with attribute
return m;
}

tokenBMon *outside::GetMessage() { return 0; }


///////////////////////////////////////////////////////////////////////////

ubrick lexBMon::LexBMon(register void *a, register typeProto **n,
		       register typeProto *tn)
{

//
// Get the symbol from the local queue or ask for it from outside
//

register tokenBMon *t = local.GetMessage();	// next symbol
while ( !t )					// que is empty
 if ( flagBM&F_askBM )
   t = out.AskForMessage();			// ask for symbol from outside
 else
 {
   *n = 0;
   return 0;					// or exit...
 }

#ifdef INPUS
Input->Line = out.Line;
#endif

//
// This symbol is for another object
//

if ( *n != t->Destination.Address )     // not to this object - exit...
{
 *n = t->Destination.Address;		// store new destination address
 return 0;
}

//
// This symbol is for current object
//

register ubrick r = t->Content;         // symbol
if ( r == Other )			// undefined?
 r = (*n)->UserLexBMon(t, a);		// call user to resolve it...
register void *c = t->ContentScript;	// attribute
if ( c )				// copy attribute
 moveAttribute((*n)->Tables_us->Lr1Tables->Tokens + r, a, c);
else
 a = 0;

tn->LastSymbol = r;			// remember last symbol
tn->LastAttribute = a;			//	    last attribute
tn->LastSender = t->ReturnAddress.Address;//	    last object
local.Free();                           // free the symbol
return r;                               // done
}


///////////////////////////////////////////////////////////////////////////

//
// Find first symbol to unblocked object, or first symbol to a blocked
// object that comes from the object itself,
// then move that symbol in front of the local queue
//
tokenBMon *localQue::GetMessage()
{
register tokenBMon **a = &(local.MesBusy);
register tokenBMon *t = local.MesBusy, *ta = local.MesBusy;
while ( t )
{
 typeProto *d = t->Destination.Address;
 if ( !d->IsFrozen() && (!d->IsBlocked() || d == t->ReturnAddress.Address) )
 //if ( !t->Destination.Address->IsBlocked() ||
 //     t->Destination.Address == t->ReturnAddress.Address )
 {
   if ( t != local.MesBusy )
   {
      *a = t->Link;			// move it in front
      t->Link = local.MesBusy;
      local.MesBusy = t;
      if ( local.MesLast == t )		// update the last symbol
	  local.MesLast = ta;
   }
   return t;				// return the found symbol
 }
 ta = t;
 a = &(t->Link);			// get next symbol
 t = t->Link;
}
return out.GetMessage();		// No tokens in the local queue,
}				        // return a token sent from outside

void localQue::Print(register ostream &o)	// only for debugging
{
register tokenBMon *t = MesBusy;
register int i = 0;
o << "\n\n--- Local que content:";
while ( t )
 o << "\n  " << t, t->Print(o), t = t->Link, i++;
o << "\n  Total elements in local que <" << i << '>' << endl;
}


typeProto::~typeProto() 		         // destructor
{
if ( AutoSignalBits )
 delete [] AutoSignalBits;
if ( Delay )
 delete Delay;
if ( Stack )
 delete Stack;
}


