/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////


#ifndef __atribus_h_
#define __atribus_h_

/******************************************************************_COMMENT

This file defines various constants used in MUSKOX, and a C++ class or 
typedef for MUSKOX attributes. This file is automatically included in
all the .h files produced by the MUSKOX compiler.

The constants and symbols defined are:


STACK				the depth of the parsing stack for
				the state transition engine.

DEBUGON				if defined, allows logging of state transitions
				in the state transition engine.

TIMING				if defined, counts time which MUSKOX compiler 
				uses to parse the source files. Used only
				for MUSKOX compiler when bootstrapping,
				not for the framework, neither for a standalone
				parser.

NOACTION			if defined, prohibits execution of any
				action specified in MUSKOX grammar. Used only
				for MUSKOX compiler when bootstrapping,
				not for the framework, neither for a standalone
				parser.

UNIONACT			currently undefined so that the _cpy_ in
				skeletus.h is defined as struct. You should not
				define this constant since not all the C++
				compilers work correctly when _cpy_ is
				defined as union. _cpy_ is used by the
				State Transition Engine in order to copy
				attrubutes using either memcpy, or
				the attribute's copy constructor.

MSKXBIG_			if defined, State transition engine will use
				long numbers for the LR1 tables. If not defined,
				short numbers will be used instead. Usually
				MSKXBIG_ should not be defined,


COPYCONSTRUCTOR			if not defined, State Transition Engine
				will always use memcpy() to copy the attributes
				on the stack. If defined, then the copy
				constructor will be used if defined, or
				memcpy() otherwise.

FAST_PARSER			used only for the stand-alone version
				of the State Transition Engine: for
				Parse() and ParseInRealTime() of the lr1Tables
				class. If defined, ignores processing of
				triggers, reductors, and skipped symbols.

INPUS				used only for the stand-alone version
				of the State Transition Engine: for
				Parse() and ParseInRealTime() of the lr1Tables
				class. If defined, allows definitions of the
				input class that is used to scan literal
				names of MUSKOX terminals and nonterminals
				from the input stream. You should define
				INPUS if you plan to use the input class for
				the stand alone State Transition Engine.

sAtribus			defined as a class if BMON constant is defined,
				or defined as a typedef long otherwise. BMON is
				automatically defined in the .h files produced
				by MUSKOX compiler with the '-b' or '-n' flags.
				This is a normal case for the MUSKOX specs
				compiled for inclusion in the ARM.

GENON				allows definitions of the generator data
				structures. 

COMMENT_******************************************************************/

#ifdef __PYRAMID__
# define __V10_COMPATIBILITY__ __PYRAMID__
#endif

#undef STACK
//#define STACK

//#undef DEBUGON
#define DEBUGON

#undef TIMING
//#define TIMING

#undef NOACTION
//#define NOACTION

#undef UNIONACT
//#define UNIONACT

#undef MSKXBIG_
//#define MSKXBIG_

#undef COPYCONSTRUCTOR
//#define COPYCONSTRUCTOR

//#undef FAST_PARSER
#define FAST_PARSER

#ifndef INPUS
 #define INPUS
#endif

//#undef GENON
#define GENON

#if defined(BMON)

class typeProto;

/////////////////////////////////////////////////////////////////////
class	sAtribus		// default attribute for the ARM
{
public:
	typeProto *Node;	// sender node
	union			// message
	{
	 long  Number;
	 char *Word;
	};
	sAtribus(register typeProto *t, register char *c)
			{ Node = t, Word = c; }
	sAtribus(register typeProto *t, register long l)
			{ Node = t, Number = l;}
	sAtribus() 	{ Node = (typeProto *)0, Number = 0; }
	sAtribus &operator = (register long &n)
			{ Node = (typeProto *)0, Number = n; return *this; }
	sAtribus(register long &n)
			{ Node = (typeProto *)0, Number = n; }
};
/////////////////////////////////////////////////////////////////////

#else				// #ifndef BMON

#ifdef __V10_COMPATIBILITY__
union   sAtribus
{
        long    Number;
        char   *Word;
        sAtribus() { Number = 0; }
        sAtribus &operator = (register long &n) { Number = n; return *this; }
        sAtribus(long x) { Number = x; }
        sAtribus(int x) { Number = x; }
        sAtribus(char *x) { Word = x; }
};

#else

/////////////////////////////////////////////////////////////////////
# ifndef __V10_COMPATIBILITY__
    typedef long sAtribus_;	// default synthesized attribute for 
				// standalone State Transition Engine
#   define sAtribus sAtribus_
# endif
#endif				// #if defined(BMON)
/////////////////////////////////////////////////////////////////////

#endif

#endif
