/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2                                      //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

#ifndef __generus_h_
#define __generus_h_

#ifndef __atribus_h_
#include "atribus.h"
#endif

#ifndef __inpus_h_
#include "inpus.h"
#endif

#ifndef __skeletus_h_
#include "skeletus.h"
#endif


enum    visitType {
		    PROTOVISIT_,
		    SIMPLEVISIT_,
		    FULLVISIT_,
		    FULLVISITSINGLE0_,
		    FINALVISIT_,
		    EXPANDVISIT_,
		    GENVISIT_
};


extern long Level;

class	bounds
{
public:
	long	Length;			// max length of generated strings
	bounds(register long l)
	{ Length = l; }
};


class	delta				// per rule
{
public:
	ubrick	DerivedLength;		// min length of derived string
	ubrick	NonTerminals;		// # of nonterminals in right part
};


class	protoVisit;
class	initVisits;
class   expandVisit;
class   simpleVisit;


enum flags {                            // flags for visits
	    F_hide    = 1,		// do not print the visit out
	    F_reached = 2,              // the visit is on the path to target
	    F_printed = 4, 		// the visit has been printed out
	    F_single0 = 8		// visit for single 'alpha : ;'
	   };

extern protoVisit  *freeStore;
extern protoVisit  *atEnd;
extern protoVisit  *currLoc;
#define CHUNK	16384

class	protoVisit
{
public:
	ubrick       State;     // state number
	ubrick	     Flag;	// flags
	protoVisit  *PrevVisit; // previous visitState on the path
	protoVisit  *NextVisit; // next visitState
	protoVisit  *SameVisit; // link to the visit for the same state
	brick	     Passed;	// current length of the path
	long	     Level;	// level of expansion
static	generator   *Generator;	// generator data
static	lr1Tables   *Lr1Tables; // tables...
	protoVisit  *next;	// pointer to the next allocated space
	protoVisit()
	     { State = 0, Flag = 0, PrevVisit = 0, Level = ::Level,
	       Passed = 0, NextVisit = 0, SameVisit = 0; }
	protoVisit(register ubrick state, register protoVisit *prev)
	     { State = state, Flag = 0, PrevVisit = prev, Level = ::Level,
	       Passed = 0, SameVisit = 0, NextVisit = 0,
	       prev->NextVisit = this; }
	protoVisit(register ubrick state)
	     { State = state, Flag = 0, Level = ::Level, Passed = 0,
	       PrevVisit = 0, NextVisit = 0, SameVisit = 0; }
virtual	~protoVisit()
	     { if ( PrevVisit ) PrevVisit->NextVisit = 0; }
virtual	simpleVisit *GetVisit() { return 0; }
virtual protoVisit *Advance(initVisits *) { return NextVisit; }
virtual protoVisit *Fork(initVisits *) { return 0; }
virtual	protoVisit *ExpandAdvance(register initVisits *);
virtual	protoVisit *ExpandFork(register initVisits *);
virtual char *GetToken() { return Lr1Tables->DebugHash[State]->Word; }
virtual	int IsNonTerminal()
{
	register ubrick t = Lr1Tables->DebugHash[State]->Result;
	return (t>=Generator->MaxTerminal)?t:BRICKFFFF;
}
virtual int IsItNonTerminal()
{ return Lr1Tables->DebugHash[State]->Result >= Generator->MaxTerminal; }
virtual ubrick Delta()
{
	register ubrick x = Lr1Tables->DebugHash[State]->DerivedLength;
	return (Generator->Flag&F_messages)?((x==BRICKEFFF)?1:x):(x != 0);
}
virtual void Restore(expandVisit *, initVisits *);
virtual visitType   Type() { return PROTOVISIT_; }
void *operator new(size_t size);
void  operator delete(void* p);
};


extern protoVisit stop;


class	simpleVisit : public protoVisit
{
public:
	ubrick  *Edges;         // edges to advance
	ubrick   Counter;       // number of edges
	ubrick	 Current;	// current edge for advancement
	ubrick	 Forks;		// number of forks undertaken
	simpleVisit(register ubrick state, register protoVisit *p)
		    : protoVisit(state, p)
		    {
		      register generator *gen = Generator;
		      Edges = gen->Edges + gen->GenShiftHash[state] + 1;
		      Counter = gen->Edges[gen->GenShiftHash[state]];
		      Current = Forks = 0;
		    }
	simpleVisit(register ubrick state) : protoVisit(state)
		    {
		      register generator *gen = Generator;
		      Edges = gen->Edges + gen->GenShiftHash[state] + 1;
		      Counter = gen->Edges[gen->GenShiftHash[state]];
		      Current = Forks = 0;
		    }
virtual	simpleVisit *GetVisit() { return new simpleVisit(*this); }
	protoVisit *Advance_(initVisits *);
	protoVisit *Fork_(initVisits *);
virtual	protoVisit *Advance(register initVisits *iv)
				{ return Advance_(iv); }
virtual	protoVisit *Fork(register initVisits *iv)
				{ return Fork_(iv); }
virtual visitType Type() { return SIMPLEVISIT_; }
};

class	expandVisit : public protoVisit
{
public:
	protoVisit *Expand;	// expand Expand->IsNonTerminal()
	ubrick   Token;		// token to expand
	brick    Remains;	// Remains from initVisits saved here
	brick    NonTerminals;	// unexpanded nonterminals on the path
	ubrick  *Rules;         // rules to expand
	delta   *Deltas;	// derived lengths and # of nonterminals
	ubrick   Counter;       // number of rules
	ubrick	 Current;	// current rule for expansion
	ubrick	 Forks;		// number of forks undertaken
	void	 Init(register ubrick token,
		      register protoVisit *expand, register brick remains,
		      register brick nonTerminals, register delta *d)
		    {
		      register generator *gen = Generator;
		      Token = token;
		      token -= gen->MaxTerminal;
		      token <<= 1;
		      Rules = gen->Lengths + gen->Rules[token];
		      Deltas = d + (gen->Rules[token]>>1);
		      Counter = gen->Rules[token + 1];
		      Current = Forks = 0;
		      Generator = gen;
		      Flag |= F_hide;
		      Expand = expand;
		      Remains = remains;
		      NonTerminals = nonTerminals;
		    }
	expandVisit(register protoVisit *expand, register ubrick token,
		    register protoVisit *p, register brick remains,
		    register brick nonTerminals, register delta *d)
		    : protoVisit(BRICKFFFF, p)
		    { Init(token, expand, remains, nonTerminals, d); }
	expandVisit(register protoVisit *expand, register ubrick token,
		    register brick remains, register brick nonTerminals,
		    register delta *d) : protoVisit(BRICKFFFF)
		    { Init(token, expand, remains, nonTerminals, d); }
virtual ubrick Delta()
{
	register ubrick x = Lr1Tables->Find(Token)->DerivedLength;
	return (Generator->Flag&F_messages)?((x==BRICKEFFF)?1:x):(x != 0);
}
virtual	protoVisit *ExpandFork(register initVisits *iv);
virtual char *GetToken()  { return Lr1Tables->Find(Token)->Word; }
virtual visitType Type() { return EXPANDVISIT_; }
};

class	genVisit : public protoVisit
{
public:
	ubrick   Token;		// token to expand
	genVisit(register ubrick token, register protoVisit *p)
		    : protoVisit(BRICKFFFF, p) { Token = token; }
	genVisit(register ubrick token)
		    : protoVisit(BRICKFFFF) { Token = token; }
virtual	int IsNonTerminal()
	    {
	     register ubrick t = Token;
	     return (t >= Generator->MaxTerminal)?t:BRICKFFFF;
	    }
virtual int IsItNonTerminal() { return Token >= Generator->MaxTerminal; }
virtual ubrick Delta()
{
	register ubrick x = Lr1Tables->Find(Token)->DerivedLength;
	return (Generator->Flag&F_messages)?((x==BRICKEFFF)?1:x):(x != 0);
}
virtual char *GetToken() { return Lr1Tables->Find(Token)->Word; }
virtual visitType Type() { return GENVISIT_; }
};

class	reductions
{
public:
	ubrick  *Reductions2;           // 2 or more reductions to reduce
	ubrick   Reduction;             // single reduction
	ubrick   Counter;               // number of reductions
	ubrick   Current;		// current reduction to reduce
	ubrick   Forks;			// number of forks undertaken
	reductions(ubrick state, generator *);
};


class	fullVisit : public simpleVisit, public reductions
{
public:
	fullVisit(register ubrick state, register protoVisit *prev)
		  : simpleVisit(state, prev), reductions(state, Generator) {}
	fullVisit(register ubrick state)
		  : simpleVisit(state), reductions(state, Generator){}
virtual	simpleVisit *GetVisit()
		    { return new simpleVisit(*((simpleVisit *)this)); }
virtual visitType Type() { return FULLVISIT_; }
};


class	fullVisitSingle0 : public fullVisit
{
public:
	protoVisit **Steps;			// Steps from initvisits
	fullVisitSingle0(register ubrick state, register protoVisit *prev,
		  register protoVisit **s)
		  : fullVisit(state, prev) { Steps = s; }
	fullVisitSingle0(register ubrick state, register protoVisit **s)
		  : fullVisit(state) { Steps = s; }
virtual	simpleVisit *GetVisit() { return Steps[Edges[0]]->GetVisit(); }
virtual	protoVisit *ExpandAdvance(register initVisits *){ return GetVisit(); }
virtual visitType Type() { return FULLVISITSINGLE0_; }
};

class	finalVisit : public protoVisit, public reductions
{
public:
	finalVisit(register ubrick state, register protoVisit *prev)
		   : protoVisit(state, prev), reductions(state, Generator) {}
	finalVisit(register ubrick state)
		   : protoVisit(state), reductions(state, Generator) {}
virtual	simpleVisit *GetVisit() { return 0; }
virtual	protoVisit *ExpandAdvance(register initVisits *) { return 0; }
virtual visitType Type() { return FINALVISIT_; }
};


class	pass
{
public:
	unsigned long	Counter;			// counter of passes
	void    Clear() { Counter = 0; }
	pass() { Clear(); }
};


class	visit
{
public:
	pass 	Edges;				// passes for shifts
	void	Clear()
		{ Edges.Clear(); }
	visit() { Clear(); }
};


class initVisits
{
public:
	visit  *Visits;       // array of visits
	delta  *Delta;	      // array of derived lengths and # of nontermnls
	protoVisit **Steps;   // array of sample states to feed GetVisit()
	ubrick  Target;	      // target state
	protoVisit *TargetState;//sample target state
	protoVisit *Start;    // start state
	ostream   &Io;	      // output file
	brick	**Distance;   // array of distances between nodes
	ubrick  Number;	      // number of states in automaton
	ubrick  PathCounter;  // counter of paths
	brick   Remains;      // number of steps till the end (total-Length)
	brick   NonTerminals; // unexpanded nonterminals on the path
	generator *Generator;
	lr1Tables *Tables;
	lex       *Lex;
	initVisits(bounds *, ubrick to, generator *, lex *, lr1Tables *, 
		   ostream &f);
	~initVisits();
	protoVisit *CreateVisit(register ubrick state);  // create a visit
							 // from scratch
	protoVisit *GetVisit(register ubrick state)      // get a visit from
							 // Steps
		    {
		     register protoVisit *v = Steps[state];
		     register protoVisit *n = v->GetVisit();
		     if ( n )
		     {
		       n->SameVisit = v->SameVisit;
		       v->SameVisit = n;
		     }
		     return n;
		    }
	brick       SoFarPassed(register protoVisit *p)	// path length so far
		    {
		     return p->SameVisit?p->SameVisit->Passed:0;
		    }
	void	    Delete(register protoVisit *v)      // delete a visit
		    {
		     Steps[v->State]->SameVisit = v->SameVisit;
		     delete v;
		    }
	void	    Flush();  			// flush generated path
	int	    Expand(ubrick target);	// expand rules on path
	protoVisit *Expand_(register protoVisit *, register ubrick,
			    register expandVisit *);
	protoVisit *Shift(simpleVisit *, ubrick);
	int	    CheckDistance(simpleVisit *, ubrick);
	void	    DeltaCalc(ubrick, generator *, lr1Tables *);
	void	    PrintStack(register protoVisit *);
	void	    PrintWord(register protoVisit *);
};

void freeAll();

#endif
