/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

#ifndef __nodus_h_
#define __nodus_h_

#include <stdio.h>
#include <time.h>

#ifndef __atribus_h_
#include "atribus.h"
#endif

#ifndef __inpus_h_
#include "inpus.h"
#endif

#ifndef __skeletus_h_
#include "skeletus.h"
#endif

#ifndef __skelbmon_h_
#include "skelbmon.h"
#endif

#ifdef __V10_COMPATIBILITY__
# include "bmlog.h"
#else
# include "bml.h"
#endif

/******************************************************************_COMMENT

This file contains definitions that define components of the standard framework:

typeDescriptorBMon			this class defines an element of the
					array of types that is generated by
					the MUSKOX compiler for each MUSKOX
					spec. typeDescriptorBMon defines the 
					type's name and a pointer to mskxBMon
					class that in turn contains pointers
					to the parsing tables, lexical analyzer,
					error object, etc.

initStaticTypes				a handle to the array of MUSKOX types
					generated by MUSKOX compiler for each
					MUSKOX spec.  Each handle points to the 
					beginning of the array and to the
					next handle in the list. Elements of 
					these arrays are of the class 
					typeDescriptorBMon (see above).

					Each array of types is pointed
					to by the handle of the class 
					initStaticTypes. All the handles are
					linked in a list at the startup time
					of the configuration monitor by the
					contructor of the initStaticTypes
					class. The head of the list is
					a pointer typeList defined in pc.cpp.
					This list is used to find the MUSKOX
					type using lookUpTypes() function
					defined in pc.cpp.

initStaticNodes				a handle to the array of static MUSKOX 
					nodes generated by the MUSKOX compiler
					for each MUSKOX spec. Each handle points
					to the beginning of the array and to the
					next handle in the list. Elements of 
					these arrays are pointers to MUSKOX
					nodes (i.e pointers to typeProto).

					All the handles are linked in a list
					at the startup time of the configuration
					monitor by the constructor of the
					initStaticNodes class. The head of the
					list is nodeList defined in pc.cpp. This
					list is used by main() to create a list
					of muskox nodes (that starts with
					allNodes) that then is used by the
					Clone() member of typeProto to create
					a dynamic node, and by Delete() to
					delete a dynamic node.
					
autoSignal				this struct defines which symbol 
					(Signal) should be sent when an
					object shifts a symbol defined in
					AtState. Used for the automatic
					state change notification with
					marked symbols.

nodePointer				this class defines an element of the 
					list of parents or the list of children.
					Each node has a list of its children
					and a list of its parents at run-time.
					For static MUSKOX nodes, these
					lists are formed at the start-up
					time of the configuration monitor by
					the Resolve() member functions of 
					each MUSKOX type called from the main().
					For dynamic nodes children and parent
					lists are maintained by the typeProto
					methods Link() and Unlink().

					The Trans member points to a 
					translation table that is used by
					the Translate() method of typeProto
					called from the communication methods
					Send() to translate binary codes for
					the same symbols between different 
					MUSKOX types. The AutoSignal member
					points to the array of autoSignal
					structs that define symbols that
					are sent by the State Transition Engine
					upon shifting symbols on the stack.

					nodePointer class represents only 
					variable links at run-time. Fixed links
					are represented by nodePointerFixed
					which is derived from nodePointer
					(see below).

nodePointerFixed			this class is derived from nodePointer
					and represents a fixed link in the
					list of children. The ChildFixed
					member points within the type to the 
					pointer that points to the child.
					The childArgName points to the
					character string that denotes that
					child's name.

childsAndParents			this class is a base class for 
					typeProto. It points to the list of
					node's children (Children), to the list
					of node's parents (Parents), and to the
					list of translation tables (Trans)
					that translate binary codes for symbols
					between the node and it's children.
					The list of translation tables is shared
					between all the nodes of the particular
					MUSKOX type.
					
os					this class is merely a placeholder
					for the functionality of the operating
					system. Currently it is empty.

Nodeblocked				this enumeration defines a flag in
					the typeProto Flags member that
					indicates that the node is currently 
					blocked.

DynamicNode				enumeration defines if the node is a
					dynamic one.

typeProto				this is the main base class of all
					MUSKOX nodes. Please refere to the 
					definition of this class for the
					explanation of its members.

SendState				These macros send the actual current
SendBackState				state from a MUSKOX action, trigger,
SendChildrenState			or reductor. The state is taken from
SendParentsState			the top of the parsing stack, not
					from the member CurrentState member
					of typeProto.

TransTable				defines an element of a list of the
					translation tables. The list is created
					for a parent and a child pair of types
					and is shared among all the pairs
					of parent/child objects. Each
					table defines symbol codes for some
					target type based on the symbols from 
					the current type: table[i] has the
					binary code in the target type of the 
					symbol which binary code is i in the 
					current type. The translation table
					is used by the communication
					methods of typePRoto.

linkIterator				a class that allows to iterate through
					the object links. Used in the
					link() and unlink() methods from
					typeProto. This is an unpublished
					interface.

childrenIterator			a class that iterates through the
					children of a parent object. This
					is a part of the published interface.

COMMENT_******************************************************************/


class	mskxBMon;
class	errorBMon;

////////////////////////////////////////////////////////////////////////

class	typeDescriptorBMon		// element of the array of types
{
public:
	char	 *TypeName;		// name of type
	mskxBMon *Tables;		// address of tables
};

class initStaticTypes;
extern initStaticTypes *typeList;

////////////////////////////////////////////////////////////////////////

class   initStaticTypes                 // pointer to static array of types
{
public:
	initStaticTypes *Link;
	typeDescriptorBMon *Types;
	initStaticTypes(typeDescriptorBMon *);
};

class initStaticNodes;
extern initStaticNodes *nodeList;

////////////////////////////////////////////////////////////////////////

class	initStaticNodes			// pointer to static array of nodes
{
public:
	initStaticNodes *Link;
	typeProto      **Nodes;
	initStaticNodes(typeProto **, void (*)());
};

class	typeProto;
class	transTable;
class	nodePointerFixed;

////////////////////////////////////////////////////////////////////////

struct	autoSignal
{
	brick	AtState;
	brick   Signal;
};

////////////////////////////////////////////////////////////////////////

class	nodePointer			// element of a list of parents/children
{
protected:
	typeProto   *ParentChild;	// reference to the parent/child
	nodePointer *Link;		// next in the list
	nodePointerFixed *Down;		// pointer to the list of children
	long	     ManyChildren;	// many children for this link
	transTable  *Trans;		// translation table
	autoSignal  *AutoSignal;	// auto signal translation table

	mskxBMon    *DerivedFrom(mskxBMon *); // check the type

public:
	nodePointer(register typeProto *relative, register nodePointer *np)
		   { ParentChild = relative, Link = np, Trans = 0,
		     AutoSignal = 0; Down = 0; ManyChildren = 0; }
virtual ~nodePointer();
virtual nodePointer *ClearFixed();              // dummy for a fixed child
virtual char        *ChildName();             
virtual typeProto  **IsFixed();                 // not a fixed child
	typeProto   *GetReference() { return ParentChild; }
	nodePointer *Next() { return Link; }
friend class childsAndParents;
friend class typeProto;
friend class linkIterator;
friend class childrenIterator;
};

////////////////////////////////////////////////////////////////////////

class	nodePointerFixed : public nodePointer	// fixed child
{
protected:
	typeProto   **ChildFixed;	// pointer to the child pointer
	char	     *ChildArgName;	// name of the argument for this child

public:
	nodePointerFixed(register typeProto *r, register nodePointer *np,
			 register typeProto *&cf, register char *ca,
			 register int v) : nodePointer(r, np)
		   {
		    cf = r, ChildFixed = &cf, ChildArgName = ca;
		    if ( v )
		      ManyChildren = 2;		// always # of children + 1
		   }
virtual char         *ChildName();
virtual nodePointer  *ClearFixed();
virtual typeProto   **IsFixed();
friend class typeProto;
};

////////////////////////////////////////////////////////////////////////

class	childsAndParents		// children and parents of the node
{
protected:
	nodePointer *Children;		// all children
	nodePointer *Parents;		// all parents
	transTable  *Trans;		// translation tables

public:
	void  InitTransTable(typeProto *, typeProto *, // init. trans. tables
			     nodePointer *);
	childsAndParents() { Children = 0, Parents = 0, Trans = 0; }

#ifdef __V10_COMPATIBILITY__
        nodePointer *ChildsPtr() { return Children; }
	nodePointer *ParentsPtr() { return Parents; }
#endif
friend class childrenIterator;
};

class rule;
class tokenBMon;

////////////////////////////////////////////////////////////////////////

class os				// init os-related stuff
{
#ifdef __V10_COMPATIBILITY__
protected:
static int Run;
public:
	os();
#else
public:
	os(){};
#endif
};

////////////////////////////////////////////////////////////////////////

enum { NodeBlocked = 1, NodeFrozen = 2, DynamicNode = 4 };

////////////////////////////////////////////////////////////////////////


class localQue;
class typeProto;

typeProto *lookUpObjects(char *name);

#ifdef __V10_COMPATIBILITY__
  class basic;
# define NodeName ObjectName
# define lookUpNodes lookUpObjects
#endif

////////////////////////////////////////////////////////////////////////

class typeProto : public childsAndParents,// all objects are derived
					  // from this type

#if defined(__sparc) && defined(__SVR4)   // only for sparc C++ 4.0.1
		  virtual public actions,
		  virtual public os
#else
		  public actions,
		  public os
#endif

{
public:

////////////////////////////////////////////////////////////////////////
//
///// published interface
//

////////////////////////////////////////////////////////////////////////
//
// PUBLISHED MEMBERS AND METHODS
//
// Lifecycle Members
//
//	char      *ObjectName
//      typeProto *LastSender
//      long       LastSymbol
//      void      *LastAttribute
//      long       CurrentState
//      long       TopStack(long l)
//      char      *TopStackSymbol(long l)
//      char      *GetSymbol(long symbol)
//      void       Block()
//      void       UnBlock()
//      long       IsBlocked()
//      void       Freeze()
//      void       UnFreeze()
//      void       IsFrozen()
//      void       RePost()
//      long       AllChildrenState(long state, char *name = 0)
//	long       AnyChildState(long state, char *name = 0)
//
// Framework Methods
//
//	virtual long       Link(typeProto *child, char *name = 0)
//	virtual long       UnLink(typeProto *child, char *name = 0)
//	        long       Verify()
//              long       IsDynamic()
//              typeProto *Clone(char *newName)
//	virtual void       SetSelf()
//      static  typeProto *Find(char *name)
//	        void       Delete()
//
///////////////////////////////////////////////////////////////////////


//
// published data members
//
	char      *ObjectName;		 // this node name
	typeProto *LastSender;		 // last sender to this node
	long       LastSymbol;		 // last symbol recieved by this node
	void      *LastAttribute;	 // attribute of the last symbol ...
	long       CurrentState;	 // last non-hidden state
////////////////////////////////////////////////////////////////////////

//
// published member-functions
//
////////////////////////////////////////////////////////////////////////
#ifdef __V10_COMPATIBILITY__
	long	   IsBlocked() { return NodeFlags&NodeBlocked;}//is node blocked?
	void	   BlockNode() { NodeFlags |= NodeBlocked; }   // block node
	void	   UnBlockNode() { NodeFlags &= ~NodeBlocked; }// unblock node
	long       IsFrozen() { return NodeFlags&NodeFrozen; } // is node frozen
	void       Freeze() { NodeFlags |= NodeFrozen; }       // freeze node
	void       UnFreeze() { NodeFlags &= ~NodeFrozen; }    // unfreeze node
#else
	long	   IsBlocked() { return Flags&NodeBlocked; } // is node blocked?
	void	   Block() { Flags |= NodeBlocked; }	     // block node
	void	   UnBlock() { Flags &= ~NodeBlocked; }	     // unblock node
	long       IsFrozen() { return Flags&NodeFrozen; }   // is node frozen?
	void       Freeze() { Flags |= NodeFrozen; }         // freeze node
	void       UnFreeze() { Flags &= ~NodeFrozen; }      // unfreeze node
#endif


virtual	long       Link(typeProto *child, char *name = 0);   // link a child
virtual	long       UnLink(typeProto *child, char *name = 0); // unlink a child

	typeProto *Clone(char *newName);	             // clone a node

virtual	void       SetSelf();            // set the static type::Self

	void       Delete();	         // delete this node from library

static typeProto  *Find(char *name)	 // finds an object with a given name
		   { return lookUpObjects(name); }

	long       Verify();	         // check validity of links

#ifdef __V10_COMPATIBILITY__
	long	   IsDynamic() { return NodeFlags&DynamicNode; } // is node dynamic?
#else
	long	   IsDynamic() { return Flags&DynamicNode; } // is node dynamic?
#endif

	void	   RePost(); 		 // repost delay que to the local que

	long	   TopStack(long p = -1); // get the stack element

	char	  *TopStackSymbol(long p);// get the stack element

	char      *GetSymbol(register long symbol) // return symbol name or
		   {                               // "U N K N O W N"
		     return Tables_us->Lex->GetToken((ubrick)symbol);
		   }

	long	   AllChildrenState(long state,  // all children in state?
				    char *n = 0);// possible link name
	long	   AnyChildState(long state,     // any child in state?
				 char *n = 0);   // possible link name
////////////////////////////////////////////////////////////////////////

//
// unsupported at this time Generator 
//
////////////////////////////////////////////////////////////////////////
	long	   Generate(long length, 	// generate sample stream
			    long number, ostream &out);
////////////////////////////////////////////////////////////////////////

//
// published communication functions
//
////////////////////////////////////////////////////////////////////////
	void  Send(long);		   // send to itself
	void  Send(char *);		   // send to itself
	void  Send();			   // send to itself the last token
	void  Send(long, void *);	   // send to itself a string
	void  Send(char *, void *);	   // send to itself a token(string)
					   // and a value

	void  SendBack(long);		   // send back to the last sender
	void  SendBack(char *);		   // send back to the last sender
	void  SendBack();		   // send back to the last sender
	void  SendBack(long, void *);	   // send back to the last sender strng
	void  SendBack(char *, void *);	   // send back to the last sender
					   // a token(string) and a value

        // for sendChildren:
	// n -- optional name of link: 0 -- all children, !=0 -- link named n

	void  SendChildren(long, char *n=0);	      // send to children
	void  SendChildren(char *, char *n=0);	      // send to children
	void  SendChildren();		      	      // send to children
						      // last token
	void  SendChildren(long, void *, char *n=0);  // send to children
						      // a string
	void  SendChildren(char *, void *, char *n=0);// send to children
						      // a token(string) and
						      // a value

	void  SendParents(long);	   // send to all parents
	void  SendParents(char *);	   // send to all parents
	void  SendParents();		   // send to all parents last token
	void  SendParents(long, void *);   // send to all parents a string
	void  SendParents(char *, void *); // send to all parents
					   // a token(string) and a value

	void  Send(typeProto *, long);	        // send to some node
	void  Send(typeProto *, char *);        // send to some node
	void  Send(typeProto *);	        // send to some node last token
	void  Send(typeProto *, long, void *);  // send to some node a string
	void  Send(typeProto *, char *, void *);// send to some node a
						// token(string) and a value

	void  SendBackFront(long);	   // send back to the last sender
	void  SendBackFront(char *);	   // send back to the last sender
	void  SendBackFront();		   // send back to the last sender
	void  SendBackFront(long, void *); // send back to the last sender strng
	void  SendBackFront(char *, void *);// send back to the last sender
					   // a token(string) and a value

	// for sendChildren:
	// n -- optional name of link: 0 -- all children, !=0 -- link named n

	void  SendChildrenFront(long, char *n=0);     // send to children
	void  SendChildrenFront(char *, char *n=0);   // send to children
	void  SendChildrenFront();	      	      // send to children
						      // last token
	void  SendChildrenFront(long, void *, char *n=0);  // send to children
						      // a string
	void  SendChildrenFront(char *, void *, char *n=0);// send to children
						      // a token(string) and
						      // a value

	void  SendParentsFront(long);	   // send to all parents
	void  SendParentsFront(char *);	   // send to all parents
	void  SendParentsFront();	   // send to all parents last token
	void  SendParentsFront(long, void *);// send to all parents a string
	void  SendParentsFront(char *, void *);// send to all parents
					   // a token(string) and a value

////////////////////////////////////////////////////////////////////////
//
///// end of the published interface
//

//
// State Transition Engine
//
////////////////////////////////////////////////////////////////////////
	brick Parse(typeProto **);	   // parse incoming symbols
////////////////////////////////////////////////////////////////////////


//
// unpublished data members
//
////////////////////////////////////////////////////////////////////////

#ifdef __V10_COMPATIBILITY__
	long	NodeFlags;		 // various flags
#else					 // #ifdef __V10_COMPATIBILITY__
	long	   Flags;		 // various flags
#endif

	long	   OrderNumber;		 // order number of the static node
	char      *TablesName_us;	 // name of tables
	mskxBMon  *Tables_us;		 // address of tables
	stack     *Stack;		 // parsing stack
	errorBMon *Error;		 // error recovery object
	void      *This;  		 // to safely cast into the derived type
	localQue  *Delay;		 // delay queue
static	localQue  *Local;		 // pointer to the local queue
	char      *AutoSignalBits;	 // bits for autosignal
	typeProto();			 // constructor
virtual ~typeProto(); 		         // destructor (defined in pc.cpp)
////////////////////////////////////////////////////////////////////////


//
// The methods below are redefined by the MUSKOX compiler
// for each individual MUSKOX type
//
///////////////////////////////////////////////////////////////////////

public:

virtual brick Action0(ubrick, stack *);           //      a : {action};
virtual brick ActionSkip0(ubrick);                // skip a : {action};
virtual rule *ActionN(ubrick, stack *);           //      a : something{action};
virtual rule *ActionO(ubrick, stack *);           // single goto ActionN
virtual rule *ActionSkipN(ubrick);                // skip a : something{action};

virtual void  ExecuteTrigger(ubrick, stack *);    // execute a trigger
virtual void  ExecuteReductor(stack *, ubrick);   // execute a reductor

virtual void Init();   			   // initialize the node
virtual void Resolve();   		   // resolve references to children
virtual typeProto *New();           	   // make a duplicate
virtual void AllFixedLinks();  		   // create all fixed links
///////////////////////////////////////////////////////////////////////


//
//  various unpublished member-functions
//
////////////////////////////////////////////////////////////////////////
virtual void   SetState(register ubrick s);
virtual ubrick GetCurrentState();
	long   Link(typeProto &child, char *name);// link a child
	long   LinkChildren(char *, typeProto * = 0);// link children
	long   UnLinkChildren(typeProto *);	 // unlink children
	void   UnLinkParent(typeProto *);        // unlink a parent
virtual void   Delete(typeProto *);	         // delete this from library
	mskxBMon  *Verify(nodePointer *, int);   // check a particular link
	typeProto *Add(char *, char *);	         // add a new node of this type
    	void       Bring(char *, char *);        // works like Add, after new()
	typeProto *Copy(char *newName);	         // copy a node
	void       Copy(sAtribus &);		 // copy this node
virtual long  UserRecover(ubrick *) {return 0;}; // user error recov. procedure
virtual	brick UserLexBMon(tokenBMon *, void *);  // resolve undefined token
virtual int   Modify(char *, char *);	         // modify this node
	long  IsHidden(register long t)		 // is symbol hidden?
		     { return Tables_us->Lr1Tables->Tokens[(ubrick)t].Hidden; }
	long  HasException(mskxBMon **, mskxBMon *);// check for except_
	brick Convert(char *c) 		         // get the value of token
		     { return Tables_us->Lex->FindToken(c); }
	brick Translate(typeProto *, long);// translate token to the other type
	void  PostPone(long, void *);	   // postpone the message in Wait que
	void  SendAutoSignals(brick, void *);    //send auto signals

////////////////////////////////////////////////////////////////////////


//
// printing members
//
////////////////////////////////////////////////////////////////////////
virtual void Print(register ostream &o);	// print the node
virtual	void PrintName();
////////////////////////////////////////////////////////////////////////


//
// These unpublished member-functions are called when the node is created
//
////////////////////////////////////////////////////////////////////////
	void FillAutoSignals();		   // fill auto signal bits
	void InitStackAndError();	   // init stack and error objects
	void Address();			   // get address of parsing tables
	void Address(typeProto *&, char *, // get address of the named
		    char *, int, int v=0); // node, link parents & children
	void Address(typeProto *&, char *, // fixed link in AllFixedLinks()
	       		         int v=0);
	void Address(char *);		   // put node on the list of variable
					   // nodes
	void Address(typeProto *);	   // put node on the list of variable
					   // nodes
	void Address0(typeProto *);	   // link parent, init transtable
	void MakeAutoReports(typeProto *,  // make autoReports from fixed child
		       nodePointer *, int);
	void MakeAutoReports0(typeProto *, // make autoReports from fixed child
		       nodePointer *, autoReport *);
	void MakeAutoReportsForAnyKid(typeProto *,   // autoReports 4 any child
		       nodePointer *);
///////////////////////////////////////////////////////////////////////


#ifdef __V10_COMPATIBILITY__

# define Childs Children
# define LinkChilds LinkChildren
# define UnLinkChilds UnLinkChildren
# define LastContentScript LastAttribute
# define LastToken LastSymbol
# define AllChilds AllChildren
# define SendChilds SendChildren

	static pid_t   CmdPid;           // pid of hvdisp -u
	static int     ConfigChanged;    // set when hvmod is issued

virtual char    *GetComment() { return (char *)NULL; };
virtual char    *GetDeviceName() { return (char *)NULL; };
virtual char    *GetMountPoint() { return (char *)NULL; };
virtual char    *GetOnlineScript() { return (char *)NULL; };
virtual char    *GetOfflineScript() { return (char *)NULL; };
virtual char    *GetFaultScript() { return (char *)NULL; };
virtual int      GetPiece() { return 0; };
virtual int      GetPriority() { return 0; };
virtual int      GetAutoSwitchOver() { return 0; };
virtual int      GetAutoStartUp() { return 0; };
virtual int      AllowActDeact() { return 0; };

virtual int     GetVirLink(){return 0;};
virtual char    *GetVirParent(){return (char *)NULL; };

virtual void    DumpFields(FILE* fp)
		{
		    fprintf(fp,"*** ERROR *** Missing DumpFields() for %s\n",
				TablesName_us+5);
		};
virtual basic  *GetBasic() { return 0; };

#else					

virtual typeProto *GetBasic();

#endif					 
//
// End of Pyramid RELIANT data
//
///////////////////////////////////////////////////////////////////////

};

////////////////////////////////////////////////////////////////////////
//
// macros to send the current state from actions, triggers, and reductors
//

#ifdef __V10_COMPATIBILITY__
# define SendChildsState SendChildrenState
#endif

#define SendState(x)        Send(x, stackus->Stack->State)
#define SendBackState       SendBack(stackus->Stack->State)
#define SendChildrenState   SendChildren(stackus->Stack->State)
#define SendParentsState    SendParents(stackus->Stack->State)

////////////////////////////////////////////////////////////////////////

class	transTable			// translation tables for relatives
{
public:
	transTable *Link;		// next translation table
	mskxBMon   *Relative;		// the relative type
	brick      *Table;		// the table itself
	void        Fill(register lexBMon *from, register lexBMon *to);//fill in
	transTable(register typeProto *to, register transTable *tl,
		   register typeProto *from)
		  {
		   register lexBMon *l = from->Tables_us->Lex;
		   Relative = to->Tables_us;
		   Link = tl;
		   Table = new brick[l->Size];
		   Fill(l, Relative->Lex);
		  }
};

////////////////////////////////////////////////////////////////////////

#ifdef __V10_COMPATIBILITY__
# define LOCK_FILE "/usr/opt/reliant/.cm_lock"
#endif

typeProto *lookUpNodes(char *nodeName);
mskxBMon  *lookUpTypes(char *typeName);

class linkIterator
{
public:
	nodePointer *Current;
	nodePointer *CurrentDown;
	char        *Name;
	linkIterator(nodePointer *, char *);
	nodePointer *Next();
};

class childrenIterator : private linkIterator	// published iterator
{
public:
	childrenIterator(register typeProto *object, register char *link) :
				linkIterator(object->Children, link) {}
	typeProto *More()
	{
	 register nodePointer *n = Next();
	 return n?n->ParentChild:0;
	}
};
#endif
