/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

#ifndef __pc_h_
#define __pc_h_

#ifndef __atribus_h_
#include "atribus.h"
#endif

#ifndef __inpus_h_
#include "inpus.h"
#endif

#ifndef __skeletus_h_
#include "skeletus.h"
#endif

#ifndef __skelbmon_h_
#include "skelbmon.h"
#endif

#ifndef __nodus_h_
#include "nodus.h"
#endif

/******************************************************************_COMMENT

This file contains definitions required for standard base framework:


enum {base framework flags}	various flags that control execution of the
				framework.

bmNode				an element of the list of handles of all MUSKOX
				objects in the ARM. The list
				starts from allNodes defined in pc.cpp.
				bmNode is not the object itself - it is 
				merely a list of pointers to typeProto.

nodeAddress			this class contains virtual memory address
				and symbolic name of the MUSKOX object. 
				this class is used in tokenBMon class to
				hold the source and destination address
				of a token in the local queue.

tokenBMon			this class defines a token -- an element of a 
				local queue. Each token represents a message or
				a state through the value of 'brick Content'.
				'void *ContentScript' points to the attribute
				of the token, if non-zero. Also, the token
				contains addresses of the object that has sent
				this token, and the address of the object
				to which this token has been sent. All the
				tokens in the local queue are linked in a
				list (Link'). 

				Major member functions of tokenBMon
				are functions Make() -- they fill in the values
				for the Content (token code) and ContentScript
				(token attribute) from its various arguments.
				These member functions are called from
				the communication methods of the typeProto
				class (see nodus.h).

localQue			this class defines a local queue of symbols.
				There is a single local queue 'local' defined
				in pc.cpp that holds all the symbols addressed
				to the objects in the ARM.
				Also, each object has its own local queue
				'Local' that holds postponed symbols.

				Elements of the queue are of class tokenBMon.
				The queue consists of two link-lists: busy
				symbols (MesBusy) and  free symbols (MesFree).
				A symbol from the free list is moved into
				the busy list in member functions Append()
				(at the end of the queue) and InFront() (in
				front of the queue). These functions are
				called by the communication methods from the
				typeProto class when a symbol is sent to a
				MUSKOX object.

				When MesFree is NULL, a new tokenBMon is
				allocated from the heap using new(). This
				new symbol is put on the busy list. When a symbol 
				is retrieved by the lexical analyzer LexBMon()
				from skelBMon class (see skelbmon.h), then
				this symbol is stored back on the free list
				by invoking the Free() method of localQueue.

				Method MergeQueue() is used to merge postponed
				symbols. Method PostPone() is used to postpone
				the symbol on the object's local queue.
				Method GetMessage() is used to move
				a symbol in front of the queue for MUSKOX 
				objects: if the object is blocked, then the
				symbol that has been sent by this same object
				is moved.

outside				a class to retreive symbols coming to the ARM
				from outside of the ARM (e.g. from .inp file).

COMMENT_******************************************************************/


enum    				// base framework flags
{ 
	F_debugBM  = 1,			// printout debugging information
	F_askBM    = 2,			// ask for a symbol if queue is empty
	F_silentBM = 4,			// be silent
	F_askAttr  = 8			// ask for attribute
};


extern long flagBM;                     // base framework flags

class	bmNode				// an object
{
public:
	typeProto *Node;		// the object in the library
	bmNode    *Link;		// next object
	bmNode(register typeProto *n, register bmNode *ln) {Node = n,Link = ln;}
};


typeProto *lookUpObjects(char *);	// finds an object by its name

class	nodeAddress			// object address
{
public:
	typeProto *Address;
	char      *Name;
	nodeAddress &operator = (typeProto *d)
			{ Address = d, Name = 0; return *this; }
	nodeAddress &operator = (register char *n)
			{ Address = lookUpObjects(n); Name = n; return *this; }
	void Print(register ostream &o)
	     {
	      register typeProto *t = Address;
	      if ( !t )
	      {
	       o << "Address unknown-";
	       if ( !Name ) 
	         o << "Name unknown";
	       else
	         o << "Name <" << Name << '>';
	      }
	      else
	       o << t->ObjectName;
	     }
};


class	tokenBMon			// token format for base framework
{					// (symbol from a local que)
public:
	tokenBMon      *Link;		// link-list in the local queue
	nodeAddress	Destination;	// addressee
	nodeAddress	ReturnAddress;	// sender
	brick		Content;	// state or message code generated
					// by MUSKOX compiler
	void	       *ContentScript;	// attribute, = 0 - not specified
	void Print(register ostream &o)	// print the token (only for dbgging)
	     {
	      o << "\n\n  Content <" << Content << "\t";
	      o << Destination.Address->TablesName_us << "::";
	      o << Destination.Address->Tables_us->Lex->GetToken(Content);
	      o<< '>';
	      o << "\n  ContentScript <" << ContentScript << '>';
	      o << "\n  Destination <";
	      Destination.Print(o);
	      o << '>';
	      o << "\n  ReturnAddress <";
	      ReturnAddress.Print(o);
	      o << '>' << endl;
	     }
	void FillContent0(register char *cs)	// store attribute
	     {
	      register typeProto *t = Destination.Address;
	      if ( t )
		Content = t->Convert(cs);
	      if ( !(stack::Flag&PASSBYVALUE) )
		ContentScript = 0;
	     }

							     //
	void FillContent(register long c, register void *cs) // store attribute
							     //
	     {
	      register typeProto *t = Destination.Address;
	      register typeProto *s = ReturnAddress.Address;
	      if ( t && s )
		Content = s->Translate(t, (brick)c);
	      if ( !(stack::Flag&PASSBYVALUE) )
		ContentScript = cs;
	      else
		moveAttribute(t->Tables_us->Lr1Tables->Tokens + Content,
							ContentScript,cs);
	     }

							     //
	void FillContent(register char *c, register void *cs)// store attribute
							     //
	     {
	      register typeProto *t = Destination.Address;
	      register typeProto *s = ReturnAddress.Address;
	      if ( t )
		Content = t->Convert(c);
	      if ( !(stack::Flag&PASSBYVALUE) )
		ContentScript = cs;
	      else
		moveAttribute(t->Tables_us->Lr1Tables->Tokens + Content,
							ContentScript,cs);
	     }

					//
					// Make the token, called
					// by Send() communication methods
					// from typeProto class
					//

	void Make(register char *d, register char *r, register char *cs )
	     {
	      Destination = d;
	      ReturnAddress = r;
	      FillContent0(cs);
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
		   bml(TLEV) << "Make char * char * char *:";
		   if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
	void Make(register char *d, register char *r, register long c)
	     {
	      Destination = d;
	      ReturnAddress = r;
	      Content = (brick)c;
	      if ( !(stack::Flag&PASSBYVALUE) )
		ContentScript = 0;
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
		   bml(TLEV) << "Make char * char * brick:";
		   if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
	void Make(register typeProto *d, register typeProto *r, 
		  register char *cs) 
	     {
	      Destination = d;
	      ReturnAddress = r;
	      FillContent0(cs);
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
		   bml(TLEV) << "Make typePRoto * typePRoto * char *:";
		   if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
	void Make(register typeProto *d, register typeProto *r,
		  register long c)
	     {
	      Destination = d;
	      ReturnAddress = r;
	      Content = (brick)c;
	      if ( !(stack::Flag&PASSBYVALUE) )
		ContentScript = 0;
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
		   bml(TLEV) << "Make typePRoto * typePRoto * brick:";
		 if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
	void Make(register typeProto *d, register typeProto *r,
		  register long c, register void *cs)
	     {
	      Destination = d;
	      ReturnAddress = r;
	      FillContent((brick)c, cs);
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
	           bml(TLEV) << "Make typeProto * typeProto * brick void *:"; 
		   if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
	void Make(register char *d, register char *r,
		  register long c, register void *cs)
	     {
	      Destination = d;
	      ReturnAddress = r;
	      FillContent((brick)c, cs);
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
	           bml(TLEV) << "Make char * char * brick void *:"; 
		   if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
	void Make0(register typeProto *d, register typeProto *r,
		   register long c, register void *cs)
	     {
	      Destination = d;
	      ReturnAddress = r;
	      Content = (brick)c;
	      if ( !(stack::Flag&PASSBYVALUE) )
		ContentScript = cs;
	      else
		moveAttribute(d->Tables_us->Lr1Tables->Tokens + Content,
							ContentScript,cs);
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
		   bml(TLEV) << "Make typeProto * typeProto * brick void *:";
		   if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
	void Make(register typeProto *d, register typeProto *r, 
		  register char *c, register void *cs)
	     {
	      Destination = d;
	      ReturnAddress = r;
	      FillContent(c, cs);
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
		   bml(TLEV) << "Make typeProto * typeProto * brick void *:";
		   if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
	void Make(register char *d, register char *r, 
		  register char *c, register void *cs) 
	     {
	      Destination = d;
	      ReturnAddress = r;
	      FillContent(c, cs);
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
		   bml(TLEV) << "Make char * char * brick void *:";
		   if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
	void PostPone(register typeProto *d, register typeProto *r,
		  register long c, register void *cs)
	     {
	      Destination = d;
	      ReturnAddress = r;
	      Content = (brick)c;
	      if ( !(stack::Flag&PASSBYVALUE) )
		ContentScript = cs;
	      else
		moveAttribute(d->Tables_us->Lr1Tables->Tokens + Content,ContentScript,cs);
#	      if defined(DEBUGON)
		if ( flagBM&F_debugBM )
		{
		   bml(TLEV) << "PostPone typeProto * typeProto * brick void *:";
		   if ( (1 << TLEV) & bmLogLevel )
		       Print(*bmLog);
		}
#	      endif
	     }
};


					//
class	localQue			// queue of symbols (states & messages)
{					//
public:
	tokenBMon  *MesFree;	// free message que
	tokenBMon  *MesLast;	// last busy message
	tokenBMon  *MesBusy;	// busy message que (the first one)
	localQue()  { MesFree = MesLast = MesBusy = 0; }
	void	    MergeQue(register localQue *l) // merge que l to this
		    {
#                   if defined(DEBUGON)
                      if ( flagBM&F_debugBM )
                      {
                           bml(QLEV) << "Merge to local que:";
                             if ( (1<< QLEV) & bmLogLevel )
                                Print(*bmLog);
                      }
#                   endif
		     if ( l && l->MesBusy ) // there is something on the que l
		     {
		       if ( MesLast )
			 MesLast->Link = l->MesBusy;
		       else
			 MesBusy = l->MesBusy;
		       MesLast = l->MesLast;
		       l->MesLast = 0;
		       l->MesBusy = 0;
		     }
		    }
	void Print(register ostream &o);	// only for debugging
	void         Free() 			// free the token for reuse
		     {
#		      if defined(DEBUGON)
			if ( flagBM&F_debugBM )
			{
			   bml(QLEV) << "Free local que:";
			   if ( (1<< QLEV) & bmLogLevel )
			       Print(*bmLog);
			}
#		      endif
		      register tokenBMon *b = MesBusy;
		      if ( b )
			MesBusy = b->Link, b->Link = MesFree, MesFree = b;
		      if ( !MesBusy )
			MesLast = 0;
		     }
	tokenBMon *Append()	// put a symbol at the tail of local que
		   {
#		    if defined(DEBUGON)
		      if ( flagBM&F_debugBM )
                      {
			 bml(QLEV) << "Append local que:";
   			     if ( (1<< QLEV) & bmLogLevel )
			        Print(*bmLog);
                      }
#		    endif
		    register tokenBMon *f = MesFree;
		    if ( !f )
		    {
		      f = MesFree = new tokenBMon;
		      f->Link = 0;
		      if ( stack::Flag&PASSBYVALUE )
			 f->ContentScript = new char[(int)stack::TotalAtrMaxSize];
		    }
		    if ( MesLast )
		      MesLast->Link = f;
		    else
		      MesBusy = f;
		    MesLast = f,
		      MesFree = f->Link,
			f->Link = 0;
		    return f;
		   }
	tokenBMon *InFront()	// put a symbol in front of que
		   {
#		    if defined(DEBUGON)
		      if ( flagBM&F_debugBM )
		      {
			 bml(QLEV) << "InFront local que:";
		 	   if ( (1<< QLEV) & bmLogLevel )
			        Print(*bmLog);
		      }
#		    endif
		    register tokenBMon *f = MesFree;
		    if ( !f )
		    {
		      f = MesFree = new tokenBMon;
		      f->Link = 0;
		      if ( stack::Flag&PASSBYVALUE )
			 f->ContentScript = new char[(int)stack::TotalAtrMaxSize];
		    }
		    MesFree = f->Link,
		      f->Link = MesBusy,
			MesBusy = f;
		    if ( !MesLast )
		      MesLast = f;
		    return f;
		   }
	tokenBMon *GetMessage();	// move symbol in front of the queue
};


class outside               		// return a symbol sent from the
{                                       // the outside
public:
	long Line;
        outside() { Line = 0;}
virtual tokenBMon *GetMessage();        // redefine it if needed
virtual	tokenBMon *AskForMessage();
};



void printHelp();

#endif
