/////////////////////////////////////////////////////////////
//                                                         //
// Copyright 1994, 1995, 1996                              //
// MASTER INFORMATION SYSTEMS CORPORATION                  //
// MUSKOX Version 1.2					   //
// All Rights Reserved                                     //
//                                                         //
/////////////////////////////////////////////////////////////

#ifndef __skeletus_h_
#define __skeletus_h_

#ifndef __atribus_h_
#include "atribus.h"
#endif

#ifndef __inpus_h_
#include "inpus.h"
#endif


/******************************************************************_COMMENT

This file defines data for the MUSKOX State Transition Engine:

macros BRICK...			define various sizes of the parsing tables for
				the State Transition Engine (engine). MUSKOX 
				compiler generates these tables from the type 
				grammar.

offsets				these offsets are used by the engine while 
				shifting the symbols onto the parsing stack.

enums				define various enumerations that distinguish
				different kinds of states for engine.

_cpy_				defines how atributes will be copied on
				the stack: by memcpy or by a copy-constructor.
				See atribus.h for UNIONACT that controls 
				definition of _cpy_

packIt				defines the layout of the elements of the engine 
				symbol table.

moveAttribute			defines a function or a macro to copy attributes
				on the stack. See atribus.h for COPYCONSTRUCTOR
				that controls definition of moveAttribute.

STACK				defines the parsing stack size for the engine.

RNDUP				define the varioant of sizeof() that rounds
mskxSize()			the argument modulo sizeof(void *).

stackElement			defines the layout of the stack element
				for the engine.

PASSBYVALUE			distinguish how attributes will be passed
PASSBYREFERENCE			with the symbols on the local queue

stack				defines the structure of the parsing stack
				for each MUSKOX node.

stackSave			an element of the save stack that is used
				by the engine to save current stack position
				while verifying the validity of the
				applied reduction.

keyWord				defines the layout of the tables of symbols
Dim()				(terminals, literals, and nonterminals). These
				tables are generated by the MUSKOX compiler.
				Dim() calculates the size of each table.
			
rule				defines an element of the array of state 
				transition rules that is generated by the
				MUSKOX compiler.

F_realTime			enumeration that defines realtime and
F_BMon                          standard framework options

lex				defines the data layout for the MUSKOX lexical
				analyzer. This layout is used by the stand-alone
				engine. The ARM uses lexBMon.

act_				typedef for the action invokation function.

actions				defines a base class for typeProto. Contains
				action invokation functions that invoke
				actions, triggers, and reductors.

actKind				enumerates kinds of actions that are deferred
				until engine veryfies the validity of the reduction

actionSave			element of the stack of actions that are saved
				by engine while it is determining the validity
				of the reduction.

generator			is not used in the current version

F_NOPOP				enumerator that defines options of
F_NOSKIP			MUSKOX Error processing routines

error				error recovery class. Defines methods
				invoked when an unexpected symbol arrives
				to an object and a parsing error occures.

Parsing flags			various flags for the State Transition Enigne

lr1Tables			defines the layout of the parsing tables
				generated by the MUSKOX compiler for every
				MUSKOX type. These tables are shared between
				objects of the same type and are used by the
				engine in the course of parsing incoming symbols.

mskx				contains entry points to all the major 
				run-time data structures of the MUSKOX type.
				generated by the MUSKOX compiler. Used only
				in the stand-alone engine. 

typeDescriptor			element of the array of MUSKOX types that
				is generated by the MUSKOX compiler. Used only
				for the stand-alone engine.

SHIFT				macros used in the engine
REDUCE2

mylex				a derived class from lex: user should
				redefine the lexical analyzer in this class
				for customized lexical analysis.

myError				a derived class from error: user should
				redefine error recovery and error reporting
				routines in this class for customized error
				processing

COMMENT_******************************************************************/


//
// macros BRICK...
//
//////////////////////////////////////////////////////////////////////////
#	define BRICKSHORT
	typedef unsigned short ubrick;	// the building brick of the tables
	typedef short brick;
	typedef unsigned char smubrick;
#	define BRICKSHIFT      16
#	define BRICKHEAD       ((unsigned short)0xE000)
#	define BRICKFFFF       ((unsigned short)0xFFFF)
#	define BRICKEFFF       ((unsigned short)0xEFFF)
#       define BRICKHEADREDUCE ((unsigned short)0xF000)
#	define BRICKHEADREDUCESHIFT 12
#	define BRICKHEADSHIFT  13
# if !(defined(__sparc) && defined(__SVR4)) // not on Sparc C++ 4.0.1
#	define BRICKPACKSHIFT     4	    // format of the packIt structure
#	define BRICKPACKHIDDEN    1
#	define BRICKPACKPUBLIC    1
#	define BRICKPACKSKIP      1
#	define BRICKPACKCOPY      1
#	define BRICKPACKWAIT      1
#       if defined(__BCPLUSPLUS__) 
#		define BRICKPACKOFFSET    11
#	else
#		define BRICKPACKOFFSET    16
#	endif
# endif
//////////////////////////////////////////////////////////////////////////


//
// offsets
// constants for offsets in shift procedures
//
//////////////////////////////////////////////////////////////////////////
#define H1 (((ubrick)B_STATE1_)<<BRICKHEADSHIFT)
#define H2 (((ubrick)B_STATE2_)<<BRICKHEADSHIFT)
#define H3 (((ubrick)B_STATE3_)<<BRICKHEADSHIFT)
#define H4 (((ubrick)B_STATE4_)<<BRICKHEADSHIFT)
//////////////////////////////////////////////////////////////////////////

//
// enums
//
// constants in the low bits of ShiftHash array indicate the type of state
//
//////////////////////////////////////////////////////////////////////////
enum 	   {
		B_REGSTATE_   = 0,      // more than 4 outgoing edges
		B_SIMPLE1_    = 1,	// simple state with 1 outgoing edge
		B_STATE1_     = 2,      // 1 outgoing edge
		B_STATE2_     = 3,      // 2 outgoing edges
		B_STATE3_     = 4,      // 3 outgoing edges
		B_STATE4_     = 5,      // 4 outgoing edges
		B_FINALSTATE_ = 6       // 0 outgoing edges - final state,
					// or a single rule a : [action];
	   };

//
//constants in the low bits of ReduceHash array indicate the type of state
//

enum       {
		B_ERROR_                = 0,    // simple state at reduction

////////////////// states with one reduction

		B_SIMPLESTATE_          = 0,    // simple state
		B_REDUCE1_              = 1,    // just 1 reduction
		B_REDUCE2MORE_          = 2,    // 2 or more reductions
		B_REDUCE0_ACTION_       = 3,    // a : {action};
		B_REDUCEN_ACTION_       = 4,    // a : some things {action};
		B_SINGLE0_              = 5,    // single rule a : ;
		B_SINGLE0_ACTION_       = 6,    // single rule a : {action};
		B_ACCEPT_               = 7,    // unique accept state

// reductions with single goto

		B_REDUCE1_SINGLE_       = 8,    // just 1 reduction
		B_REDUCEN_ACTION_SINGLE_= 9,    // a : bla bla {action}

////////////////// states with more than one reduction

		B_ERROR_2MORE_          = 0,    // no reduction in 2MORE
		B_REDUCE1_2MORE_        = 1,    // no action in 2MORE
		B_REDUCE0_ACTION2MORE_  = 2,    // 0_ACTION in 2MORE
		B_REDUCEN_ACTION2MORE_  = 3,    // N_ACTION in 2MORE

// reductions with single goto

		B_REDUCE1_2MORE_SINGLE_       = 4, // no action in 2MORE
		B_REDUCEN_ACTION2MORE_SINGLE_ = 5  // N_ACTION in 2MORE
	   };
//////////////////////////////////////////////////////////////////////////


//
// _cpy_
//
//////////////////////////////////////////////////////////////////////////
typedef long (*_copy_)(void *, void *);	// use copy constructor for attribute's
					// movement within the stack

#ifdef UNIONACT
 union	_cpy_				// copy attributes within the stack
#else
 struct	_cpy_				// copy attributes within the stack
#endif
{
	_copy_  Method;			// use copy constructor
	ubrick	SizeOfAttribute;	// use memcpy
};
//////////////////////////////////////////////////////////////////////////


//
// packIt
//
//////////////////////////////////////////////////////////////////////////
class packIt {			// defines a symbol in the parsing tables
public:
	smubrick        Mask;
	smubrick        Local;

#if defined(__sparc) && defined(__SVR4) // only for sparc C++ 4.0.1
	ubrick 	Shift;
	ubrick 	Hidden;
	ubrick 	Public;
	ubrick 	Skip;
	ubrick 	Copy;
	ubrick 	Wait;
	ubrick 	Offset;
#else
	unsigned int 	Shift:BRICKPACKSHIFT;
	unsigned int	Hidden:BRICKPACKHIDDEN;
	unsigned int    Public:BRICKPACKPUBLIC;
	unsigned int    Skip:BRICKPACKSKIP;
	unsigned int    Copy:BRICKPACKCOPY;
	unsigned int	Wait:BRICKPACKWAIT;
	unsigned int	Offset:BRICKPACKOFFSET;
#endif

 	_cpy_		Move;
	ubrick		Trigger;
};


//
// moveAttribute
//
//////////////////////////////////////////////////////////////////////////

			// move attributes onto the stack

#ifdef COPYCONSTRUCTOR
	static inline void moveAttribute(register packIt *p,
				 	 register void *to, 
					 register void *from)
	{
	 if ( p->Copy )
 	  (void)((*(p->Move.Method))(to, from));
	 else
 	  memcpy(to, from, p->Move.SizeOfAttribute);
	}
#else
# define moveAttribute(p,to,from) memcpy(to, from, (p)->Move.SizeOfAttribute)
#endif
//////////////////////////////////////////////////////////////////////////


//
// STACK
//
//////////////////////////////////////////////////////////////////////////

			// stack size

#ifndef STACK
#	define STACK 50
#endif
//////////////////////////////////////////////////////////////////////////


//
// RNDUP
// mskxSize()
//
//////////////////////////////////////////////////////////////////////////

			// round up the size

#define RNDUP (sizeof(void *) - 1)
#define mskxSize(x) ((sizeof(x)+RNDUP)&~RNDUP)
//////////////////////////////////////////////////////////////////////////


//
// stackElement
//
//////////////////////////////////////////////////////////////////////////
class	stackElement			// element of the engine parsing stack
{
public:
	ubrick	 State;			// automaton state
	char    *Atribut;		// the attribute
	stackElement() { State = 0, Atribut = 0; }
};
//////////////////////////////////////////////////////////////////////////


//
// PASSBYVALUE
// PASSBYRFERENCE
//
//////////////////////////////////////////////////////////////////////////
enum
{
	PASSBYVALUE = 1,
	PASSBYREFERENCE = 2
};
//////////////////////////////////////////////////////////////////////////


//
// stack
//
//////////////////////////////////////////////////////////////////////////
struct stackSave;

class	stack				// engine parsing stack
{
public:
	int	       Depth;		// remember these parameters
	packIt	      *PackIt;		// to copy the stack for
	long	       Size;		//  recursive
	ubrick	       State;		//    compilation

	char          *AtrStackLimit;	// attribute stack limit
	long	       AtrMaxSize;	// maximum size of the attribute
	char	      *Atribut;		// $$ attribute area
	char	      *LexAtribut;	// lexical analyzer attribute area
	char	      *LexAtribut2;	// lexical analyzer attribute area 2
	stackElement  *StackLimit;	// state stack limit
	stackElement  *StackHigh;	// highest state stack value possible
	stackElement  *Stack;		// current of state stack
static  long	       TotalAtrMaxSize;	// total max of AtrMaxSize
static  long	       Flag;		// varios flags
	stackSave     *Save;		// remember stackSave in realtime
	stackSave     *SaveEnd;         // end of stackSave in realtime
	void Init(int, packIt *, long, ubrick);
	stack(int depth, packIt *p, long size, ubrick state)
	     { this->Init(depth, p, size, state); }
	stack(const stack &s)		// copy constructor for recursive comp.
	     { this->Init(s.Depth, s.PackIt, s.Size, s.State); }
	stack(){}
	~stack()
	{
	 delete [] Atribut;
	 delete [] LexAtribut;
	 delete [] LexAtribut2;
	 delete [] AtrStackLimit;
	 delete [] StackLimit;
	}
	void Reset(register ubrick state)
	{ Stack = StackLimit, Stack->State = state; }
	int CalcMaxAtrSize(packIt *, long);
	ubrick GetSave(stackElement *);		// get stack symbol in real time
};
//////////////////////////////////////////////////////////////////////////


//
// stackSave
//
//////////////////////////////////////////////////////////////////////////
struct stackSave			// save stack on goto in realtimeparse
{
	stackElement *Stack;		// remember stack position
	ubrick	      State;		// remember state
};
//////////////////////////////////////////////////////////////////////////


//
// keyWord
// Dim()
//
//////////////////////////////////////////////////////////////////////////
class	keyWord				// table of names of MUSKOX symbols
{
public:
	ubrick  DerivedLength;		// min length of the derived string
	ubrick	Length;			// length of the word
	char   *Word;			// the word itself
	ubrick	Result;			// result returned by lexer
};

#define DIM(x) (int)(sizeof(x)/sizeof(keyWord))
//////////////////////////////////////////////////////////////////////////


//
// rule
//
//////////////////////////////////////////////////////////////////////////
class	rule				// rule
{
public:
				// goto after reduce or
	brick   State;          // single goto state if optimized
	brick	Reduce;         // # of elements in the right part
};
//////////////////////////////////////////////////////////////////////////


//
// F_realTime
// F_BMon
//
//////////////////////////////////////////////////////////////////////////
enum { F_realTime = 1, F_BMon = 2 };
//////////////////////////////////////////////////////////////////////////


//
// lex
//
//////////////////////////////////////////////////////////////////////////
// There are several user member functions for lexical analysis:
//
//	PureUserLex() 	may be redefined and then called
//                      when user does not want to have any MSKX
//			lexical analysis.  Must return one of the predefined
//			symbols values.  Definitions of symbols
//			is generated by MSKX.
//
//      Lex()		called instead of PureUserLex() when
//			user wants to look for the MSKX defined terminals
//			via LookUp() function and via another user defined
//			lexical analyzer UserLex() (see below).  Returns
//			Result, which is a code of the symbol found
//			by GetToken(), LookUp(), or UserLex().
//
//      GetToken()	must be redefined by user so that it extracts
//			a symbol from the input stream.  If it finds a symbol,
//			then it may set Token to the symbol found.  If it
//			returns 1, then it also must set Result which
//			then will be returned by the calling function Lex().
//			If it returns 0, it must set Token so that the
//			calling function Lex() will invoke LookUp() functions
//			to search for the Token found in the look-up tables
//			Literals, Terminals, or NonTerminals.
//
//			Usually, input stream may be divided
//			by symbols which are words, combinations of
//			special symbols (like ++), separators (like ,),
//			spaces (like \s, \n), etc.
//
// LookUpLiteral()	looks up the tables of symbols Literals, Terminals, and
// LookUpTerminal()     NonTerminals against the Token found in the input
// LookUpNonTerminal()	stream.  If succeeds, returns 1 and sets the Result.
//			Otherwise returns 0.
//
//	UserLex()	called when GetToken() and all LookUp...() failed.
//
//
//	When Parse() is called to parse the source (see below),
//	one of it's arguments is a pointer to member function of
//	class lex.  It must be set either to PureUserLex(), or to Lex().
//
//	Usually, lexical analyzers expect only literals or values of terminals
//	to come from the input stream.  However, the MSKX lexical analyser
//	may be redefined by user so that it may expect names of nonterminals
//	or names of terminals in the input stream.  For that three arrays
//	Literals, Terminals, and NonTerminals are supplied in the lex
//	class.  These arrays contain symbol numbers and symbol's literal
//	presentation.
//
//	In order to provide flexibility and at the same time effeciency
//	of lexical analysis, lex has five function pointers: lex1,...,lex5.
//	Lex() calls lex1, lex2, ... up to lex5 one after another untill
//      the next called function returns 1 and sets the Result to return
//	from Lex().  For example, if user does not want to recognize
//	names of terminals and nonterminals in the input stream,
//	the pointers may be set this way:
//
//	Lex1  =	GetToken;
//	Lex2  =	LookUpLiteral;
//	Lex3  = UserLex;
//
//	If user expects names of nonterminals, the sequence may be:
//
//	Lex1  =	GetToken;
//	Lex2  =	LookUpLiteral;
//	Lex3  =	LookUpNonTerminal;
//	Lex4  =	UserLex;
//
//	The extreme case may look like this:
//
//	Lex1  =	GetToken;
//	Lex2  =	LookUpLiteral;
//	Lex3  = LookUpTerminal;
//	Lex4  =	LookUpNonTerminal;
//	Lex5  =	UserLex;
//
//	Member function DefaultLexSet() is called from the constructor and
//	sets the first 3 lexers as GetToken, LookUpLiteral, and UserLex.
//
//	Member function Nop() is a noop, returning 1.
//////////////////////////////////////////////////////////////////////////
class lex
{
public:
	keyWord  *Literals;		// array of literals
	keyWord  *Terminals;		// array of terminals
	keyWord	 *NonTerminals;		// array of nonterminals
	ubrick	  LiteralsSize;         // sizes
	ubrick	  TerminalsSize;        //       of these
	ubrick	  NonTerminalsSize;     //                arrays
	keyWord  *Token;                // lexem found by GetToken()
	ubrick	  Result;		// result returned by lexer
	ubrick	  Other;		// error returned by lexer
	ubrick	  EndOfStream;		// end of stream
	ubrick	  Alternative;		// return from Lex() if others fail
	int	  Flag;			// flags
	input	 *Input;		// input buffer
	ubrick	 (lex::*Lex1)(void *);  // pointers to lexical analyzers
	ubrick	 (lex::*Lex2)(void *);
	ubrick	 (lex::*Lex3)(void *);
	ubrick	 (lex::*Lex4)(void *);
	ubrick	 (lex::*Lex5)(void *);
virtual ubrick   UserLex(void *);          // lexical analyzers
virtual ubrick   LookUpLiteral(void *);
virtual ubrick   LookUpTerminal(void *);
virtual ubrick   LookUpNonTerminal(void *);
	ubrick   GetLiteral(char *);
	ubrick   GetTerminal(char *);
	ubrick   GetNonTerminal(char *);
	ubrick   GetToken(register char *);
virtual char* GetToken(ubrick);
virtual ubrick   GetToken(void *);
virtual ubrick   SetAlternative(void *);
virtual ubrick   Nop(void *);
	void  DefaultLexSet()        	// set default sequence of lexers
	{
	 Lex1 = &lex::GetToken;
	 Lex2 = &lex::LookUpLiteral;
	 Lex3 = &lex::UserLex;
	 Lex4 = &lex::Nop;
	 Lex5 = &lex::Nop;
	}
	void  DebugLexSet()        	// set debugging sequence of lexers
	{
	 Lex1 = &lex::GetToken;
	 Lex2 = &lex::LookUpLiteral;
	 Lex3 = &lex::LookUpTerminal;
	 Lex4 = &lex::LookUpNonTerminal;
	 Lex5 = &lex::UserLex;
	}
	lex(register keyWord *lit,  register ubrick litSize,
	    register keyWord *term, register ubrick termSize,
	    register keyWord *non,  register ubrick nonSize,
	    register ubrick endOfStream, register ubrick other,
	    register keyWord *key,  register input *inp,
	    register void (lex::*set)())
	{
	 EndOfStream      = endOfStream;
	 Other 		  = Result = Alternative = other;
	 Terminals        = term;
	 NonTerminals     = non;
	 Literals         = lit;
	 TerminalsSize    = termSize;
	 NonTerminalsSize = nonSize;
	 LiteralsSize     = litSize;
	 Token            = key;
	 Input		  = inp;
	 Flag             = 0;
	 (this->*set)();
	}
	lex() {}
virtual ubrick PureUserLex(void *);
virtual ubrick Lex(void *);
	void Store()
	{
#ifdef INPUS
	 Input->Store();
	 Token->Word = Input->Next - 2;
	 Token->Length = 1;
#endif
	}

#ifdef INPUS
	void Store(register char *x)
	{
	 Token->Word = x;
	 Token->Length = (ubrick)(Input->Next - x - 1);
	}
#else
	void Store(register char *)
	{
	}
#endif
};
//////////////////////////////////////////////////////////////////////////


//
// act_
//
//////////////////////////////////////////////////////////////////////////
typedef brick (*act_)(stack *);	// action functions
//////////////////////////////////////////////////////////////////////////


//
// actions
//
//////////////////////////////////////////////////////////////////////////
class actions				// muskox actions
{
public:
	int	 DefReductor;		// trigger index of the reductor
	int	 ReduceBy;		// length of the rule to be reduced
	lex	*Lex;			// lexical analyzer
        long     ActionFlag;            // action flag
        enum     { TransOn = 1 };
        void  TranslateOff() { ActionFlag &= ~TransOn; }
	void  TranslateOn()  { ActionFlag |= TransOn; }
        long  IsTranslateOn(){ return ActionFlag&TransOn; }
	actions() { Lex = 0; DefReductor = 0; ReduceBy = 0; ActionFlag = 0; }
	int	 GetToken(register char *c) { return Lex->GetToken(c); }
virtual	ubrick	 GetCurrentState();                // redefine these two in
virtual void	 SetState(ubrick);  		   // typeproto
virtual void     SendAutoSignals(brick, void *);   // defined in typeProto

///////
/////// these functions know about arrays of actions, triggers, and reductors
///////

virtual brick Action0(ubrick, stack *);            //      a:{ action };
virtual brick ActionSkip0(ubrick);                 // skip a:{ action };
virtual rule *ActionN(ubrick, stack *);            //      a:something{action};
virtual rule *ActionO(ubrick, stack *);            // single goto ActionN
virtual rule *ActionSkipN(ubrick);                 // skip a:something{action};

virtual void  ExecuteTrigger(ubrick, stack *);     // execute a trigger
virtual void  ExecuteReductor(stack *, ubrick);    // execute a reductor
};
//////////////////////////////////////////////////////////////////////////

class lr1Tables;

//
// actKind
//
//////////////////////////////////////////////////////////////////////////
enum	actKind				// in real time parsing, remember
{					// the kind of defferred action
	ACTION0,			// so that it will be executed
	ACTIONN,			// in actionSave::Execute()
	ACTIONNS,
	DEFTRIGGER,
	TRIGGER,
	REDUCTOR,
	AUTOSIGNAL
};
//////////////////////////////////////////////////////////////////////////


//
// actionSave
//
//////////////////////////////////////////////////////////////////////////
class actionSave			// save actions in realtime parse
{
public:
	stackElement *Stack;			  // remember stack position
	actKind       Action;			  // type of action
	ubrick        Index;			  // remember action function
	stackSave    *Save;			  // remember current stackSave
	static void Execute(actions *, actionSave *, // execute saved actions
			    actionSave *, lr1Tables *, stack *);
};
//////////////////////////////////////////////////////////////////////////


//
// generator
//
//////////////////////////////////////////////////////////////////////////
enum  { F_messages = 1 };               // do not generate states

class generator                         // generator tables
{
public:
	ubrick  *GenShiftHash;          // hash for generator' shift states
        ubrick  *Edges;                 // outgoing edges
        ubrick  *GenReduceHash;         // hash for generator' reduce states
        ubrick  *Reductions2;           // reductions (states with >= 2 red.)
        ubrick   StatesNumber;          // number of states
        ubrick   MaxTerminal;           // maximum terminal value
        ubrick   NonTerminalsCounter;   // number of nonterminals
        ubrick  *Rules;                 // entries to rules (indxd by lft prt)
        ubrick  *Lengths;               // pos. in Members and lengths of rules
        ubrick  *Members;               // members of right parts
        ubrick   Flag;                  // flags for generation
	ostream  *Cout;			// output stream
	ubrick  *FindTransition(ubrick state, ubrick symbol, lr1Tables *);
	generator(
		  ubrick  *genShiftHash = 0,
		  ubrick  *edges = 0,
		  ubrick  *genReduceHash = 0,
		  ubrick  *reductions2 = 0,
		  ubrick   statesNumber = 0,
		  ubrick   maxTerminal = 0,
		  ubrick   nonTerminalsCounter = 0,
		  ubrick  *rules = 0,
		  ubrick  *lengths = 0,
		  ubrick  *members = 0,
		  ostream *ccout = &cerr,
		  ubrick   flag = 0
		 )
		{
			GenShiftHash = genShiftHash;
			Edges = edges;
			GenReduceHash = genReduceHash;
			Reductions2 = reductions2;
			StatesNumber = statesNumber;
			MaxTerminal = maxTerminal;
			NonTerminalsCounter = nonTerminalsCounter;
			Rules = rules;
			Lengths = lengths;
			Members = members;
			Flag = flag;
			Cout = ccout;
		}
};

//////////////////////////////////////////////////////////////////////////


//
// F_NOPOP
// F_NOSKIP
//
//////////////////////////////////////////////////////////////////////////
enum
{
	F_NOPOP  = 1,			// do not pop stack on errors
	F_NOSKIP = 2			// do not skip symbols on errors
};
//////////////////////////////////////////////////////////////////////////


//
// error
//
//////////////////////////////////////////////////////////////////////////
class error                    		// error recovery class
{
public:
	int	Number;			// number of errors
	int	UserResult;             // UserRecover() sets this to return
					// from Recover() (see Recover())
	int	SkipTokens;		// how many symbols to skip until
					// parsing stops
	int	Flag;			// flags
	error(register int s)
		  { Number = 0, UserResult = 0, SkipTokens = s,
			 Flag = F_NOPOP | F_NOSKIP; }
	error() {Number = 0, UserResult = 0, SkipTokens = 100,
			 Flag = F_NOPOP | F_NOSKIP; }
virtual void StackOverflow(lr1Tables *);
virtual void SaveStackOverflow(lr1Tables *);
virtual	int  UserRecover(lr1Tables *, stack *, lex *,
			 ubrick (lex::*)(void *), ubrick *)
			 { return 0; }
virtual	int  Recover(lr1Tables *, stack *, lex *,
		     ubrick (lex::*)(void *), ubrick *, void *);
virtual void ReportError(lr1Tables *, stack *, lex *, ubrick *);
virtual int  ErrorLimit();
};
//////////////////////////////////////////////////////////////////////////


//
// parsing flags
//
//////////////////////////////////////////////////////////////////////////
enum
{
	F_DEBUG		= 1,		// log parsing proccess
	F_ACTION	= 2,		// execute actions while parsing
	F_DEBUGRECOVER  = 4, 		// log state recovery in parsing process
	F_INTERACTIVE   = 8		// report parsing errors on cerr
};

enum
{
	F_goto          = 1		// Goto in progress
};
//////////////////////////////////////////////////////////////////////////


//
// lr1Tables
//
//////////////////////////////////////////////////////////////////////////
class lr1Tables {			// LR1 parsing tables
public:
	char      *LanguageName;// language name
	ubrick    *ReduceHash;  // hash for reductions
	keyWord  **DebugHash;	// state hash for debugging
	ubrick    *Reduce1;     // array of reductions
	ubrick    *Reduce2Table;// array of states and lookaheads
				// with reductions
	ubrick    *ShiftHash;   // hash for shifts
	ubrick    *Out1;        // array of shifts of states with 1 shift
	ubrick    *Out2;	//                                2 shifts
	ubrick    *Out3;	//				  3 shifts
	ubrick    *Out4;	//				  4 shifts
	ubrick    *Rows;	// array of packed shifts for >= 5 shifts
	ubrick    *Translate;   // translate targets from select rules
	packIt    *Tokens;	// decoding bits in Rows
	ubrick     EndOfStream;	// message, returned at the end of input
				// stream
	ubrick	   Error;	// state for error processing
	ubrick	   Trigger;	// message for default trigger
	ubrick	   Reductor;	// message for reductor
	ubrick     Accept;      // state, to which the whole stream
				// is reduced
	ubrick     Start;	// start state
	ubrick	   Dummy;	// never returned by lexer
	ubrick     StartState;	// Start state
	brick      Init;	// init state
	brick      ErrorMessage;// error message
	stack     *Stack;       // pointer to the stack
	lex       *Lex;		// pointer to the lexical analyzer
	error     *ErrorProc;	// error proccessing
static	int        Flag;	// flag (F_DEBUG, F_ACTION)
static	ostream   *Cerr;	// error report and debugging device
	generator *Generator;   // generator data structure
	long	   DebugFlag;   // runtime flags for logging and debugging
	lr1Tables(register char      *name,
		  register ubrick    *reduceHash,
		  register keyWord  **debugHash,
		  register ubrick    *reduce1,
		  register ubrick    *reduce2Table,
		  register ubrick    *shiftHash,
		  register ubrick    *out1,
		  register ubrick    *out2,
		  register ubrick    *out3,
		  register ubrick    *out4,
		  register ubrick    *rows,
		  register ubrick    *tran,
		  register packIt    *tokens,
		  register ubrick      endOfStream,
		  register ubrick      err,
		  register ubrick      trig,
		  register ubrick      reduc,
		  register ubrick      accept,
		  register ubrick      start,
		  register ubrick      dummy,
		  register ubrick     startState,
		  register brick      init,
		  register brick      errorMessage,
		  register stack     *s,
		  register lex	     *l,
		  register error     *errorProc,
		  register generator *g)
	{
	 LanguageName = name;
	 ReduceHash   = reduceHash;
	 DebugHash    = debugHash;
	 Reduce1      = reduce1;
	 Reduce2Table = reduce2Table;
	 ShiftHash    = shiftHash;
	 Out1         = out1;
	 Out2         = out2;
	 Out3         = out3;
	 Out4         = out4;
	 Rows         = rows;
	 Translate    = tran;
	 Tokens       = tokens;
	 EndOfStream  = endOfStream;
	 Error        = err;
	 Trigger      = trig;
	 Reductor     = reduc;
	 Accept	      = accept;
	 Start        = start;
	 Dummy        = dummy;
	 StartState   = startState;
	 Init         = init;
	 ErrorMessage = errorMessage;
	 Stack        = s;
	 Lex 	      = l;
	 ErrorProc    = errorProc;
	 Flag         = F_ACTION;
	 Cerr         = &cerr;
	 Generator    = g;
	 DebugFlag    = 0;
	}

//
// stand-alone engine (non-real time parse)
//
/////////////////////////////////
virtual ubrick Parse(stack *, lex *, ubrick (lex::*)(void *), actions *,
		     error *);

virtual	int   Success(register stack *s, register lex *l,
		      register ubrick (lex::*lx)(void *),
                      register actions *t, register error *e);
/////////////////////////////////

//
// engine (real-time parse)
//
/////////////////////////////////
virtual ubrick ParseInRealTime(stack *, lex *, ubrick (lex::*)(void *),
			       actions *, error *);

virtual	int   SuccessInRealTime(register stack *s, register lex *l,
				register ubrick (lex::*lx)(void *),
				register actions *t, register error *e);
/////////////////////////////////

//
// generator methods
//
/////////////////////////////////
#ifdef GENON
 virtual int   GenerateFromTo(lex *, ubrick from, ubrick to, long length,
			      long number, ostream &output);

 virtual int   Generate(lex *, long length, long number, ostream &output);
#endif
/////////////////////////////////

//
// Translation methods
//
/////////////////////////////////

ubrick TranslateToken(ubrick);
/////////////////////////////////

//
// used in error recovery
//
/////////////////////////////////
virtual int   FindEdgeAndMakeTransition(stackElement **stack, // find symbol,
					ubrick symbol);// transit, update stack

virtual int   FindEdge(ubrick state, ubrick symbol);   // find symbol only
/////////////////////////////////


//
// debugging methods
//
/////////////////////////////////
void  DebugSet(register ostream *x) { Flag |= F_DEBUG; Cerr = x; }
static  void  DebugSet()      { Flag |=  F_DEBUG; }
static  void  DebugReset()    { Flag &= ~F_DEBUG; }
void  ActionSet()     { Flag |=  F_ACTION; }
void  ActionReset()   { Flag &= ~F_ACTION; }
void  StackReset()    { Stack->Reset(StartState); }

void  Debug(stackElement *);		      // printout shift/goto
void  Debug(ubrick);            	      // printout reduction
void  PrintGoto();			      // printout goto 
void  DebugError(stackElement *);	      // printout error condition
void  DebugSkip(brick);         	      // skip next symbol on error
void  DebugBack(stackElement *);	      // pop the stack on error
void  DebugWait();			      // wait for symbol after error
void  DebugRecover(stackSave *, stackSave *); // recover in real time parsing
void  DebugPostPone(stackElement *, brick);   // printout postponed message
void  DebugLex(char *, char *);		      // indicate lexer has been called
/////////////////////////////////

//
// find textual representation of the symbol
//
/////////////////////////////////
keyWord *Find(register ubrick symbol);
/////////////////////////////////

//
// store the attribute
//
/////////////////////////////////
	void  StoreAtr(register stackElement *s, register void *a,
		       register ubrick t)
	{
	 moveAttribute(Tokens + t, s->Atribut, a);
	}
/////////////////////////////////
};
//////////////////////////////////////////////////////////////////////////


//
// mskx
//
//////////////////////////////////////////////////////////////////////////
class	mskx      	// entry point to all the parsing data
{
public:
	lr1Tables *Lr1Tables;
	stack     *Stack;
	lex       *Lex;
	ubrick     (lex::*Lx)(void *);
	actions   *Node;
	error	  *Error;
	mskx(register lr1Tables *lr1,
	     register stack *s,
	     register lex *l,
	     register ubrick (lex::*lx)(void *),
	     register actions *n,
	     register error *e)
	{
	 Lr1Tables = lr1;
	 Stack     = s;
	 Lex       = l;
	 Lx        = lx;
	 Node 	   = n;
	 Error     = e;
	 Node->Lex = l;
	}

//
// non-real time parsing
//
///////////////////
	int Success();
//
// real-time parsing
//
///////////////////
	int SuccessInRealTime();
//
// generator
//
///////////////////
#ifdef GENON
	int Generate(long length, long number, ostream &f);
#endif
///////////////////
};
//////////////////////////////////////////////////////////////////////////


//
// typeDescriptor
//
//////////////////////////////////////////////////////////////////////////
class   typeDescriptor                  // element of the array of types
{
public:
	char     *TypeName;             // name of type
	mskx     *Tables;               // address of tables
};
//////////////////////////////////////////////////////////////////////////


//
// SHIFT
// REDUCE2
//
//////////////////////////////////////////////////////////////////////////
#define SHIFT(out, x, h) if ( out[transit+(x-h)] == token )	   \
			 {                                         \
			  (++stck)->State=out[transit+(x-h+1)];    \
			  continue;                                \
			 }

#define REDUCE2(x, t, ex) { register ubrick *u = Reduce2Table + x;	\
			    x = *u++;					\
			    register ubrick c = *u++;			\
			    while ( c )					\
			    {						\
			     do						\
			       if ( *u++ == t )				\
				  goto ex;				\
			     while ( --c );				\
				  x = *u++;			    	\
				  c = *u++;			   	\
				 }				   	\
				 ex:;					\
			  }
//////////////////////////////////////////////////////////////////////////

//
// mylex
//
//////////////////////////////////////////////////////////////////////////
class myLex : public lex
{
public:
					// define this for custom lexical
					// analysis
virtual ubrick PureUserLex(void *);

	myLex(lex *l) : lex() { *(lex *)this = *l; }	// constructor
};
//////////////////////////////////////////////////////////////////////////


//
// myerror
//
//////////////////////////////////////////////////////////////////////////
class myError : public error
{
public:
					// define this for custom error
					// recovery
virtual int UserRecover(lr1Tables *, stack *, lex *,
			ubrick (lex::*)(void *), ubrick *);

					// define this for custom error							// reports
virtual void ReportError(lr1Tables *, stack *, lex *, ubrick *);

	myError(register error *e) : error() 		// constructor
	       { *(error *)this = *e; Flag &= ~F_NOSKIP; }

virtual int ErrorLimit();		// define this for custom error
					// limit
};
//////////////////////////////////////////////////////////////////////////

#endif
